#!/usr/bin/perl -w

# Copyright 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use Carp 'croak';
use List::Util 'max','min','sum','pairmap';
use Math::BaseCnv 'cnv';
use POSIX 'floor';
use Math::PlanePath::Base::Digits 'bit_split_lowtohigh';

use Math::PlanePath::AlternatePaper 118; # v.118 for xyxy_to_n_either()
use Math::PlanePath::Base::Digits
  'bit_split_lowtohigh', 'digit_split_lowtohigh',
  'digit_join_lowtohigh', 'round_down_pow';
use Math::PlanePath::Base::Generic
  'round_nearest';

use FindBin;
use lib "$FindBin::Bin/../tools";
use MyAlternate ':all';
use MyFLAT;
use MyFSM;
our @ipc;
$|=1;

# uncomment this to run the ### lines
use Smart::Comments;

{
  # permute LtoH state numbers
  my @old_state_table   = ([0,0],      # 0 = unused
                           [1,26],     # 1 = 2a
                           [10,14],    #   2 = 3d
                           [8,11],     #   3 = 4a
                           [4,3],      # 4 = 1a
                           [1,9],      #   5 = 2b
                           [8,11],     #   6 = 3g
                           [2,18],     #   7 = 4g
                           [4,12],     # 8 = 1b
                           [5,22],     #   9 = 2c
                           [10,21],    #   10 = 3f
                           [19,15],    #   11 = 4c
                           [8,11],     # 12 = 1c
                           [5,17],     #   13 = 2e
                           [2,13],     #   14 = 3e
                           [8,16],     #   15 = 4b
                           [19,15],    # 16 = 1d
                           [2,18],     #   17 = 2d  -> 3d,3b
                           [5,17],     #   18 = 3b
                           [23,27],    #   19 = 4d
                           [8,16],     # 20 = 1e
                           [2,13],     #   21 = 2f
                           [2,18],     #   22 = 3c
                           [23,24],    #   23 = 4f
                           [19,20],    # 24 = 1f
                           [0,0],      #   25 unused
                           [5,22],     #   26 = 3a
                           [19,20],    #   27 = 4e
                           [4,6],      # 28 = T
                           [5,7],      # 29 = U
                           [28,29]);   # 30 = S

  my @old = (undef, '2a','3d','4a',
             '1a', '2b','3g','4g',   # this row starts for low bits 0..3
             '1b', '2c','3f','4c',
             '1c', '2e','3e','4b',
             '1d', '2d','3b','4d',
             '1e', '2f','3c','4f',
             '1f',undef,'3a','4e',
             'T','U','S');
  my @new = (undef, '2a','3a','4a',
             '1a', '2b','3g','4g',  # this row starts for low bits 0..3
             '1b', '2c','3b','4c',
             '1c', '2d','3c','4b',
             '1d', '2e','3d','4d',
             '1e', '2f','3e','4f',
             '1f',undef,'3f','4e',
             'T','U','S');

  my %old_name_to_i = (map {defined $old[$_] ? ($old[$_] => $_) : ()} 0 .. $#old);
  my %new_name_to_i = (map {defined $new[$_] ? ($new[$_] => $_) : ()} 0 .. $#new);
  my %name_old_to_new;
  foreach my $i (1 .. $#old) {
    defined $old[$i] or next;
    die if defined $name_old_to_new{$old[$i]};
    $name_old_to_new{$old[$i]} = $new[$i];
  }
  my %name_new_to_old;
  foreach my $i (1 .. $#new) {
    defined $new[$i] or next;
    die if defined $name_new_to_old{$new[$i]};
    $name_new_to_old{$new[$i]} = $old[$i];
  }
  require Clone;
  my @new_table;
  foreach my $old_i (0 .. $#old_state_table) {
    my $from_name = $old[$old_i];
    if (! defined $from_name) {
      die if defined $new_table[$old_i];
      $new_table[$old_i] = [0,0];
      next;
    }
    ### $old_i
    ### $from_name

    my @row = @{$old_state_table[$old_i]};
    ### @row
    foreach my $bit (0,1) {
      my $old_to = $row[$bit];
      next if $old_to==0;
      my $to_name = $old[$old_to];
      my $new_to = $new_name_to_i{$to_name};
      $row[$bit] = $new_to;
      ### $old_to
      ### $to_name
      ### $new_to
    }

    my $new_i = $new_name_to_i{$from_name};
    die if defined $new_table[$new_i];
    $new_table[$new_i] = \@row;
    ### @row
    ### $new_i
  }

  my $prefix = '  = (';
  foreach my $i (0 .. $#new_table) {
    my $row = $new_table[$i];
    my $str = join(',',@$row);
    my $comma = ($i < $#new_table ? ',' : ');');
    printf "%s%-10s # ", $prefix, "[$str]$comma";
    if ($i) {
      if ($i%4!=0) { print "  "; }
    }
    my $name = $new[$i] // 'unused';
    print "$i = $name";
    $prefix = '     ';
    if ($row->[0]) {
      print "  -> ",join(',',map {$new[$_]} @$row);
    }
    print "\n";
  }
  exit 0;
}
{
  # A073057 concatenated morphisms, MyFSM union
  my $f = A073057_flat(1);
  MyFLAT::FLAT_print_tikz($f, flow => 'east');
  $f->MyFLAT::view;
  exit 0;
}

{
  # A073057 concatenated morphisms, MyFSM union
  my @fs;
  my @tables;
  foreach my $a (1..4) {
    my $f = A073057_flat($a);
    $f = $f->MyFLAT::reverse;
    $f = $f->intersect(FLAT::Regex->new ('(0|1)(0|1) (0|1)*')->as_dfa);
    $f = $f->MyFLAT::minimize(verbose=>1);
    $fs[$a] = $f;

    my @table;
    foreach my $state ($f->get_states) {
      foreach my $bit (0,1) {
        my @to = $f->successors($state,$bit);
        @to==1 or die "oops, not a DFA";
        $table[$state]->[$bit] = $to[0];
      }
    }
    $tables[$a] = \@table
  }

  my $flip = 0;
  my %table;
  my %accepting;
  my %seen;
  my @pending = ([(map {$fs[$_]->get_starting} 1 .. 4),
                  ($flip ? (0) : ())]);
  my $initial;
  while (@pending) {
    my $statearef = shift @pending;
    my $key = join(',',@$statearef);
    # ### $statearef
    # ### $key
    my $parity;
    if ($flip) { $parity = pop @$statearef; }
    @$statearef == 4 or die;
    next if $seen{$key}++;
    $initial //= $key;
    my @accepting;
    foreach my $i (0 .. 3) {
      my $a = $i+1;
      if ($fs[$a]->is_accepting($statearef->[$i])) {
        push @accepting, $a;
      }
    }
    ### @accepting
    @accepting <= 1 or die;
    if (@accepting) {
      $accepting{$key} = "r$accepting[0]";
    }
    foreach my $bit (0,1) {
      my $bit = $bit;
      if ($flip && $parity) { $bit = 1-$bit; }
      my @to = map {my $a = $_+1;
                    my $from = $statearef->[$_];
                    $tables[$a]->[$from]->[$bit]} 0 .. 3;
      if ($flip) {
        $parity = 1-$parity;
        push @to, 1-$parity;
      }
      my $to_key = join(',',@to);
      $table{$key}->{$bit} = $to_key;
      push @pending, \@to;
    }
  }
  ### %table
  ### %accepting

  my $fsm = MyFSM->new
    (table      => \%table,
     initial    => $initial,
     accepting  => \%accepting,
     title      => "LtoH",
     flow       => 'west',
    );

  $fsm->simplify (verbose => 1);

  $fsm->renumber_states;
  ($initial) = $fsm->initial_states;
  ### $initial
  # single 0-bit not accepting
  delete $fsm->{'accepting'}->{$fsm->transition($fsm->initial_states,0)};
  delete $fsm->{'accepting'}->{$fsm->initial_states};

  # accepting type within name
  foreach my $state ($fsm->accepting_states) {
    if (defined (my $accepting = $fsm->{'accepting'}->{$state})) {
      $fsm->rename_state($state, "$state $accepting");
    }
  }

  # $fsm->view;
  print "num states ",$fsm->num_states,"\n";

  my %state_name_to_num;
  my @state_num_to_name;
  my $set_name_to_num = sub {
    my ($state, $num) = @_;
    if (defined $state_name_to_num{$state}) { die; }
    if (defined $state_num_to_name[$num]) { die "oops duplicate num=$num"; }
    $state_name_to_num{$state} = $num;
    $state_num_to_name[$num] = $state;
  };
  foreach my $a (1 .. 4) {
    ### $a
    {
      my $s = $fsm->transition($initial,($a-1)&1);
      $s = $fsm->transition($s,($a-1)>>1);
      $set_name_to_num->($s, ($a-1)+4);
    }
    my $upto = ($a-1) % 4;
    if ($upto == 0) { $upto += 4; }
    foreach my $state ($fsm->accepting_states) {
      next if defined $state_name_to_num{$state};
      next unless $fsm->{'accepting'}->{$state} eq "r$a";
      ### $upto
      if (defined $state_num_to_name[$upto]) { $upto += 4; }
      $set_name_to_num->($state, $upto);
      $upto += 4;
    }
  }
  {
    my $upto = 28;
    foreach my $state (sort $fsm->non_accepting_states) {
      $set_name_to_num->($state, $upto++)
    }
  }

  my @table;
  foreach my $state ($fsm->states) {
    my $from = $state_name_to_num{$state};
    foreach my $bit (0,1) {
      my $to = $fsm->transition($state,$bit);
      $table[$from]->[$bit] = $state_name_to_num{$to};
    }
  }
  foreach my $i (0 .. $#table) {
    $table[$i] //= [0,0];
  }
  my @ret = ('') x scalar(@table);
  foreach my $state ($fsm->accepting_states) {
    my $num = $state_name_to_num{$state};
    $ret[$num] = substr($fsm->{'accepting'}->{$state}, 1);
  }
  ### @ret

  foreach my $a (1 .. 4) {
    my @accepting = map {$fsm->{'accepting'}->{$_} eq "r$a"
                           ? ($state_name_to_num{$_} // die)
                           : ()}
      $fsm->accepting_states;
    my $f = MyFLAT::aref_to_FLAT_DFA
      (\@table,
       name      => "again a=$a",
       starting  => $state_name_to_num{$initial},
       accepting_list => \@accepting);
    ### @accepting
    # $f->MyFLAT::view (flow => 'west');

    MyFLAT::FLAT_check_is_equal($f,
                                $fs[$a]);
    # $f->equals() or die;
  }

  # print "states: ",join(', ',@state_names),"\n";
  # print "accepting: ";
  # foreach my $i (0 .. $#table) {
  #   my $acc;
  #   if (defined (my $state = $state_num_to_name[$i])) {
  #     $acc = $fsm->{'accepting'}->{$_};
  #   }
  #   $acc //= '  ';
  #   print $acc,', ';
  # }
  # print "\n";

  my @accepting_histogram;
  foreach my $accepting (values %{$fsm->{'accepting'}}) {
    $accepting_histogram[substr($accepting,1)]++;
  }
  ### @accepting_histogram

  # print "initial ",$initial[0],"\n";
  $#table == 30 or die;  # 28,29,30
  foreach my $i (1 .. 27) {
    if ($i == 8 || $i == 18) { print "\n"; };
    print $table[$i]->[0],",",$table[$i]->[1],";";
  }
  print "\n";

  if (0) {
    my $tf = MyFLAT::aref_to_FLAT_DFA
      (\@table,
       starting  => $state_name_to_num{$initial});
    MyFLAT::FLAT_print_tikz($tf, flow => 'west');
  }

  # 6,6,7,7

  # 1..6
  # 7..12
  # 14..20
  # 21..28

  my %tfsm_accepting = map {$_ => int($_/7)+1} 0 .. 27;
  my $tfsm = MyFSM->new
    (table_aref      => \@table,
     accepting  => \%tfsm_accepting,
     title      => "LtoH",
     flow       => 'west',
    );
  # $tfsm->view;

  exit 0;
}

{
  # A073057 concatenated morphisms MyFSM union
  my @HtoL_table = ([0,0],
                    [1,2],
                    [1,3],
                    [4,2],
                    [4,3],
                    [3,4],   # 5
                    [1,5]);  # 6
  my @fsm;
  foreach my $a (1..4) {
    my $f = MyFLAT::aref_to_FLAT_DFA
      (\@HtoL_table,
       name      => "a=$a",
       starting  => 6,
       accepting => $a);
    $f = $f->MyFLAT::reverse;
    $f = $f->MyFLAT::minimize(verbose=>1);

    my @table;
    foreach my $state ($f->get_states) {
      foreach my $bit (0,1) {
        my @to = $f->successors($state,$bit);
        @to==1 or die "oops, not a DFA";
        $table[$state]->[$bit] = $to[0];
      }
    }
    $fsm[$a] = MyFSM->new
      (table_aref => \@table,
       initial    => 6,
       accepting  => { map {$_=> "acc$a"} $f->get_accepting },
       title      => "a=$a",
       flow       => 'west',
      );
  }
  my $fsm = $fsm[1];
  foreach my $a (2 .. 4) {
    $fsm = $fsm->union($fsm[$a], type => 'merge');
    $fsm->{'flow'} = 'west';
    $fsm->simplify;
    $fsm->renumber_states;
  }
  ### accepting: $fsm->{'accepting'}

  $fsm->view;
  exit 0;
}
{
  # A073057 concatenated morphisms
  require MyFLAT;
  my @HtoL_table = ([0,0],
                    [1,2],
                    [1,3],
                    [4,2],
                    [4,3],
                    [3,4],   # 5
                    [1,5]);  # 6
  foreach my $a (1..4) {
    my $f = MyFLAT::aref_to_FLAT_DFA
      (\@HtoL_table,
       name      => "a=$a",
       starting  => 6,
       accepting => $a);
    $f = $f->MyFLAT::reverse;
    $f = $f->MyFLAT::as_nfa;

    # my $to = [$f->get_starting];
    # $f->unset_starting(@$to);
    # $to = [$f->successors($to,[0])];
    # $to = [$f->successors($to,[0])];
    # $f->set_starting(@$to);

    $f = $f->MyFLAT::minimize(verbose=>1);
    $f->MyFLAT::view;
  }
  exit 0;
}
{
  # A073057 concatenated morphisms MyFSM
  require MyFSM;
  my @HtoL_table = ([0,0],
                    [1,2],
                    [1,3],
                    [4,2],
                    [4,3],
                    [3,4],   # 5
                    [1,5]);  # 6
  my $fsm = MyFSM->new
    (table_aref => \@HtoL_table,
     initial    => 6,
     accepting  => { 1 => 'ret1',
                     2 => 'ret2',
                     3 => 'ret3',
                     4 => 'ret4' },
     title      => 'A073057',
     flow       => 'east',
    );
  $fsm->delete_unreached_states;
  print "accepting: ",$fsm->accepting_states,"\n";
  $fsm = $fsm->reverse;
  print "accepting: ",$fsm->accepting_states,"\n";
  # $fsm->simplify;
  # $fsm->renumber_states;

  $fsm->view;
  exit 0;
}
