#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.010;
use strict;
use Carp 'croak';
use List::Util 'max','min','sum';
use Math::BaseCnv 'cnv';

use FLAT;
use FindBin;
use lib "$FindBin::Bin/../tools";
use MyAlternate ':all';
use MyPrintwrap ':all';
use MyFLAT;

# uncomment this to run the ### lines
# use Smart::Comments;

# Philip Lafrance, Narad Rampersad, Randy Yee, "Some Properties of a
# Rudin-Shapiro-Like Sequence", arxiv 1408.2277

my @state_table = ([0,1],
                   [2,0],
                   [1,3],
                   [3,2]);
my $F = MyFLAT::aref_to_FLAT_DFA
  (\@state_table,
   name => 'in',
   accepting_list => [0,1]);
$F->{'flow'} = 'east';

sub in {
  my ($n) = @_;
  return $F->contains(reverse(sprintf('%b',$n))) ? 1 : -1;
}
{
  # A246016
  my @want = (1,1,-1,1,1,-1,1,1,-1,1,-1,-1,1,1,-1,1,1,-1,1,1,-1);
  my $want = join(',',@want);
  my $got = join(',',map {in($_)} 0 .. $#want);
  ### want: $want
  ### got : $got
  $want eq $got or die;
}


{
  # PmOneBits as dir, overlaps
  my %seen;
  my $x = 0;
  my $y = 0;
  foreach my $n (0 .. 200) {
    my $dir4 = PmOneBits($n) & 3;
    my $dx = $dir4_to_dx[$dir4];
    my $dy = $dir4_to_dy[$dir4];
    my $seg = xyxyseg($x,$y, $x+$dx,$y+$dy);
    my $dup = $seen{$seg};
    print "n=$n $seg ",($dup ? "duplicate $seen{$seg}" : ""),"\n";
    # $seen{$seg} //= "n=$n";
    $seen{$seg}++;
    $x += $dx;
    $y += $dy;
  }
  exit 0;

  sub xyxyseg {
    my ($x1,$y1, $x2,$y2) = @_;
    if ($x1 > $x2) {
      ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
    } elsif ($y1 > $y2) {
      ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
    }
    return "$x1,$y1,$x2,$y2";
  }
}

{
  # i(n) drawing, as dx or dy alternately
  # A246016
  my %seen;
  my $x = 0;
  my $y = 0;
  my $join = '';
  $Printwrap = 2;
  printwrap('\\draw ');
  $Printwrap += 2;
  foreach my $n (0 .. 2**7) {
    my $in = in($n);
    my $dx = ($n&1 ? 0 : $in);
    my $dy = ($n&1 ? $in : 0);
    my $seg = "$x,$y,$dx,$dy";
    my $dup = $seen{$seg};
    $seen{$seg}++;

    my ($rdx,$rdy) = ($dup ? xy_rotate_plus90($dx,$dy) : (0,0));

    my $sx = $x . chamfer($dup ? 2*$dx + $rdx : $dx);
    my $sy = $y . chamfer($dup ? 2*$dy + $rdy : $dy);
    $x += $dx;
    $y += $dy;
    my $ex = $x . chamfer($dup ? 2*-$dx + $rdx : -$dx);
    my $ey = $y . chamfer($dup ? 2*-$dy + $rdy : -$dy);
    printwrap("$join($sx,$sy)--($ex,$ey)");
    $join = '--';
    # $seen{$seg} //= "n=$n";
  }
  printwrap(";\n");
  exit 0;
}


{
  # i(n) as dx,dy overlaps
  my %seen;
  my $x = 0;
  my $y = 0;
  foreach my $n (0 .. 200) {
    my $in = in($n);
    my $dx = ($n&1 ? 0 : $in);
    my $dy = ($n&1 ? $in : 0);
    my $seg = "$x,$y,$dx,$dy";
    my $dup = $seen{$seg};
    print "n=$n $seg ",($dup ? "duplicate $seen{$seg}" : ""),"\n";
    # $seen{$seg} //= "n=$n";
    $seen{$seg}++;
    $x += $dx;
    $y += $dy;
  }
  exit 0;
}

{
  my $f = GRS_flat(1);
  $f->{'flow'} = 'east';

  $f = $F;
  $f->MyFLAT::view;

  $f = GRS_flat(1)->MyFLAT::transmute(func => sub {
                                        my ($state,$bit) = @_;
                                        my $new_bit = $bit;
                                        if ($state) { $new_bit = 1-$new_bit; }
                                        $state = $bit;
                                        return ($state, $new_bit);
                                      },
                                      # direction => 'reverse',
                                      initial => 0)
    ->MyFLAT::minimize
    ;

  $f->MyFLAT::view;
  exit 0;
}

{
  require FLAT;
  require MyFLAT;
  my $f = $F;

  # $f->MyFLAT::view;

  # LtoH is 0<->1 bit flip of HtoL
  $f = $f->MyFLAT::reverse->MyFLAT::minimize;
  $f->{'flow'} = 'west';
  $f->MyFLAT::view;

  # $f = $f->MyFLAT::transmute(func => sub {
  #                              my ($state,$bit) = @_;
  #                              my $new_bit = $bit ^ $state;
  #                              return ($bit, $new_bit);
  #                            },
  #                            direction => 'reverse',
  #                            initial => 0)
  #   # ->MyFLAT::reverse
  #   ->MyFLAT::minimize;
  # 
  # $f->MyFLAT::view;

  exit 0;
}


