/* Copyright 2017, 2018, 2019 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/


/* Work in progress, possibly not working, going to change wildly.

*/

/* assert()s and similar check code disabled by default.
   Compile with -DNAUTYEXTRA_WANT_ASSERTS=1 to have asserts.
*/
#if ! NAUTYEXTRA_WANT_ASSERTS
#if ! defined(NDEBUG)  /* if NDEBUG not already set */
#define NDEBUG
#endif
#endif
#include <assert.h>

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/naugroup.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

/* Change WANT_DEBUG to 1 for mucho debug prints. */
#ifndef WANT_DEBUG
#define WANT_DEBUG  0
#endif
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

/* Default is to detect overflows in various counts, Wiener index, etc.
   Compile with -DNAUTYEXTRA_WANT_OVERFLOW_DETECTION=0 to disable.
*/
#if ! defined(NAUTYEXTRA_WANT_OVERFLOW_DETECTION)
#define NAUTYEXTRA_WANT_OVERFLOW_DETECTION 1
#endif


/* ------------------------------------------------------------------------- */
/* Macros Used Here */

#define numberof(array)  (sizeof(array)/sizeof((array)[0]))

/* ss is a setword variable "setword ss".
   n is an int number of bits to have in ss.
   Loop over all sets of n bits in ss.
   Evaluate "body" with successive ss.
   body can goto or return etc to abort the loop.
*/
#define NAUTYEXTRA_SET1_LOOP(ss,n,body)                 \
  do {                                                  \
    assert (n >= 1);                                    \
    if (! (SETWORDSNEEDED(n) == 1)) abort();            \
    setword __NAUTYEXTRA_SET_LOOP__inc = BITT[n-1];     \
    (ss) = 0;                                           \
    do {                                                \
      body;                                             \
    } while ((ss) += __NAUTYEXTRA_SET_LOOP__inc);       \
  } while (0)

static void
INTERNAL_overflow (void)
{
  fprintf(stderr, "nautyextra integer overflow\n");
  abort();
}

/* dst,x,y are NAUTYEXTRA_TYPE_ULL variables.
   Set dst = x+y, with abort on overflow. */
#define ADD_NO_OVERFLOW_ULL(dst, x,y)                                   \
  do {                                                                  \
    NAUTYEXTRA_TYPE_ULL __ADD_NO_OVERFLOW_ULL__x = (x);                 \
    NAUTYEXTRA_TYPE_ULL __ADD_NO_OVERFLOW_ULL__y = (y);                 \
    NAUTYEXTRA_TYPE_ULL __ADD_NO_OVERFLOW_ULL__ret                      \
      = __ADD_NO_OVERFLOW_ULL__x + __ADD_NO_OVERFLOW_ULL__y;            \
    (dst) = __ADD_NO_OVERFLOW_ULL__ret;                                 \
    if (NAUTYEXTRA_WANT_OVERFLOW_DETECTION) {                           \
      if (__ADD_NO_OVERFLOW_ULL__ret < __ADD_NO_OVERFLOW_ULL__x)        \
        INTERNAL_overflow();                                          \
    }                                                                   \
  } while (0)

/* dst,x,y are NAUTYEXTRA_TYPE_ULL variables.
   Set dst = x*y, with abort on overflow.
   ENHANCE-ME: Division to detect overflow will be a bit slow.
   Maybe a mult in floating point would be better.
*/
#define MUL_NO_OVERFLOW_ULL(dst, x,y)                                   \
  do {                                                                  \
    NAUTYEXTRA_TYPE_ULL __MUL_NO_OVERFLOW_ULL__x = (x);                 \
    NAUTYEXTRA_TYPE_ULL __MUL_NO_OVERFLOW_ULL__y = (y);                 \
    NAUTYEXTRA_TYPE_ULL __MUL_NO_OVERFLOW_ULL__ret                      \
      = __MUL_NO_OVERFLOW_ULL__x * __MUL_NO_OVERFLOW_ULL__y;            \
    (dst) = __MUL_NO_OVERFLOW_ULL__ret;                                 \
    if (NAUTYEXTRA_WANT_OVERFLOW_DETECTION) {                           \
      if (__MUL_NO_OVERFLOW_ULL__x                                      \
          && __MUL_NO_OVERFLOW_ULL__ret / __MUL_NO_OVERFLOW_ULL__x      \
          != __MUL_NO_OVERFLOW_ULL__y)                                  \
        INTERNAL_overflow();                                          \
    }                                                                   \
  } while (0)

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))
#define MIN3(x,y,z)  MIN(MIN(x,y), z)


/* ------------------------------------------------------------------------- */
/* Static Workspace (thread local) */

DYNALLSTAT(char,INTERNAL_workspace,INTERNAL_workspace_sz);

#define WORKSPACE(type,var,size,name)                          \
  DYNALLOC1(char,                                              \
            INTERNAL_workspace,INTERNAL_workspace_sz,          \
            (size)*sizeof(type), name);                        \
  type *var = (type*) INTERNAL_workspace;

#define WORKSPACE2(type1,var1,size1, type2,var2,size2, name)                  \
  int __WORKSPACE2__alloc1 = (size1)*sizeof(type1);                           \
  int __WORKSPACE2__alloc2 = (size2)*sizeof(type2);                           \
  DYNALLOC1(char,                                                             \
            INTERNAL_workspace,INTERNAL_workspace_sz,                         \
            __WORKSPACE2__alloc1 + __WORKSPACE2__alloc2, name);               \
  type1 *var1 = (type1*) INTERNAL_workspace;                                  \
  type2 *var2 = (type2*) (INTERNAL_workspace + __WORKSPACE2__alloc1);

#if ! MAXN

#define WORKSPACE3(type1,var1,size1,                                          \
                   type2,var2,size2,                                          \
                   type3,var3,size3,                                          \
                   name)                                                      \
  int __WORKSPACE3__alloc1 = (size1)*sizeof(type1);                           \
  int __WORKSPACE3__alloc2 = (size2)*sizeof(type2);                           \
  int __WORKSPACE3__alloc3 = (size3)*sizeof(type3);                           \
  int __WORKSPACE3__total =                                                   \
    __WORKSPACE3__alloc1 + __WORKSPACE3__alloc2 + __WORKSPACE3__alloc3;       \
  DYNALLOC1(char, INTERNAL_workspace,INTERNAL_workspace_sz,                   \
            __WORKSPACE3__total, name);                                       \
  type1 *var1 = (type1*) INTERNAL_workspace;                                  \
  type2 *var2 = (type2*) (INTERNAL_workspace + __WORKSPACE3__alloc1);         \
  type3 *var3 = (type3*) (INTERNAL_workspace + __WORKSPACE3__alloc1           \
                          + __WORKSPACE3__alloc2);

#define WORKSPACE4(type1,var1,size1,                                          \
                   type2,var2,size2,                                          \
                   type3,var3,size3,                                          \
                   type4,var4,size4,                                          \
                   name)                                                      \
  int __WORKSPACE4__alloc1 = (size1)*sizeof(type1);                           \
  int __WORKSPACE4__alloc2 = (size2)*sizeof(type2);                           \
  int __WORKSPACE4__alloc3 = (size3)*sizeof(type3);                           \
  int __WORKSPACE4__alloc4 = (size4)*sizeof(type4);                           \
  int __WORKSPACE4__total =                                                   \
    __WORKSPACE4__alloc1 + __WORKSPACE4__alloc2                               \
    + __WORKSPACE4__alloc3 + __WORKSPACE4__alloc4;                            \
  DYNALLOC1(char, INTERNAL_workspace,INTERNAL_workspace_sz,                   \
            __WORKSPACE4__total, name);                                       \
  type1 *var1 = (type1*) INTERNAL_workspace;                                  \
  type2 *var2 = (type2*) (INTERNAL_workspace + __WORKSPACE4__alloc1);         \
  type3 *var3 = (type3*) (INTERNAL_workspace + __WORKSPACE4__alloc1           \
                          + __WORKSPACE4__alloc2);                            \
  type4 *var4 = (type4*) (INTERNAL_workspace + __WORKSPACE4__alloc1           \
                          + __WORKSPACE4__alloc2 + __WORKSPACE4__alloc3);

#endif

/* ------------------------------------------------------------------------- */
/* Diagnostics */

int nautyextra_want_progress = 0;
int nautyextra_progress_reps = 1;

static time_t prev_progress_time;
static int reps_remaining = 0;
int
nautyextra_progress(void)
{
  const int delay = 3;
  assert (nautyextra_progress_reps > 0);
  assert (reps_remaining >= 0);

  if (reps_remaining > 0) {
    reps_remaining--;
    return FALSE;
  }

  time_t t;
  if (time(&t)==-1) return FALSE;
  time_t delta = t - prev_progress_time;
  if (delta >= 0) {
    if (delta > delay) {
      if (nautyextra_progress_reps > 1) {
        nautyextra_progress_reps >>= 1;
      }
    }
    if (delta < delay) {
      int new_reps = nautyextra_progress_reps << 1;
      if (new_reps > nautyextra_progress_reps)
        nautyextra_progress_reps = new_reps;
    }
  }
  reps_remaining = nautyextra_progress_reps;
  prev_progress_time = t;
  return TRUE;
}

/* print the adjacency matrix of g, as 0s and 1s */
void
nautyextra_print_matrix (const graph *g, int m,int n)
{
  set *row = (set*) g;
  assert (n >= 0);

  int from, to;
  for (from=0; from < n; from++, row+=m) {
    for (to=0; to < n; to++) {
      putchar(' ');
      putchar(ISELEMENT(row,to) ? '1' : '0');
    }
    putchar('\n');
  }
}

/* Check vpar[1..n] values.  Return TRUE if good or FALSE if not.
   Nothing is printed, and no abort()
*/
/* int */
/* nautyextra_validate_pred_vpar (const int *vpar, int n) */
/* { */
/*   return internal_check_vpar(vpar,n, 0); /\* no abort *\/ */
/* } */


/* ------------------------------------------------------------------------- */
/* Graphical Viewing */

/* TRUE means view synchronously, meaning the view functions wait for the
   viewer to exit before returning.  The default is asynchronous.
   When asynchronous be careful not to spawn too many views or you can fork
   bomb yourself.
   (Could put some sensible limit here as protection perhaps.)
*/
int nautyextra_view_synchronous = 0;

static void
view_graph_pl (char * const *argv)
{
  DEBUG(printf("view_graph_pl()\n"));
  int pid;
  fflush(NULL);
  pid = fork();
  if (pid == -1) {
    perror("fork");
    exit(1);
  }
  if (pid == 0) {
    DEBUG(printf(" child\n"));
    fflush(NULL);
    execvp("view-graph.pl", argv);
    perror("cannot exec view-graph.pl");
    _exit(1);
  }
  DEBUG(printf(" parent, child pid %d\n", pid));
  if (nautyextra_view_synchronous) {
    waitpid(pid,NULL,0);
  }
}

/* str is a string of graph6 or sparse6.
   View it using view-graph.pl which is an experiment in my Graph-Maker-Other.
*/
void
nautyextra_view_g6 (const char *str)
{
  char * argv[4];
  argv[0] = "view-graph.pl";
  argv[1] = "--graph6";
  argv[2] = (char*) str;
  argv[3] = NULL;
  view_graph_pl(argv);
}

static const char sparse6_empty_graph[] = ":?\n";

/* View g using view-graph.pl.

   FIXME: The intention is to notice and show directedness, but currently g
   should be undirected (arcs both ways for each edge).
*/
void
nautyextra_view (const graph *g, int m, int n)
{
  DEBUG(printf("nautyextra_view() %p %d,%d\n", g,m,n));
  nautyextra_view_g6 (n==0 ? sparse6_empty_graph
                      : ntos6((graph*)g,m,n));
}

/* View sg using view-graph.pl.

   FIXME: The intention is to notice and show directedness, but currently sg
   should be undirected (arcs both ways for each edge).
*/
void
nautyextra_view_sg (const sparsegraph *sg)
{
  DEBUG(printf("nautyextra_view_sg() %p %d\n", sg, sg->nv));
  nautyextra_view_g6 (sg->nv == 0 ? sparse6_empty_graph
                      : sgtos6((sparsegraph*)sg));
}

/* View vpar[1..n] using view-graph.pl.
   The directedness of vpar is included in the display.
*/
void
nautyextra_view_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_view_vpar() %p %d\n", vpar, n));
  char **argv = malloc((n+3)*sizeof(char*));
  argv[0] = "view-graph.pl";
  argv[1] = "--vpar";
  int pos = 2;
  int v;
  int free_pos = pos;
  for (v = 1; v <= n; v++) {
    char *str = malloc(10);
    sprintf(str, "%d", vpar[v]);
    argv[pos++] = str;
  }
  argv[pos] = NULL;

  view_graph_pl(argv);

  for ( ; free_pos < pos; free_pos++)
    free(argv[free_pos++]);
  FREES(argv);
}


/* ------------------------------------------------------------------------- */
/* Sparsegraph Construction
*/

/* No, can just inline.

   nautyextra_sg_realloc_nv() grows sg->v,sg->vlen and sg->d,sg->dlen arrays
   if necessary to accommodate the given n many vertices.
   If the arrays are big enough already then unchanged.
   If grown then the existing part is retained in the new allocation.
   The actual number of vertices in sg->nv is ignored and unchanged, just
   space made available for at least n.

   Cf nauty SG_ALLOC() which is free() and malloc() so discards existing
   contents.  SG_ALLOC() good for complete new creation, vs realloc here for
   growing.
*/
void
nautyextra_sg_realloc_nv (sparsegraph *sg, int n)
{
  assert (n >= 0);
  DYNREALLOC(size_t,sg->v,sg->vlen,n, "nautyextra_sg_realloc_nv");
  DYNREALLOC(int,   sg->d,sg->dlen,n, "nautyextra_sg_realloc_nv");
}

/* No, can just inline.

   nautyextra_sg_realloc_elen() grows sg->e,sg->elen if necessary so it has
   space for up to the given elen many arcs.
   If sg->elen is big enough already then unchanged.
   If grown then the existing part is retained in the new allocation.
   The actual current number of arcs sg->nde is ignored and unchanged, just
   space made available for at least elen.

   Cf nauty SG_ALLOC() is free() and malloc() rather than realloc(), so it
   doesn't keep existing contents.  SG_ALLOC() is good for complete new
   creation, vs realloc here for growing.
*/
void
nautyextra_sg_realloc_elen (sparsegraph *sg, size_t elen)
{
  DYNREALLOC(int, sg->e,sg->elen,elen, "nautyextra_sg_realloc_elen");
}


/* ------------------------------------------------------------------------- */
/* Leaf Vertices
   FIXME: "Leaf" for childless vs degree=1.
*/

/* Return the number of leaf vertices (degree 0 or 1) in vpar[1..n].
*/
int
nautyextra_leaf_count_vpar (const int *vpar, int n)
{
  int v;
  int count = 0;
#if MAXN
  int degrees[MAXN+1];    /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(int,degrees,n+1,"nautyextra_leaf_count_vpar");
#endif
  assert (n >= 0);
  nautyextra_degrees_array_vpar(vpar,n,degrees);

  for (v = 1; v <= n; v++)
    count += (degrees[v]<=1);
  return count;
}

/* Return the number of non-leaf vertices (degree >= 2) in vpar[1..n].
*/
int
nautyextra_nonleaf_count_vpar (const int *vpar, int n)
{
  assert (n >= 0);
  return n - nautyextra_leaf_count_vpar(vpar,n);
}


/* ------------------------------------------------------------------------- */
/* Set Operations */

int
nautyextra_set_is_subset (const set *subset, const set* superset, int m)
{
  for ( ; m>0; m--)
    if (~(*subset++) & (*superset++)) return FALSE;
  return TRUE;
}

/* ss is a setword of bits.  Return the position of its n'th non-zero element.
   n=0 is the first element, n=1 the second, etc.
   n must be in the range 0 <= n < POPCOUNT(ss);

   This is equivalent to doing n many nextelement(), but implemented by
   skipping whole word popcounts.
*/
int
nautyextra_set_nthelement1 (setword ss, int n)
{
  assert (n >= 0);
  int pos;
  do {
    assert (ss != 0);
    TAKEBIT(pos,ss);
  } while (--n >= 0);
  return pos;
}

/* ss is a set.  Return the position of its n'th non-zero element.
   n=0 is the first element, n=1 the second, etc.
   n must be in the range 0 <= n < setsize(ss,m).
   There is no m parameter, the search just continues until the n'th 1-bit.

   This is equivalent to doing n many nextelement(), but implemented by
   skipping whole word popcounts.
*/
int
nautyextra_set_nthelement (const set *ss, int n)
{
  assert (n >= 0);
  for (;;) {
    setword s = *ss++;
    int spop = POPCOUNT(s);
    if (n >= spop) {
      n -= spop;
    } else {
      int pos;
      do {
        TAKEBIT(pos,s);
      } while (--n >= 0);
      return pos;
    }
  }
}


/* ------------------------------------------------------------------------- */
/* Edges, Cyclic */

/* Return the number of self-loops in sg, ie. any arc v->v.
   sg can have multiple loops at the same v, all are counted.
   Cf loopcount() for graph.
   Cf gtnauty.c static hasloops(), hasloops_sg() predicates.
*/
int
nautyextra_loopcount_sg (const sparsegraph *sg)
{
  int            n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int count = 0;
  int from;
  for (from = 0; from < n; from++) {
    const int *eptr = ee + vv[from];
    int degree = dd[from];
    int i;
    for (i = 0; i < degree; i++) {
      int to = *eptr++;
      count += (to == from);
    }
  }
  return count;
}


/* ------------------------------------------------------------------------- */
/* Multi-Edges */

/* Return TRUE if sg has any multi-edge, ie. any repeated u->v.
   Return FALSE if not.

   ENHANCE-ME: Would it be faster to sort sub-arrays like sortlists_sg()
   rather than repeatedly EMPTYSET()?  For small degrees probably yes.
*/
int
nautyextra_is_multiedged_sg (const sparsegraph *sg)
{
  int           n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int m = SETWORDSNEEDED(n);
  int from;
#if MAXN
  set seen[SETWORDSNEEDED(MAXN)];
  assert (n <= MAXN);
#else
  WORKSPACE(set,seen,m,"nautyextra_is_multiedged_sg");
#endif

  for (from = 0; from < n; from++) {
    int degree = dd[from];
    if (degree >= 2) {
      const int *eptr = ee + vv[from];
      EMPTYSET(seen,m);
      int i;
      for (i = 0; i < degree; i++) {
        int to = *eptr++;
        if (ISELEMENT(seen,to)) return TRUE;
        ADDELEMENT(seen,to);
      }
    }
  }
  return FALSE;
}


/* ------------------------------------------------------------------------- */
/* Conversions */

/* Convert tree sg to vpar[1..n].
   vpar must have space for sg->nv+1 ints.
   sg must be a connected tree (not a forest) and undirected (so edges both
   u->v and v->u).

   Vertex numbers 0..n-1 in sg become 1..n in vpar.
   root is the vertex number in sg which will become the root as root+1 in vpar.

   FIXME: Allow directed sg by taking any u->v or v->u as an edge in vpar.
   Maybe first sg vertex with no out edges as root.

   FIXME: Allow sg forest.  Root as first each component ...

   FIXME: Notice directed so inverse of directed=1 vpar_to_sg.
 */
void
nautyextra_sg_to_vpar (const sparsegraph *sg, int *vpar, int root)
{
  int            n = sg->nv;
  DEBUG(printf("nautyextra_sg_to_vpar() n=%d\n", n));
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
#if MAXN
  int queue[MAXN];
  assert (n <= MAXN);
#else
  WORKSPACE(int,queue,n,"nautyextra_sg_to_vpar");
#endif
  memset(vpar, '\0', (n+1)*sizeof(vpar[0]));

  int *qhead = queue;
  int *qtail = queue;
  *qtail++ = root;

  while (qhead != qtail) {
    assert (qhead < queue+n);
    int from0 = *qhead++;
    int from1 = from0 + 1;
    int from_parent1 = vpar[from1];
    int degree = dd[from0];
    int epos   = vv[from0];
    int to_i;
    for (to_i = 0; to_i < degree; to_i++, epos++) {
      int to0 = ee[epos];
      assert (0 <= to0 && to0 < n);

      int to1 = to0 + 1;
      if (to1 != from_parent1) {
        /* to is a child */
        assert(vpar[to1] == 0);
        vpar[to1] = from1;
        assert (qtail < queue+n);
        *qtail++ = to0;
      }
    }
  }
}


/* Convert tree vpar[1..n] to graph g,m,n.
   g must have space for m,n vertices.  Any existing content is replaced.

   Edges are between child c and parent p.
   directed=0 gives an undirected sparsegraph, so arcs both c->p and p->c.
   directed=1 gives a directed sparsegraph with edges c->p, child to parent.
   directed=2 gives a directed sparsegraph with edges p->c, parent to child.
*/
void
nautyextra_vpar_to_graph (const int *vpar, graph *g, int m, int n, int directed)
{
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (directed==0 || directed==1 || directed==2);

  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g, m,n);

  int v;
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    assert (0 <= p && p <= n);
    if (p) {
      int v0 = v-1;
      int p0 = p-1;
      if (directed != 2) ADDONEARC(g, v0,p0, m);  /* up */
      if (directed != 1) ADDONEARC(g, p0,v0, m);  /* down */
    }
  }
}

/* Convert tree vpar[1..n] to sparsegraph sg.
   sg must be already SG_INIT but any existing content is realloced as
   necessary and replaced.

   Edges are between child c and parent p.
   directed=0 gives an undirected sparsegraph, so arcs both c->p and p->c.
   directed=1 gives a directed sparsegraph with edges c->p, child to parent.
   directed=2 gives a directed sparsegraph with edges p->c, parent to child.

   The arcs in sg are not sorted.  In the current code the blocks of arcs in
   sg->e are contiguous, but don't rely on that.

   cf gentreeg.c which generates vpar and writes sparse6.  Its WriteS6()
   prints sparse6 directly from vpar, without going through sparsegraph.
 */
void
nautyextra_vpar_to_sg (const int *vpar, int n, sparsegraph *sg, int directed)
{
  DEBUG(printf("nautyextra_vpar_to_sg() n=%d directed=%d\n", n, directed));
  DEBUG(nautyextra_print_vpar(vpar,n));
  assert (nautyextra_validate_vpar(vpar,n));
  assert (directed==0 || directed==1 || directed==2);

  SG_ALLOC(*sg,n,2*n, "nautyextra_vpar_to_sg");

  /* outdegrees of vertices,
     v --> parent and if undirected parent --> v,
     except no parent of root
  */
  int *dd = sg->d;
  memset (dd, '\0', n*sizeof(dd[0]));
  int v;
  int nde = 0;
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    assert (0 <= p && p <= n);
    if (p) {
      int v0 = v-1;
      int p0 = p-1;
      if (! (directed & 2)) {
        dd[v0]++;      /* v -> p */
        nde++;
      }
      if (! (directed & 1)) {
        dd[p0]++;      /* v <- p */
        nde++;
      }
    }
  }
  sg->nv = n;
  sg->nde = nde;

  /* positions within e for the arc destinations, spaced according to degrees
   */
  size_t *vv = sg->v;
  int epos = 0;
  for (v = 1; v <= n; v++) {
    int v0 = v-1;
    DEBUG(printf("  at v=%d v0=%d outdegree=%d  epos=%d (nde=%d)\n",
                 v, v0, dd[v0], epos, nde));
    vv[v0] = epos;
    epos += dd[v0];
    assert (epos <= nde);
  }

  /* store arc destinations, incrementing vv[v0] and vv[p0] for where to
     store next
  */
  int *ee = sg->e;
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    if (p) {
      int p0 = p-1;
      int v0 = v-1;
      if (! (directed & 2))
        ee[vv[v0]++] = p0;    /* v -> p */
      if (! (directed & 1))
        ee[vv[p0]++] = v0;    /* v <- p */
    }
  }

  /* step vv[v] back to the start of its arc block in ee */
  for (v = 0; v < n; v++)
    vv[v] -= dd[v];

  /* vv[v] stepped back above is same as spacing first established */
#ifndef NDEBUG
  epos = 0;
  for (v = 1; v <= n; v++) {
    int v0 = v-1;
    DEBUG(printf("  check v=%d v0=%d vv[%d]=%d cf epos=%d\n",
                 v,v0, v0,vv[v0], epos));
    assert (vv[v0] == epos);
    epos += dd[v0];
  }
#endif
}

/* Return a string of graph6 which is tree vpar[1..n].
   Vertex numbers 1..n in vpar become 0..n-1 in the graph6 numbering.
   graph6 is undirected so the root vertex (or root vertices of a forest) are
   not distinguished in the graph6.
   The string is static allocated per sgtog6() so good until the next such call.

   FIXME: Waiting for nautyextra_vpar_to_sg() ...
*/
const char *
nautyextra_vpar_to_g6 (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_vpar_to_g6() %p n=%d\n", vpar, n));
  char *ret;
  SG_DECL(sg);
  SG_INIT(sg);
  nautyextra_vpar_to_sg (vpar, n, &sg, 0); /* undirected */
  ret = sgtog6 (&sg);
  SG_FREE(sg);
  return ret;
}

/* static const char sparse6_empty_graph[] = ":?\n"; */

/* Return a string of sparse6 which is tree vpar[1..n].
   vpar vertex numbers 1..n become 0..n-1 in the sparse6 numbering.
   sparse6 is undirected so the root vertex (or root vertices of a forest) are
   not distinguished in the graph6.
   The string is static allocated per sgtos6() so good until the next such call.

   FIXME: Waiting for nautyextra_vpar_to_sg() ...
*/
const char *
nautyextra_vpar_to_s6 (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_vpar_to_s6() %p n=%d\n", vpar, n));

  /* sgtos6() doesn't like sg.nv==0 */
  if (n == 0) return sparse6_empty_graph;

  SG_DECL(sg);
  SG_INIT(sg);
  nautyextra_vpar_to_sg (vpar, n, &sg, 0);   /* undirected */
  DEBUG(printf("  sg %p %d\n", &sg, sg.nv));
  assert (nautyextra_validate_sg (&sg));
  char *ret = sgtos6 (&sg);
  SG_FREE(sg);
  return ret;
}

/* vpar[1..n] is a tree.
   Return it as a string of digraph6.
   Vertex numbers 1..n become 0..n-1 in the digraph6 numbering.
   The string is static allocated per sgtod6() so good until the next such call.

   Edges are between child c and parent p.
   directed=0 gives an undirected sparsegraph, so arcs both c->p and p->c.
   directed=1 gives a directed sparsegraph with edges c->p, child to parent.
   directed=2 gives a directed sparsegraph with edges p->c, parent to child.

   FIXME: Waiting for nautyextra_vpar_to_sg ...
*/
const char *
nautyextra_vpar_to_d6 (const int *vpar, int n, int directed)
{
  DEBUG(printf("nautyextra_vpar_to_s6() %p %d\n", vpar, n));
  SG_DECL(sg);
  SG_INIT(sg);
  nautyextra_vpar_to_sg (vpar, n, &sg, directed);
  DEBUG(printf("  sg %p %d\n", &sg, sg.nv));
  assert (nautyextra_validate_sg (&sg));
  char *ret = sgtod6 (&sg);
  SG_FREE(sg);
  return ret;
}


/* ------------------------------------------------------------------------- */
/* Isomorphic */

/* Copy g_in to g_out with a canonical labelling.
   This is a simplified interface to densenauty() just to canonicalize.

   FIXME: digraph changes the canonical labelling ...
*/
void
nautyextra_canonise (const graph *g_in, int m, int n, graph *g_out)
{
  DEFAULTOPTIONS_GRAPH(options);
  statsblk stats;
#if MAXN
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE3(int,lab,n,
             int,ptn,n,
             int,orbits,n,
             "nautyextra_canonise");
#endif

  /* densenauty() is the full nauty() but its own workspace allocation */
  assert(g_in != g_out);
  options.digraph  = ! nautyextra_is_undirected(g_in,m,n);
  options.getcanon = TRUE;
  densenauty ((graph*)g_in, lab, ptn, orbits, &options, &stats, m,n, g_out);

  /* nauty((graph*)g_in, lab, ptn, */
  /*       NULL,       /\* active *\/ */
  /*       orbits,&options,&stats, */
  /*       NULL, 0,    /\* workspace *\/ */
  /*       m, n, g_out); */
}

/* Return TRUE if g1 and g2 are isomorphic, or FALSE if not.

   FIXME: notice directedness?  Or digraph always might be a slowdown?
   Or some degree checks first for inequality?
*/
int
nautyextra_is_isomorphic (const graph *g1, const graph *g2, int m, int n)
{
  DYNALLSTAT(graph,canon_g1,canon_g1_sz);
  DYNALLSTAT(graph,canon_g2,canon_g2_sz);

  DYNALLOC2(graph,canon_g1,canon_g1_sz,m,n,"nautyextra_is_isomorphic");
  DYNALLOC2(graph,canon_g2,canon_g2_sz,m,n,"nautyextra_is_isomorphic");

  nautyextra_canonise(g1,m,n,canon_g1);
  nautyextra_canonise(g2,m,n,canon_g2);
  return nautyextra_aresame(canon_g1,canon_g2,m,n);
}

/* Set sg_out to a canonically labeled isomorphic copy of sg_in.
   Must have sg_out different from sg_in (no in-place conversion).

   This is a simplified interface to sparsenauty() (or fcanonise_inv_sg())
   just to canonicalize.

   FIXME: Create dynamic if sg_out==NULL?
   FIXME: digraph always might be a slowdown ...
*/
sparsegraph *
nautyextra_canonise_sg(const sparsegraph *sg_in, sparsegraph *sg_out)
{
  DEFAULTOPTIONS_SPARSEGRAPH(options);
  statsblk stats;
#if MAXN
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  assert(sg_in->nv <= MAXN);
#else
  DYNALLSTAT(int, lab,    lab_sz);
  DYNALLSTAT(int, ptn,    ptn_sz);
  DYNALLSTAT(int, orbits, orbits_sz);
  int n = sg_in->nv;
  DYNALLOC1(int, lab,    lab_sz,    n, "nautyextra_canonise_sg");
  DYNALLOC1(int, ptn,    ptn_sz,    n, "nautyextra_canonise_sg");
  DYNALLOC1(int, orbits, orbits_sz, n, "nautyextra_canonise_sg");
#endif

  DEBUG(printf("nautyextra_canonise_sg()\n"));
  assert(sg_in != sg_out);

  options.digraph  = ! nautyextra_is_undirected_sg(sg_in);
  options.getcanon = TRUE;
  sparsenauty((sparsegraph*)sg_in,lab,ptn,orbits,&options,&stats,sg_out);

  /* sparsenauty() can leave sg lists in various orders, which causes
     sgtos6() to give different strings for what are otherwise aresame_sg()
     graphs.  In the interests of canonical form sort here too.  (sgtog6()
     is not affected by list order in sg.)
  */
  /* sortlists_sg(sg_out); */

  return sg_out;
}

/* Return TRUE if sg1 and sg2 are isomorphic, or FALSE if not.
   This uses sparsenauty().

   Multi-edges are not treated properly.  Think the marking in aresame_sg()
   treats multi-edge as a single each.  What does sparsenauty() do?  Could
   sort and compact the edge array for compare.  sortlists_sg() sorts sg->e
   arcs for each vertex, but doesn't compact or order within sg->e.

   FIXME: notice directedness?  Or digraph always might be a slowdown?
   Or some degree checks first for inequality?
*/
int
nautyextra_is_isomorphic_sg (const sparsegraph *sg1, const sparsegraph *sg2)
{
  SG_DECL(canon_sg1);
  SG_DECL(canon_sg2);
  int same;

  if (sg1->nv != sg2->nv || sg1->nde != sg2->nde) return FALSE;

  DEBUG(printf("nautyextra_is_isomorphic_sg()\n"));
  SG_INIT(canon_sg1);
  SG_INIT(canon_sg2);
  nautyextra_canonise_sg(sg1,&canon_sg1);
  nautyextra_canonise_sg(sg2,&canon_sg2);
  same = aresame_sg(&canon_sg1,&canon_sg2);
  SG_FREE(canon_sg1);
  SG_FREE(canon_sg2);
  return same;
}

/* Return TRUE if g and sg are isomorphic, or FALSE if not.
   Args similar to nauty_to_sg().
   If sg has any multi-edges then it is not isomorphic to g, since g never has
   multi-edges.
*/
int
nautyextra_is_isomorphic_gsg (const graph *g, const sparsegraph *sg, int m, int n)
{
  if (sg->nv != n) return FALSE;
  if (n == 0) return TRUE;
  if (nautyextra_is_multiedged_sg(sg)
      || nautyextra_num_arcs(g,m,n) != sg->nde) return FALSE;

  /* Is it better to g->sg or sg->g for comparison?
     Knowing that sg->nv == n means only twice the size of g in temp space.
   */
  DYNALLSTAT(graph,g2,g2_sz);
  int g2_m = SETWORDSNEEDED(n);
  DYNALLOC2(graph,g2,g2_sz,m,n,"nautyextra_is_isomorphic_gsg");
  sg_to_nauty((sparsegraph*)sg, g2, g2_m, &g2_m);
  assert(g2_m == m);
  return nautyextra_is_isomorphic (g,g2,m,n);
}


/* Return TRUE if vpar1[1..n1] and vpar2[1..n2] are isomorphic.
   Return FALSE if not.
*/
int
nautyextra_is_isomorphic_vpar2 (const int *vpar1, int n1,
                                const int *vpar2, int n2,
                                int directed)
{
  int ret;
  if (n1 != n2) return FALSE;
  SG_DECL(sg1);
  SG_DECL(sg2);

  SG_INIT(sg1);
  SG_INIT(sg2);
  nautyextra_vpar_to_sg (vpar1, n1, &sg1, directed);
  nautyextra_vpar_to_sg (vpar2, n2, &sg2, directed);
  ret = nautyextra_is_isomorphic_sg (&sg1, &sg2);

  SG_FREE(sg1);
  SG_FREE(sg2);
  return ret;
}

/* Return TRUE if vpar1,n and vpar2,n are isomorphic.
   Return FALSE if not.

   directed=1 compares the trees as directed graphs.
   The intention is to have directed=0 for undirected, so free trees,
   but that doesn't work yet.

   The time and space is linear in the tree size n.  Non-isomorphic trees
   might be identified sooner than a full run.

   If perm!=NULL then perm[0..n] is set to a permutation of the integers
   1..n which maps vpar1 to vpar2.  v in vpar1 corresponds to perm[v] in
   vpar2.  nautyextra_relabel_perm_vpar() of vpar1,perm gives vpar2.
*/
int
nautyextra_is_isomorphic_vpar (const int *vpar1, const int *vpar2, int n,
                               int directed, int *perm)
{
  /* The code broadly follows

     Dinitz, Itai and Rodeh, "On an Algorithm of Zemlyachenko for Subtree
     Isomorphism", Technicon Technical Report CS0945, 1998.
     http://www.cs.technion.ac.il/users/wwwb/cgi-bin/tr-info.cgi/1998/CS/CS0945
     http://www.cs.technion.ac.il/users/wwwb/cgi-bin/tr-get.cgi/1998/CS/CS0945.pdf

     Vertices are considered in phases upwards according to subtree height.
     The first phase is the childless vertices.  They are reckoned removed.
     The second phase is vertices which are then childless, and so on.  The
     childless are all those with subtree height 0, then next phase all with
     subtree height 1, etc.

     .               *
     .              / \
     .             *   F          F = first set of parents
     .            / \   \
     .           F   F   L        L = initial leaves (childless)
     .          / \   \
     .         L   L   L

     data[v].num_children is initialized from vpar parents.
     data[v].num_children is a 2-element array, being
     data[v].num_children[0] for vpar1 and data[v].num_children[1] for
     vpar2.  A pass over num_children identifies the initial childless
     vertices L.  Vertices in a given phase have their respective parent
     num_children decremented.  Any which become 0 are new leaves for the
     next phase, such as F above.

     Vertices of the current phase have sequence numbers phase <= t < upto.
     data[t].seq[i] = vertex number at sequence number t, where i=0,1 for
     the two trees.  This is using data[] indexed by sequence number t,
     rather than by vertex number v as done above for num_children.

     The structure of the subtree at a vertex is represented by an integer
     id.  The set of ids used is common to the vpar1 and vpar2 trees.  When
     a vertex v1 from vpar1 and vertex v2 from vpar2 have the same id it
     means the subtrees at and below v1,v2 are isomorphic.

     Leaf vertices L are id=0.  Ids form an idtree.  Suppose a vpar vertex
     has child vertices with ids = id1,id2,...  These ids are descents down
     the idtree

     .                0
     .               /
     .             id1
     .            /  \
     .       id1,x   id1,id2
     .                  \
     .                 id1,id2,id3
     .                    \
     .                   id1,id2,id3,id4

     A parent starts as id 0 for no children.  Its first child has id1 and
     is a descent to an idtree vertex representing id1.  The second child
     id2 is further descent to idtree vertex representing sequence id1,id2.
     The third child to id1,id2,id3, etc.

     The sequence id1,id2,id3 is not recorded, nor are all possible
     descents.  idtree[id].sub_id is the most recent child under id.  In the
     diagram above would have idtree[id1].sub_id = id2, since id2 is the
     newest underneath id1.

     So a parent with its id = id1, on finding child id2 to be added, looks
     at idtree[id1].sub_id, sees that it's the desired id2.  The parent id
     is changed by setting data[p].id[i] = idtree[id1].cumul which is the
     desendant id representing id1,id2.

     If idtree[id1].sub_id != id2 desired, then a new_id in the idtree is
     created.  idtree_upto is the next available id number.  id1 is then set

     .    idtree[id1].sub_id = id2
     .    idtree[id1].cumul = new_id
     .
     .    idtree[new_id].sub_id = -1        for no sub-ids in idtree

     It suffices to record only sub_id most recent idtree descent because
     children are always taken in order of increasing id.  So if the current
     sub_id doesn't match then it's a smaller id and won't be seen again, so
     can be overwritten.

     Note that id values have no significance beyond representing
     substructure, and being assigned in blocks for a given subtree height.
     In particular they are not lexicographic order of substructure.

     The pass over the phase vertices establishes ids of the parents.
     The aim is to check that the set of ids in this phase is the same in
     vpar1 and vpar2.  Vertices with a given id are put into a linked list
     for a "bucket sort".  The linked list for a given id is

     .    idtree[id].v[i] = v          first of list, i=0 vpar1, i=1 vpar2
     .    data[v].chain[i] = v2        next of list
     .    data[v2].chain[i] = v3       next of list
     .    until some vx = 0            end of list

     An empty list is immediate end idtree[id].v[i] = 0.

     A pass over those ids, by increasing id, is then made and the linked
     lists followed, to set data[t].seq[i] to the vertex numbers of the next
     phase sorted by increasing id.  The count of how many of each id in
     vpar1 and vpar2 are compared at the same time.  If a count differs then
     the sets of ids are not the same and the trees are not isomorphic.

     Each parent has at least 1 child, so each will get an id which is newly
     created in the phase.  idtree_phase is the start of the parent ids of
     this phase.  The pass to bucket sort those parents thus only needs to
     consider idtree_phase .. idtree_upto, not the whole of the id range.

     The length of this idtree_phase .. idtree_upto is at most the number of
     children, so the total of all idtree pass work is linear.  This is
     where Dinitz et al make the algorithm linear.  If the parents where
     sequenced by a direct sort by their children's ids (lexicographically)
     then that would be a full O(N*log(N)), though at worst N=n/2 vertices
     can be parents in a given phase.  The restricted range of ids allows
     the bucket sort.

     The code here varies from Dinitz et al in two minor ways.

     - The vertices in phases are determined successively, not at the start.
     The phase number is the subtree height h of the vertices in that phase,
     but this is not counted or recorded.

     - Lists of parent vertices with given id are formed by a pass over
     idtree.  Dinitz et al instead maintain such lists (or sets rather) by
     moving a parent p from one list to another when its id changes.
     Double-linked lists would make that constant-time work to move, and
     there's only one move per child.  But the pass here should be about the
     same work, perhaps a touch less being single-linked.

     The total number of ids generated is at most n for any pair of trees,
     isomorphic or not.  A full set of n ids occurs for example with a
     path-n where every vertex going upwards is a new id.

     To show the id limit, suppose the trees are isomorphic up to a phase
     with C many vertices each about to be accumulated into their parents P.

     .             /  P  \               P = all above C
     .            +-------+
     .           /    C    \             C = child vertices of phase
     .          +-----------+
     .         /             \
     .        /       M       \          M = middle other vertices
     .       /                 \
     .      +-------------------+
     .     /          F          \       F = first set of parents
     .    +-----------------------+
     .   /            L            \     L = initial leaves
     .
     .                                   whole tree L+F+M+C+P = N

     All the L initial leaves are id=0.  Each F has a first L as a child.
     This first child of each gives the same id=1 to each F.  Further
     children on Fs make at most 1 new id per child.  So ids <= 2+L-F to
     establish L and F.  F to its parents (in M) and middles M to their
     parents (within M and up to C) add at most 1 id each vertex .  So reach
     all ids of C established with at most 2+L-F + F+M = 2+L+M ids.

     Now C accumulate to their parents adding at most 1 id each.  Those ids
     might be different in vpar1,vpar2, so up to 2*C.  Total ids I then

     .    I =  2 + L     + M + 2*C
     .      =  2 + L + C + M + C
     .      <= 2 + L + F + M + C        since C<=F

     The whole tree is N = L+F+M+C+P, and subtracting I gives N-I = P-2.

     If P=1 then that is 1 parent above all of C.  The C ids are the same in
     vpar1,vpar2 so accumulate to P as just 1 new id, not 2, so that I is 1
     smaller and have N-I = P-1 = 0 and so I <= N.  Otherwise P >= 2 and
     therefore have N-I >= 0 which is also I <= N.

     The processing of vertices stops when reach a set of children C which
     are seq indexes t with phase <= t < upto where upto==n, the size of the
     trees.  These children have been compared for equal sets of ids in
     vpar1,vpar2.  For a forest they are root vertices all with the maximum
     height in the forest.  For a tree they are the single root.

     If vpar1,vpar2 are forests of isolated vertices then they are all leaf
     and all go into the initial phase so that upto==n and so also the final
     phase with no work.

     For a forest, roots < maximum height are reached and compared between
     vpar1,vpar2 before the final phase.  Like other children, they are
     applied to the id of their parent, ie. to p=0.  At the final phase it
     is the same in vpar1,vpar2, though it is never compared.  Could try to
     skip its creation, but seems easier to let it store and not worry that
     it's unused.

     Accumulating ids into p=0 treats it like an additional vertex for the
     purpose of the idtree limit calculation above, but the id limit still
     holds.  If p=0 is ever used then it is one of F or later.  If any
     singleton vertices then p=0 is in F, or if all forest components of >=2
     vertices then p=0 is set later.  p=0 does not have a parent so the
     maximum number of P parents above reduces by at least 1, so that C<=F
     in the I calculation above becomes C < F, thus reducing I by 1 and
     maintaining I <= N.


     For the permutation option, a mapping of vertices between vpar1 and
     vpar2 is found working down the tree by descending id.  For example,

     .        vpar1                 vpar2
     .      1      2               1      2
     .     / \     |              / \     |
     .    3   4    5             3   5    4

     Vertices 1,2 have different ids and map directly 1->1, 2->2.  The
     children 3,4,5 all have the same id (being leaves), but they must be
     mapped so that their parents match.  So 3,4 in vpar1 parent 1 -> 1 in
     vpar2 and its children are 3,5 so 3->3, 4->5.  Similarly 5 in vpar1 map
     -> 4 in vpar2 because their parents 2->2.

     For a given id, the vertices of vpar1 are put into linked lists under
     "p2" which is where their parent maps to.  The vertices of vpar2 with
     the id are then mapped by taking a vertex v1 from the lists.

     Vertices which are roots go into a linked list under p2=0 and are
     mapped from there the same as others.

     Whenever a list for p2 is >=2 vertices, meaning >=2 identical
     isomorphic subtrees, they can be matched up to the corresponding in
     vpar2 in various ways (list length factorial).  No attempt is made to
     arrange any particular order, so the mapping is just one of possibly
     several suitable mappings.

     The full set of mappings is the automorphism group applied to any one
     mapping, but there's nothing here for that.

     ----------------
     cf.

     V. N. Zemlyachenko, "Determining Tree Isomorphism", in Voprosy
     Kibernetiki, Proceedings of the Seminar on Combinatorial Mathematics
     (Moscow 1971), pages 54-60.

     -----
     Aho, Hoperoft and Ullman, "The Design and Analysis of Computer
     Algorithms", Addison-Wesley, 1974, pages 84-85.

     -----
     Douglas M. Campbell and David Radford, "Tree Isomorphism Algorithms:
     Speed vs. Clarity", Mathematics Magazine, volume 64, number 4, October
     1991, pages 252-261.  http://www.jstor.org/stable/2690833

     Dialogue on isomorphism problem and the Aho, Hopcoft, Ullman algorithm
     being a bit obscure, in its most compact statement at least.

     -----
     Baida Zhanga, Yuhua Tanga, Junjie Wua, Shuai Xub, "LD: A Polynomial Time
     Algorithm for Tree Isomorphism", Procedia Engineering, volume 15, 2011,
     pages 2015-2020.

     -----
     Valiente, "Simple and Efficient Subtree Isomorphism", Technical Report
     LSI-00-72-R, Technical University of Catalonia, 2000.

     Valiente, "Simple and Efficient Tree Comparison", Technical Report
     LSI-01-1-R, Technical University of Catalonia, 2001.
     http://www.cs.upc.edu/~valiente/

     http://www.cs.upc.edu/~valiente/graph-00-01-c.pdf
     Slides on tree isomorphism and mapping problems.

     GP-Test  my(i = 2+'L+'F+'M+'C, N = 'L+'F+'M+'C+'P); N-i     == 'P-2
     GP-Test  my(i = 2+'L+'F+'M+'C, N = 'L+'F+'M+'C+'P); N-(i-1) == 'P-1
  */

  DEBUG(printf("nautyextra_is_isomorphic_vpar() %p %p n=%d  perm %p\n",
               vpar1,vpar2,n,perm));
  DEBUG(printf(" vpar1  "); nautyextra_print_vpar(vpar1,n));
  DEBUG(printf(" vpar2  "); nautyextra_print_vpar(vpar2,n));
  assert (n >= 0);
  assert (directed == 1);

  struct data {
    int num_children[2];
    int seq[2];
    int id[2];
    int chain[2];
  };
  struct idtree {
    int sub_id;
    int cumul;
    int v[2];
  };
  const int idtree_size_extra = 0;
  int data_size   = n+1;
  int idtree_size = n + idtree_size_extra;
#if MAXN
  assert(n <= MAXN);
  struct data   data[MAXN+1];
  struct idtree idtree[MAXN+idtree_size_extra];
  assert (data_size   <= numberof(data));
  assert (idtree_size <= numberof(idtree));
#else
  WORKSPACE2(struct data,data,data_size,
             struct idtree,idtree,idtree_size,
             "nautyextra_is_isomorphic_vpar");
#endif
  int  i, t, v;

  /* zeros for data[v] id[] and chain[]
     zeros for idtree[id].v[]
  */
  memset (data, '\0', data_size  *sizeof(data[0]));
  assert((memset(idtree,'\x55',idtree_size*sizeof(idtree[0])), 1));
  idtree_size = idtree_size;  /*USED*/

  const int *vpars[2];
  vpars[0] = vpar1;
  vpars[1] = vpar2;

  /* establish data[v],num_children */
  for (v = 1; v <= n; v++)
    for (i = 0; i <= 1; i++)
      data[vpars[i][v]].num_children[i]++;

  DEBUG({ printf("initial num children\n");
          for (i = 0; i <= 1; i++) {
            printf("  vpar%d  ", i+1);
            for (v = 1; v <= n; v++)
              printf(" %d", data[v].num_children[i]);
            printf("\n");
          }
        });

  /* establish data[t],seq initial lists of children 0 to upto-1 */
  int new_upto[2];
  new_upto[0] = new_upto[1] = 0;
  for (v = 1; v <= n; v++) {
    for (i = 0; i <= 1; i++) {
      if (data[v].num_children[i] == 0) {
        data[v].num_children[i] = -1;
        data[new_upto[i]++].seq[i] = v;
      }
    }
  }

  DEBUG(printf("initial leaf counts %d %d\n", new_upto[0], new_upto[1]));
  int upto = new_upto[0];
  if (upto != new_upto[1]) {
    DEBUG(printf("  different, not isomorphic\n"));
    return FALSE;
  }

  /* id=0 is root of idtree and id of initial leaves */
  idtree[0].sub_id = -1;
  int idtree_upto = 1;

  int phase = 0;
  for (;;) {
    DEBUG(printf("seq phase=%d to upto=%d\n", phase, upto));
    DEBUG({ for (i = 0; i <= 1; i++) {
              printf("  seq vpar%d vertices ", i+1);
              for (t = 0; t < phase; t++)
                printf(" %d", data[t].seq[i]);
              printf(" phase=%d:", phase);
              for (t = phase; t < upto; t++)
                printf(" %d", data[t].seq[i]);
              printf("\n");
            }
          });
    if (upto >= n) {
      DEBUG(printf("end, is isomorphic\n"));

      /* End at final phase..upto-1 which is the rest of vertices 1..n,
         having been compared and found equal.
         Roots of < maximum height put same settings into p=0 id. */
      assert (upto == n);
      assert (data[0].id[0] == data[0].id[1]);
      assert (upto - phase == nautyextra_height_num_components_vpar(vpar1,n));
      assert (upto - phase == nautyextra_height_num_components_vpar(vpar2,n));

      break;
    }

    /* idtree_phase = next new id which will be added.
       After new ids are added they are idtree_phase <= id < idtree_upto.
    */
    int idtree_phase = idtree_upto;

    int new_upto[2];
    new_upto[0] = new_upto[1] = upto;

    for (t = phase; t < upto; t++) {
      for (i = 0; i <= 1; i++) {
        DEBUG(printf("  t=%d i=%d\n", t, i));

        int v = data[t].seq[i];
        int p = vpars[i][v];
        assert (1 <= v && v <= n);
        assert (0 <= p && p <= n);

        int v_id = data[v].id[i];
        int p_id = data[p].id[i];
        DEBUG(printf("  vpar%d  v=%d v_id=%d   p=%d p_id=%d\n",
                     i+1, v,v_id, p,p_id));

        assert (0 <= p_id && p_id < idtree_upto);
        assert (idtree[p_id].sub_id <= v_id);

        if (idtree[p_id].sub_id == v_id) {
          DEBUG(printf("    descend idtree to %d sub_id %d\n",
                       idtree[p_id].cumul, v_id));
          p_id = idtree[p_id].cumul;
        } else {
          DEBUG(printf("    new idtree %d sub_id %d\n", idtree_upto, v_id));
          assert(0 <= idtree_upto && idtree_upto < idtree_size);
          idtree[p_id].sub_id = v_id;
          idtree[p_id].cumul = idtree_upto;
          idtree[idtree_upto].sub_id = -1;  /* nothing below */
          idtree[idtree_upto].v[0] = idtree[idtree_upto].v[1] = 0;  /*endlist*/
          p_id = idtree_upto;
          idtree_upto++;
        }
        DEBUG(printf("    set parent %d id %d\n", p, p_id));
        data[p].id[i] = p_id;

        if (--data[p].num_children[i] == 0) {
          DEBUG(printf("    p=%d becomes leaf\n", p));
          data[new_upto[i]++].seq[i] = p;
        }
      }
    }
    DEBUG(printf("  seq new_upto %d %d\n", new_upto[0], new_upto[1]));
    phase = upto;
    upto = new_upto[0];
    if (upto != new_upto[1]) {
      DEBUG(printf("  different, not isomorphic\n"));
      return FALSE;
    }

    /* Loop over parent vertices seq which are the new phase.
       Build idtree[id].v[i] linked list of vertices with that id.
    */
    DEBUG(printf("  parents seq phase %d upto %d\n", phase, upto));
    for (t = phase; t < upto; t++) {
      for (i = 0; i <= 1; i++) {
        int v = data[t].seq[i];
        int id = data[v].id[i];
        DEBUG(printf("   vpar%d  v=%d id %d\n", i+1, v, id));
        assert (v != 0);
        assert (idtree_phase <= id && id < idtree_upto);
        assert (id != 0);  /* nothing put into root id=0 */
        data[v].chain[i] = idtree[id].v[i]; idtree[id].v[i] = v;
      }
    }

    /* by ids, lists of vertices, to make seq in ascending id order,
       and compare counts in vpar1,2
    */
    DEBUG(printf("  idtree_phase %d idtree_upto %d\n", idtree_phase, idtree_upto));
    new_upto[0] = new_upto[1] = phase;
    int id;
    for (id = idtree_phase; id < idtree_upto; id++) {
      DEBUG(printf("  id %d\n", id));
      int count[2];
      count[0] = 0;
      count[1] = 0;
      for (i = 0; i <= 1; i++) {
        for (v = idtree[id].v[i]; v!=0; v = data[v].chain[i]) {
          data[new_upto[i]++].seq[i] = v;     /* by increasing id */
          count[i]++;
        }
      }
      DEBUG(printf("    counts %d %d\n", count[0], count[1]));
      if (count[0] != count[1]) {
        DEBUG(printf("    different, not isomorphic\n"));
        return FALSE;
      }
    }
    DEBUG(printf("  traversed chains new_upto %d %d\n",
                 new_upto[0], new_upto[1]));
    assert (new_upto[0] == upto);
    assert (new_upto[1] == upto);
  }

  if (perm) {
    perm[0] = 0;

    for (i = 0; i <= n; i++)
      data[i].num_children[0] = 0;

    /* data[p2].num_children[0] is the "head" of the linked lists,
       ie. the first v in vpar1.  Then data[v].chain[0] next, until 0.
    */
    upto--;
    while (upto >= 0) {
      phase = upto;
      int v         = data[phase].seq[0];
      int target_id = data[v].id[0];
      DEBUG(printf("perm phase=%d is v=%d target_id=%d\n",
                   phase, v, target_id));
      assert (1 <= v && v <= n);
      assert (0 <= target_id && target_id < idtree_upto);
      while (upto >= 0) {
        int v = data[upto].seq[0];                     /* v in vpar1 */
        DEBUG(printf(" upto=%d vpar1 v=%d id=%d\n",
                     upto, v, data[v].id[0]));
        assert (1 <= v && v <= n);
        if (data[v].id[0] != target_id) break;

        int p = vpar1[v];
        assert (0 <= p && p <= n);
        int p2 = (p ? perm[p] : 0);
        if (p) assert (1 <= perm[p] && p <= n);

        DEBUG(printf("  link at p2=%d (head was %d)\n",
                     p2, data[p2].num_children[0]));
        data[v].chain[0] = data[p2].num_children[0];  /* chain = head */
        data[p2].num_children[0] = v;                  /* head = v */
        upto--;
      }

      DEBUG({ int p2;
              for (p2=0; p2<=n; p2++) {
                int v = data[p2].num_children[0];
                if (v) {
                  printf(" list p2=%d v=%d chain", p2, v);
                  while ((v=data[v].chain[0])) {
                    printf(" %d", v);
                  }
                  printf("\n");
                }
              }});

      DEBUG(printf(" to vpar2 upto=%d to phase=%d\n", upto, phase));
      for (i = upto+1; i <= phase; i++) {
        int v2 = data[i].seq[1];             /* from vpar2 */
        int p2 = vpar2[v2];
        DEBUG(printf(" i=%d vpar2 v2=%d p2=%d\n", i, v2, p2));

        int v = data[p2].num_children[0];    /* v int vpar1 */
        data[p2].num_children[0] = data[v].chain[0];
        perm[v] = v2;
        DEBUG(printf("  set perm %d -> %d, list head now %d\n",
                     v, v2, data[p2].num_children[0]));
        assert (1 <= v && v <= n);
        assert (1 <= v2 && v2 <= n);
      }
#ifndef NDEBUG
      /* all vertices taken from vpar1 to put in perm */
      for (i = 0; i <= n; i++) assert(data[i].num_children[0] == 0);
#endif
    }

    DEBUG({ printf("perm ");
            printf(" [%d] ", perm[0]);
            for (v = 1; v <= n; v++)
              printf(" %d", perm[v]);
            printf("\n");
          });

  } else {
    DEBUG(printf("  no perm calculation\n"));
  }
  return TRUE;
}

/* Return TRUE if sg and vpar[1..n] are isomorphic.
   Return FALSE if not.

   FIXME: sg,vpar treated as undirected ok ?
   FIXME: sg to vpar tree checking not cyclic then isomorphism trees.
*/
int
nautyextra_is_isomorphic_sg_vpar (const sparsegraph *sg, const int *vpar, int n)
{
  DEBUG(printf("nautyextra_is_isomorphic_sg_vpar() %p %p %d\n", sg, vpar, n));

  if (sg->nv != n
      || sg->nde != (n ? 2*(n-1) : 0)) return FALSE;
  if (n == 0) return TRUE;

  DYNALLSTAT(graph,g,g_sz);
  int m = SETWORDSNEEDED(n);
  DYNALLOC2(graph,g,g_sz,m,n,"nautyextra_is_isomorphic_sg_vpar");
  nautyextra_vpar_to_graph(vpar, g,m,n, 0);

  return nautyextra_is_isomorphic_gsg (g,sg, m,n);
}

int
nautyextra_aresame_vpar (const int *vpar1, const int *vpar2, int n)
{
  DEBUG(printf("nautyextra_aresame_vpar() %p %p n=%d\n", vpar1,vpar2, n));
  assert (n>=0);
  return memcmp(vpar1+1, vpar2+1, n*sizeof(*vpar1)) == 0;
}

/* ------------------------------------------------------------------------- */
/* Self-Complement */

/* g is a graph, either directed or undirected.
   Return TRUE if g is self-complement, meaning isomorphic to its complement().
   Return FALSE if not.

   FIXME: Waiting for isomorphic.
*/
int
nautyextra_is_selfcomplement (const graph *g, int m, int n)
{
  assert (n>=0);

  /* must have n==0,1 mod 4 for even total number of edges, so n==2,3 not

     GP-Test  vector(20,n,n--; binomial(n,2)%2==0) == \
     GP-Test  vector(20,n,n--; n%4==0 || n%4==1)

     GP-Test  vector(20,n,n--; binomial(n,2)%2==1) == \
     GP-Test  vector(20,n,n--; n%4==2 || n%4==3)
  */
  if (n&2) return FALSE;

  DYNALLSTAT(graph,g2,g2_sz);
  DYNALLOC2(graph,g2,g2_sz,m,n,"nautyextra_is_selfcomplement");
  memcpy(g2, g, n*m*sizeof(g[0]));
  complement(g2,m,n);

  unsigned long g_edges;
  int g_mindeg;
  int g_mincount;
  int g_maxdeg;
  int g_maxcount;
  boolean g_eulerian;
  degstats((graph*)g,m,n, &g_edges, &g_mindeg, &g_mincount, &g_maxdeg, &g_maxcount,
           &g_eulerian);

  unsigned long g2_edges;
  int g2_mindeg;
  int g2_mincount;
  int g2_maxdeg;
  int g2_maxcount;
  boolean g2_eulerian;
  degstats(g2,m,n, &g2_edges, &g2_mindeg, &g2_mincount, &g2_maxdeg, &g2_maxcount, &g2_eulerian);
  assert(g_edges + g2_edges == n*(n-1)/2);

  if (g_edges != g2_edges
      || g_mindeg != g2_mindeg
      || g_mincount != g2_mincount
      || g_maxdeg != g2_maxdeg
      || g_maxcount != g2_maxcount
      || g_eulerian != g2_eulerian) return FALSE;

  return nautyextra_is_isomorphic (g,g2,m,n);
}


/* ------------------------------------------------------------------------- */
/* Sequences */

void
nautyextra_seq_upascend (const int *vpar, int n, int ahead, int *seq)
{
  DEBUG(printf("nautyextra_seq_upascend() n=%d ahead=%d \n", n, ahead);
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);
  assert (ahead==0 || ahead==1);
  int i;
#if MAXN
  int num_children[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(int,num_children,n+1,"nautyextra_seq_upascend");
#endif

  nautyextra_num_children_array_vpar(vpar,n, num_children);

  int upto = 0;
  for (i = 1; i <= n; i++) {
    DEBUG(printf(" i=%d\n", i));
    int v = i;
    if (num_children[v] != 0) continue;

    for (;;) {
      assert (num_children[v] == 0);

      DEBUG(printf("  store seq[%d] = %d\n", upto,v));
      assert(upto<n);
      seq[upto++] = v;

      num_children[v] = -1;
      if (! (v = vpar[v])) break; /* to parent */
      assert (v != i);

      assert(num_children[v] > 0);
      if (--num_children[v]) break;

      if (!ahead && v>i) break;
    }
  }
  DEBUG(printf(" end upto=%d\n", upto));
  assert(upto==n);
}
void
nautyextra_from_upascend (const int *vec, int n, int ahead, int *vpar)
{
  DEBUG(printf("nautyextra_from_upascend() n=%d ahead=%d\n", n, ahead));
  DEBUG({ int i;
          printf("vec ");
          for (i = 0; i < n; i++) printf(" %d", vec[i]);
          printf("\n");
        });
  assert (n >= 0);
  assert (ahead==0 || ahead==1);
  assert (n==0 || vec[n-1]==0);
  int i;
#if MAXN
  int num_children[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(int,num_children,n+1,"nautyextra_from_upascend");
#endif

  nautyextra_num_children_array_vpar(vec-1,n,num_children);
  DEBUG({ int v;
          printf("num children ");
          for (v = 1; v <= n; v++) printf(" %d", num_children[v]);
          printf("\n");
        });

  int upto = 0;
  for (i = 1; i <= n; i++) {
    DEBUG(printf(" i=%d upto=%d  num_children=%d\n",
                 i, upto, num_children[i]));
    int v = i;
    while (num_children[v] == 0) {
      num_children[v] = -1;
      int p = vec[upto++];
      DEBUG(printf("  set v=%d parent %d\n", v, p));
      vpar[v] = p;
      if (!p) break;
      num_children[p]--;
      v = p;
      if (!ahead && v>i) {
        DEBUG(printf("  break for p>i\n"));
        break;
      }
    }
  }
  assert(upto == n);
}

/* Sort key[1..n] by ascending order.
   Set seq[1..n] to i,j,k,... so that key[i] < key[j] < ...
*/
void
nautyextra_seq_bucket_sort (const int *keys, int n, int *seq)
{
  DEBUG(printf("nautyextra_seq_bucket_sort() n=%d keys=%p to seq=%p\n",
               n, keys, seq));
  assert (n >= 0);
#if MAXN
  int chain[MAXN+1];
  assert(n <= MAXN);
#else
  WORKSPACE(int,chain,n+1, "nautyextra_seq_bucket_sort");
#endif
  assert (chain != keys);

  memset (seq, '\0', (n+1)*sizeof(seq[0]));
  memset (chain, '\0', (n+1)*sizeof(chain[0]));  /* all end-of-lists */

  if (n) {
    int min_key = keys[1];
    int max_key = min_key;
    int i;
    for (i = 2; i <= n; i++) {
      min_key = MIN(min_key, keys[i]);
      max_key = MAX(max_key, keys[i]);
    }
    int max_k = max_key - min_key;
    DEBUG(printf("key range %d to %d, adjusted 0 to %d\n",
                 min_key, max_key, max_k));
    assert(max_k <= n);

    for (i = 1; i <= n; i++) {
      int k = keys[i] - min_key;
      assert (0 <= k && k <= n);
      chain[i] = seq[k]; seq[k] = i;
    }

    assert(seq[0] != 0);      /* the smallest key */
    assert(seq[max_k] != 0);  /* the biggest key */

    int upto = n;
    int k;
    for (k = max_k; k >= 0; k--) {
      DEBUG(printf(" key k=%d\n", k));
      int i = seq[k];   /* first of linked list */
      while (i) {
        DEBUG(printf("  i=%d  store to %d\n", i, upto));
        assert (upto >= 1);
        assert (upto >= k);
        seq[upto--] = i;
        i = chain[i];
      }
    }
    assert (upto == 0);
  }
  seq[0] = 0;
}

void
nautyextra_seq_subtree_height_vpar (const int *vpar, int n, int *seq)
{
  DEBUG(printf("nautyextra_seq_subtree_height_vpar()\n"));
  assert (n >= 0);
#if MAXN
  int heights[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  /* with initial n+1 room for nautyextra_subtree_heights_vpar() num children */
  WORKSPACE(int,heights,2*(n+1), "nautyextra_seq_subtree_height");
  heights += n+1;
#endif

  nautyextra_subtree_heights_vpar(vpar,n, heights);
  nautyextra_seq_bucket_sort(heights,n, seq);
}

void
nautyextra_max_ancestor_array_vpar (const int *vpar, int n, int *max_ancestor)
{
  assert (n >= 0);
  assert (max_ancestor != NULL);

  memset (max_ancestor, '\0', (n+1)*sizeof(max_ancestor[0]));
  int i;
  for (i = 1; i <= n; i++) {
    if (max_ancestor[i]) continue;
    int v = i;
    int prev = 0;
    int m = 0;
    for (;;) {
      max_ancestor[v] = -prev; prev=v;
      v = vpar[v];
      if (!v) break;   /* at root */

      m = max_ancestor[v];
      if (m) {
        if (m<0) {
          int s = prev;
          for (;;) {
            m = MAX(m,s);
            if (s == v) break;
            s = -max_ancestor[s];
          }
        }
        break;
      }

      while (prev > 0) {
        v = prev; prev = -max_ancestor[v];    /* downwards */
        max_ancestor[v] = m = MAX(m,v);
      }
    }
  }
}


/* ------------------------------------------------------------------------- */

/* Return TRUE if sg has an arc u-->v.
   Return FALSE if not.

   ENHANCE-ME: A version which assumed sg has sorted arc lists could bsearch.
*/
int
nautyextra_is_arc_sg (const sparsegraph *sg, int u, int v)
{
  const int *dd = sg->d;
  int degree    = dd[u];
  const int *ee = sg->e + sg->v[u];
  while (degree--) {
    int to = *ee++;
    if (to == v) return TRUE;
  }
  return FALSE;
}

/* Return TRUE if tree vpar[1..n] is connected.
   Return FALSE if not.
   Connected here means <= 1 roots, so an empty tree is reckoned connected.
   A forest of 2 or more component trees is not connected.

   cf. isconnected() for graph.
*/
int
nautyextra_is_connected_vpar (const int *vpar, int n)
{
  assert (n >= 0);
  int num_roots = -2;
  int v;
  for (v = 1; v <= n; v++)
    if (vpar[v]==0 && ++num_roots==0)
      return FALSE;
  return TRUE;
}

/* Return TRUE if g has any cycle u--v--...--u, or return FALSE if not.
   g must be undirected, so both u-->v and v-->u arcs.
   g must be connected.

   The strategy is to start from vertex 0 and expect to follow all its
   neighbours along and spreading outwards.  If g is a tree (acyclic) then
   no previous vertex will be seen, except the immediately preceding parent
   of each.

   Similar code to nauty girth(), but here stopping at any cycle, not
   continuing on to look for the smallest.
*/
int
nautyextra_is_cyclic (const graph *g, int m, int n)
{
  assert (n >= 0);

  /* Cannot EMPTYSET() on seen==NULL which occurs when n==0. */
  if (n == 0) return FALSE;

#if MAXN
  int queue[2*MAXN];
  set seen[SETWORDSNEEDED(MAXN)];
  assert(n <= MAXN);
#else
  WORKSPACE2(int,queue,2*n,
             set,seen, m,   "nautyextra_is_cyclic");
#endif
  DEBUG(printf("nautyextra_is_cyclic() n=%d\n", n));
  EMPTYSET(seen,m);

  int *qhead = queue;
  int *qtail = queue;
  *qtail++ = 0;
  *qtail++ = 0;
  ADDELEMENT(seen,0);
  while (qhead != qtail) {
    int parent = *qhead++;
    int from   = *qhead++;
    DEBUG(printf("  at from=%d parent %d  queue remaining %d, seen %s\n",
                 from, parent, qtail-qhead, nautyextra_set_str(seen,n)));
    assert (0 <= from && from < n);
    assert (0 <= parent && parent < n);
    assert (qhead <= queue + 2*n);

    set *from_row = ((set*) g) + from;
    int to;
    for (to = -1; (to = nextelement(from_row,m,to)) >= 0; ) {
      if (to == parent) continue;
      if (ISELEMENT(seen,to)) {
        DEBUG(printf("  cycle at %d to %d\n", from, to));
        return TRUE;
      }

      ADDELEMENT(seen,to);
      *qtail++ = from;
      *qtail++ = to;
      assert (qtail <= queue + 2*n);
    }
  }

  DEBUG(printf("  end of queue, no cycle found\n"));
  return FALSE;
}


int
nautyextra_is_cyclic_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_is_cyclic() n=%d\n", n));
  assert (n >= 0);

  /* Cannot EMPTYSET() on seen==NULL which occurs when n==0. */
  if (n == 0) return FALSE;

#if MAXN
  int seen[MAXN+1];
  assert(n <= MAXN);
#else
  WORKSPACE(int,seen,n+1,"nautyextra_is_cyclic_vpar");
#endif
  memset(seen, '\0', (n+1)*sizeof(*seen));

  int i;
  for (i=1; i<=n; i++) {
    int v = i;
    do {
      int s = seen[v];
      if (s) {
        if (s == i) return TRUE;  /* revisit to this search, is a cycle */
        break;  /* reach something already found acyclic */
      }
      seen[v] = i;
      v = vpar[v];
      assert (0 <= v && v <= n);
    } while (v);
  }
  return FALSE;
}

/* Return TRUE if vpar[1..n] has any duplicated leaf, meaning two leaf
   vertices with the same attachment vertex.
*/
int
nautyextra_any_duplicated_leaf_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_any_duplicated_leaf_vpar() n=%d\n", n));
  assert (n >= 0);

  int v;
  struct data {
    int num_children;
    int num_leaf_attach;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_any_duplicated_leaf_vpar");
#endif

  memset(data, '\0', (n+1)*sizeof(*data));
  for (v = 1; v <= n; v++)
    data[vpar[v]].num_children++;

  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    if (p) {
      if (data[v].num_children == 0) {
        DEBUG(printf("v=%d leaf p=%d\n", v, p));
        if (data[p].num_leaf_attach++ >= 1) {
          DEBUG(printf("  duplicated leaf at p=%d\n", p));
          return TRUE;
        }
      }

      if (vpar[p]==0 && data[p].num_children == 1) {
        DEBUG(printf("v=%d parent p=%d root no other children, so parent is leaf\n",
                     v, p));
        if (data[v].num_leaf_attach++ >= 1) {
          DEBUG(printf("  duplicated leaf at v=%d\n", p));
          return TRUE;
        }
      }
    }
  }
  return FALSE;
}

/* Return the longest limb height in tree vpar[1..n].
   This is the longest path length from a leaf up to a branch vertex or the
   root, whichever comes first.
*/
int
nautyextra_limb_height_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_limb_height_vpar() n=%d\n", n));
  assert (n >= 0);

  int i;
  struct data {
    int num_children;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_limb_height_vpar");
#endif

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  int max_limb = 0;
  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children != 0) continue;

    DEBUG(printf(" v=%d leaf\n", v));
    int limb = 0;
    while ((v = vpar[v])) {
      limb++;
      if (data[v].num_children != 1) break;
    }
    DEBUG(printf("   limb %d\n", limb));
    max_limb = MAX(max_limb, limb);
  }
  return max_limb;
}

/* Return the longest limb length in tree vpar[1..n].
   This is the longest path length from a leaf to a branch vertex.

   The path is for vpar as a free tree, so it can go up to the root and down
   again to a branch.
*/
int
nautyextra_limb_length_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_limb_length_vpar() ");
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
  struct data {
    int num_children;
    int root_limb_height;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_limb_length_vpar");
#endif

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  int max_limb = 0;
  for (i = 1; i <= n; i++) {
    int v = i;
    int num_children = data[v].num_children;
    if (num_children == 1) continue;
    int start_leaf = (num_children==0);

    DEBUG(printf("v=%d num_children %d go up start_leaf=%d\n",
                 v, num_children, start_leaf));
    int limb = 0;
    for (;;) {
      int p = vpar[v];
      if (!p) {
      ROOT:
        DEBUG(printf("  stop at root v=%d num_children %d\n", v, num_children));
        if (num_children <= 2) {
          int root_limb_height = data[v].root_limb_height;
          DEBUG(printf("  <=2 children, root_limb_height %d %d\n",
                       root_limb_height >> 1, root_limb_height & 1));

          data[v].root_limb_height
            = root_limb_height | ((limb << 1) + start_leaf);
          DEBUG(printf("  set root_limb_height to %d %d\n",
                       data[v].root_limb_height >> 1,
                       data[v].root_limb_height & 1));

          int to_leaf = (num_children==1);
          limb += root_limb_height >> 1;
          int root_limb_height_from_leaf = root_limb_height&1;
          if (start_leaf || root_limb_height_from_leaf || to_leaf) {
            goto MAX_LIMB;
          } else {
            goto NEXT_V;
          }
        }
        break;
      }
      v = p;
      num_children = data[v].num_children;
      limb++;
      DEBUG(printf("  v=%d num_children %d\n", v, num_children));
      if (num_children != 1) {
        DEBUG(printf("  stop at branch v=%d num_children %d\n",
                     v, num_children));
        if (!vpar[v]) goto ROOT;
        break;
      }
    }

    if (start_leaf) {
    MAX_LIMB:
      DEBUG(printf("  limb %d\n", limb));
      max_limb = MAX(max_limb, limb);
    }

  NEXT_V:
    ;
  }
  return max_limb;
}


/* ------------------------------------------------------------------------- */

/* Search g for triangles.
   g must be undirected (edges both ways) and no self-loops.
   Each triangle is

         a
        / \     with a < b < c
       b---c

   and calls with the 3 vertices and the given user "data" pointer,

       stop = (*callback) (g,m,n, a,b,c, data);

   The vertices are ordered a < b < c.

   If (*callback) returns 0 then the search continues.
   If (*callback) returns non-0 then the search stops and
   nautyextra_triangle_search() returns with that non-zero value.
   If the search runs to completion then nautyextra_triangle_search()
   returns 0.

   The current code is a slightly naive search over all vertices a, looking
   for suitable neighbours b,c.  This means 2 neighbours with an edge
   between.  a can have more then 2 neighbours.  All suitable pairs of them
   are triangles.
*/
int
nautyextra_triangle_search (const graph *g, int m, int n,
                            int (*callback)(const graph*,int,int,
                                            int,int,int,void *),
                            void *data)
{
  /* Cf nauty numtriangles().  Inner loop here is similar. */

  DEBUG(printf("nautyextra_triangle_search() %p n=%d\n", g, n));
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));

  set *a_row = (set*) g;
  int a;
  for (a = 0; a < n-2; a++, a_row += m) {
    DEBUG(printf(" a=%d\n", a));
    int b;
    for (b = a; (b = nextelement(a_row,m,b)) >= 0; ) {
      set *b_row = GRAPHROW(g,b,m);

      /* look at a_row BITAND b_row for edges to some c in both */
      int i = SETWD(b);
      const set *ai_row = a_row + i;
      const set *bi_row = b_row + i;
      setword w = *ai_row++ & *bi_row++;
      w &= BITMASK(SETBT(b));  /* keep bits > b */
      for (;;) {
        while (w) {
          int t;
          TAKEBIT(t,w);
          int c = TIMESWORDSIZE(i) + t;
          int ret = (*callback)(g,m,n, a,b,c, data);
          if (ret) return ret;
        }
        if (++i >= m) break;
        w = *ai_row++ & *bi_row++;
      }
    }
  }
  return 0;
}


/* ------------------------------------------------------------------------- */

/* Search g,m,n for diamond subgraphs.
   g must be undirected (edges both ways) and no self-loops.
   Each diamond sought is

         c
        / \     with a < b and c < d
       a---b
        \ /
         d

   and calls with the 4 vertices and the given user "data" pointer,

       stop = (*callback) (g,m,n, a,b,c,d, data);

   If (*callback) returns 0 then the search continues.
   If (*callback) returns non-0 then the search stops and
   nautyextra_diamond_search() returns with that non-zero value.
   If the search runs to completion then nautyextra_diamond_search()
   returns 0.

   The vertices are ordered a < b and c < d.  The order between those pairs
   depends on the graph and can be anything.

   induced=0 allows with or without edge c-d.  An edge c-d makes a
   complete-4 subgraph.  In this case the search will give all combinations
   of ways to take a pair a,b and pair c,d, which means 4 choose 2 = 6 ways.
   For bigger or other complete subgraphs likewise all combinations of
   pairs.

   An induced=0 search can be thought of as looking for two triangles with
   an edge in common -- that edge being a,b.  Every edge which is common to
   two triangles is reported, but just once each such edge.  If an edge has
   several triangles around it then c,d will take those triangles in all
   combinations 2 at a time, but with c<d so that the 2 triangles are c,d
   just one way around, not also d,c.

   induced=1 requires no edge c-d, so that the diamond is an induced
   subgraph of g.

   The current code is a slightly naive search over all vertices a, looking
   for suitable neighbours b,c,d.  This means edge b-c and b-d, and if
   induced then no c-d.  The search is b>a, but then c starts as any
   neighbour at all, then d>c.
*/
int
nautyextra_diamond_search (const graph *g, int m, int n, int induced,
                                   int (*callback)(const graph*,int,int,
                                                   int,int,int,int, void *),
                                   void *data)
{
  DEBUG(printf("nautyextra_diamond_search() %p n=%d\n", g, n));
  assert (n >= 0);

  /*
    .   c
    .  / \     with a < b
    . a---b    and  c < d
    .  \ /
    .   d
  */
  int a;
  set *a_row = (set*) g;
  for (a = 0; a < n-1; a++, a_row += m) {                  /* a all */
    DEBUG(printf(" a=%d\n", a));

    int b;
    for (b = a; (b = nextelement(a_row,m,b)) >= 0; ) {     /* a < b */
      DEBUG(printf("  b=%d\n", b));
      assert (a < b);
      set *b_row = GRAPHROW(g,b,m);

      int c;
      for (c = -1; (c = nextelement(a_row,m,c)) >= 0; ) {  /* c any, c!=b */
        assert (c != a); /* no self-loop */
        if (c == b) continue;
        assert (c != a); /* no self-loop */
        DEBUG(printf("   c=%d\n", c));
        if (! ISELEMENT(b_row,c)) continue;  /* no b--c edge, not diamond */
        set *c_row = GRAPHROW(g,c,m);

        int d;
        for (d = c; (d = nextelement(a_row,m,d)) >= 0; ) { /* c<d, d!=b */
          assert (d != a); /* no self-loop */
          if (d == b) continue;
          DEBUG(printf("    d=%d\n", d));
          assert (c < d);

          if (! ISELEMENT(b_row,d)) continue;  /* no b--d edge, not diamond */
          if (induced
              && ISELEMENT(c_row,d)) continue; /* have c--d edge, not induced */

          DEBUG(printf("     found %d,%d,%d,%d\n", a,b,c,d));
          int ret = (*callback)(g,m,n, a,b,c,d, data);
          if (ret) return ret;
        }
      }
    }
  }
  return 0;
}
static int
INTERNAL_diamond_count_callback (const graph *g, int m, int n,
                                int a, int b, int c, int d, void *data)
{
  (*(int*)data)++;
  DEBUG(printf("  diamond %d,%d,%d,%d  (count %d)\n", a,b,c,d, *(int*)data));
  return 0;
}
int
nautyextra_diamond_count (const graph *g, int m, int n, int induced)
{
  int count = 0;
  nautyextra_diamond_search (g,m,n, induced,
                             &INTERNAL_diamond_count_callback, &count);
  return count;
}


/* a,b,c,d are vertices of a diamond in undirected graph g,m,n
   (two triangles with a common edge).

   .      c
   .     / \
   .    a---b
   .     \ /
   .      d

   Return TRUE if both triangles are odd.  An odd triangle has some other
   vertex which has edges going to an odd number of a,b,c, so either 1 or 3
   edges.  Likewise a,b,d odd.

   If there is an edge c-d then that causes d to have edges to all 3 of
   a,b,c so that it is odd, and vice-versa c to a,b,d so it is odd, and so
   that the return is TRUE.
*/
int
nautyextra_diamond_is_both_odd (const graph *g, int m,
                                int a, int b, int c, int d)
{
  DEBUG(printf("nautyextra_diamond_is_both_odd() m=%d %d,%d,%d,%d\n",
               m, a,b,c,d));
  /* assert (n >= 0); */
  /* assert (m >= SETWORDSNEEDED(n)); */
  /* assert (0 <= a && a < n); */
  /* assert (0 <= b && b < n); */
  /* assert (0 <= c && c < n); */
  /* assert (0 <= d && d < n); */
  assert (a != b);
  assert (a != c);
  assert (a != d);
  assert (b != c);
  assert (b != d);
  assert (c != d);
  assert (ISELEMENT(GRAPHROW(g,a,m),b));
  assert (ISELEMENT(GRAPHROW(g,a,m),c));
  assert (ISELEMENT(GRAPHROW(g,a,m),d));
  assert (ISELEMENT(GRAPHROW(g,b,m),a));
  assert (ISELEMENT(GRAPHROW(g,b,m),c));
  assert (ISELEMENT(GRAPHROW(g,b,m),d));
  assert (ISELEMENT(GRAPHROW(g,c,m),a));
  assert (ISELEMENT(GRAPHROW(g,c,m),b));
  assert (ISELEMENT(GRAPHROW(g,d,m),a));
  assert (ISELEMENT(GRAPHROW(g,d,m),b));

  /* a^b^c = 0 if even, 1 if odd
     looking for both a^b^c and a^b^d somewhere */
  int i;
  const set *a_row = GRAPHROW(g,a,m);
  const set *b_row = GRAPHROW(g,b,m);
  const set *c_row = GRAPHROW(g,c,m);
  const set *d_row = GRAPHROW(g,d,m);

  DEBUG(printf("  a row %s\n", nautyextra_set_str(a_row,m*WORDSIZE)));
  DEBUG(printf("  b row %s\n", nautyextra_set_str(b_row,m*WORDSIZE)));
  DEBUG(printf("  c row %s\n", nautyextra_set_str(c_row,m*WORDSIZE)));
  DEBUG(printf("  d row %s\n", nautyextra_set_str(d_row,m*WORDSIZE)));

  int both = 0;
  for (i = 0; i < m; i++) {
    setword a_xor_b = a_row[i] ^ b_row[i];
    if (a_xor_b ^ c_row[i]) both |= 1;
    if (a_xor_b ^ d_row[i]) both |= 2;
    if (both == 3) {
      DEBUG(printf("  yes\n"));
      return TRUE;
    }
  }
  DEBUG(printf("  no, only %d\n", both));
  return FALSE;
}

/* a,b,c are vertices a triangle in g,m.

   .      c
   .     / \
   .    a---b

   Return TRUE if the triangle is even.  For an even triangle, any other
   vertex has edges to an even number of the vertices a,b,c, so either no
   edge to any, or to exactly 2 of them.
*/
int
nautyextra_triangle_is_even (const graph *g, int m, int a, int b, int c)
{
  /* don't need n, just m is enough to search rows a,b,c */
  /* assert (n >= 0); */
  /* assert (m >= SETWORDSNEEDED(n)); */
  /* assert (0 <= a && a < n); */
  /* assert (0 <= b && b < n); */
  /* assert (0 <= c && c < n); */

  assert (a != b);
  assert (a != c);
  assert (b != c);
  assert (ISELEMENT(GRAPHROW(g,a,m),b));
  assert (ISELEMENT(GRAPHROW(g,a,m),c));
  assert (ISELEMENT(GRAPHROW(g,b,m),a));
  assert (ISELEMENT(GRAPHROW(g,b,m),c));
  assert (ISELEMENT(GRAPHROW(g,c,m),a));
  assert (ISELEMENT(GRAPHROW(g,c,m),b));

  const set *a_row = GRAPHROW(g,a,m);
  const set *b_row = GRAPHROW(g,b,m);
  const set *c_row = GRAPHROW(g,c,m);
  int i;
  for (i = 0; i < m; i++)
    if (a_row[i] ^ b_row[i] ^ c_row[i])    /* a^b^c = 0 if even, 1 if odd */
      return FALSE;

  return TRUE;
}

/* ------------------------------------------------------------------------- */

#ifndef NDEBUG
/* Return true if "ancestor" is an ancestor of v, with protection against
   infinite loops if vpar is circular.
*/
static int
debug_safe_is_ancestor_vpar (const int *vpar, int v, int ancestor)
{
  DEBUG(printf("debug_safe_is_ancestor_vpar() %d to %d\n", v, ancestor));
  assert (ancestor >= 1);
  int count = 0;
  int n = 0;
  do {
    n = MAX(n,v+3);
    if (++count > n) break;
    DEBUG(printf(" v=%d\n", v));
    assert (v >= 1);
    if (v == ancestor) return TRUE;
    v = vpar[v];
  } while (v);
  return FALSE;
}
#endif

/* v1 up to ancestor p then down to v2 is a path.
   Set *c1p and *c2p to the centre vertices of that path.
   If the path length is even there is a single centre and c1,c2 are the same.
   If the path length is odd there are two centres and c1 is the first.
   Can have v1==v2==p==0 in which case set c1=c2=0;
   If c2p is NULL then that vertex is not stored.  This can be used if the
   path is known to have a single centre.

   FIXME: Omit p?  Or p=0 common ancestor unknown.
*/
void
nautyextra_path_centres_vpar (const int *vpar, int v1, int v2, int p,
                              int *c1p, int *c2p)
{
  assert (debug_safe_is_ancestor_vpar(vpar,v1,p));
  assert (debug_safe_is_ancestor_vpar(vpar,v2,p));
  DEBUG(printf("nautyextra_path_centres_vpar() %d %d p %d\n", v1,v2,p));
  int c1 = 0;
  int c2 = 0;
  int swap = 0;
  if (v1) {
    assert (p >= 1);
    for (;;) {
      DEBUG(printf(" at %d %d\n", v1, v2));
      assert (v1 >= 1);
      assert (v2 >= 1);
      if (v2 == p) {
        break;
      }
      if (v1 == p) {
        v1 = v2;
        swap = 1;
        break;
      }
      c1 = v1;
      c2 = v2;
      v1 = vpar[v1];
      v2 = vpar[v2];
    }

    /* centres of v1 -- p */
    c1 = v1;
    c2 = vpar[v1];
    for (;;) {
      DEBUG(printf(" at v=%d c=%d,%d\n", v1, c1,c2));
      assert (v1 >= 1);

      if (v1 == p) {
        DEBUG(printf(" even length path, single centre\n"));
        c2 = c1;
        break;
      }
      v1 = vpar[v1];
      if (v1 == p) {
        DEBUG(printf(" odd length path, two centres\n"));
        break;
      }
      v1 = vpar[v1];

      /* centres advance once for each two v1 */
      c1 = c2;
      c2 = vpar[c2];
    }

    assert (c1 >= 1);
    assert (c2 >= 1);
  }
  DEBUG(printf(" return %d,%d\n",c1,c2));
  if (swap) { int t = c1; c1 = c2; c2 = t; }
  *c1p = c1;
  if (c2) *c2p = c2;
}

/* Set hist[d] to the number of vertices in vpar[1..n] which have degree d.
   hist[] must have room for n ints.

   FIXME: Maybe return max degree.  Max degree would be <= n-1.
   Maybe memset only as much as needed, so application knowing a degree
   limit could give small array.
*/
void
nautyextra_degree_histogram_vpar (const int *vpar, int n, int *hist)
{
  int v;
#if MAXN
  int degrees[MAXN+1];    /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  DYNALLSTAT(int,degrees,degrees_sz);
  DYNALLOC1(int,degrees,degrees_sz,n+1,"nautyextra_degree_histogram_vpar");
#endif
  assert (n >= 0);

  nautyextra_degrees_array_vpar(vpar,n,degrees);
  memset (hist, '\0', n*sizeof(hist[0]));
  for (v = 1; v <= n; v++) {
    assert (0 <= degrees[v] && degrees[v] < n);
    hist[degrees[v]]++;
  }
}

/* Return the number of vertices in vpar[1..n] which have children.
   (This differs from non-leaf vertices in that a root with one child has
   children, but is a leaf so not non-leaf.)
*/
int
nautyextra_parent_count_vpar (const int *vpar, int n)
{
  int m = SETWORDSNEEDED(n+1);       /* 1 to n inclusive */
#if MAXN
  set seen[SETWORDSNEEDED(MAXN+1)];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(set,seen,m,"nautyextra_parent_count_vpar");
#endif
  memset (seen,'\0',m*sizeof(seen[0]));
  int v;
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    if (p) ADDELEMENT(seen,p);
  }
  return setsize(seen, m);
}

/* Return the number of childless vertices in vpar[1..n], being not the
   parent of anything else.
   This differs from leaf vertices in that a root is a leaf if it has just
   one child.
*/
int
nautyextra_childless_count_vpar (const int *vpar, int n)
{
  return n - nautyextra_parent_count_vpar(vpar,n);
}

/* g is an undirected graph.
   Return the number of edges in g.

   cf nauty degstats() which gives edges and more.
*/
int
nautyextra_num_edges (const graph *g, int m, int n)
{
  DEBUG(printf("nautyextra_num_edges() m=%d,n=%d\n", m,n));
  int count = 0;
  int v;
  for (v = 0; v < n; v++) {
    count += setsize((set*)g, m);
    g += m;
  }
  DEBUG(printf("  arcs %d\n", count));
  assert ((count & 1) == 0);
  return count >> 1;
}

/* For set depths[1..n] to the depth of each vertex in tree vpar[1..n].
   Depth is distance of a vertex from its root.  Each root is depth 0, their
   children depth 1, etc.
*/
void
nautyextra_depths_vpar (const int *vpar, int n, int *depths)
{
  DEBUG(printf("nautyextra_depths_vpar() n=%d\n", n));

  /* At each i search upwards for a depths[v] already known, or the root.
     depths[v]=-1 initially for not known, then >=0 when known.
     Keep a linked list downwards through the v traversed, stored in depths[].
     Fill them in back down again starting from the found depth or from -1
     above the root.  Pre-increment means the latter sets depths[root] = 0.
  */

  int i;
  for (i = 1; i <= n; i++)
    depths[i] = -1;

  for (i = 1; i <= n; i++) {
    int v = i;
    int prev = 0;
    int d;
    for (;;) {
      d = depths[v];
      if (d >= 0) break;
      depths[v] = prev; prev = v; v = vpar[v];
      if (!v) break;
    }
    while (prev) {
      v = prev; prev = depths[prev]; depths[v] = ++d;
    }
  }
}

/* For tree vpar[1..n], set heights[1..n] so that height[v] is the height of
   the subtree below each vertex v.
   If v is a leaf then its subtree height is 0.
   If v has just one level of children then height 1, etc.
*/
void
nautyextra_subtree_heights_vpar (const int *vpar, int n, int *heights)
{
  DEBUG(printf("nautyextra_subtree_heights_vpar() n=%d\n", n);
        printf(" vpar "); nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
#if MAXN
  int num_children[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(int,num_children,n+1,"nautyextra_subtree_heights_vpar");
#endif

  nautyextra_num_children_array_vpar(vpar,n, num_children);

  memset(heights+1, '\0', n*sizeof(*heights));
  for (i = 1; i <= n; i++) {
    DEBUG(printf("i=%d\n", i));
    int v = i;
    if (num_children[v]) continue;

    for (;;) {
      int h = heights[v] + 1;
      DEBUG(printf("consider v=%d parent %d\n", v, vpar[v]));
      assert (num_children[v] == 0);

      num_children[v] = -1;
      v = vpar[v];     /* parent */
      if (! v) break;

      int v_height = heights[v];
      heights[v] = MAX(v_height, h);
      DEBUG(printf(" set v=%d height %d\n", v, heights[v]));

      assert(num_children[v] > 0);
      if (--num_children[v]) break;
    }
  }
}

/* Return the number of roots in vpar[1..n] whose height is equal to the
   maximum height in the whole forest.

   FIXME: Name num components ?
   nautyextra_height_num_components(vpar,n) ?
*/
int
nautyextra_height_num_components_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_height_num_components_vpar() n=%d\n", n);
        printf("vpar "); nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
  struct data {
    int num_children;
    int height;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  /* WORKSPACE(struct data,data,n+1,"nautyextra_height_num_components_vpar"); */
  struct data *data = alloca((n+1)*sizeof(struct data));
#endif
  int maximum_height = 0;
  int num_roots = 0;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children) continue;

    for (;;) {
      int h = data[v].height;
      assert (data[v].num_children == 0);
      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        /* root */
        if (h > maximum_height) {
          maximum_height = h;
          num_roots = 1;
        } else {
          num_roots += (h == maximum_height);
        }
        break;
      }

      h++;
      int v_height = data[v].height;
      data[v].height = MAX(v_height, h);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;
    }
  }
  return num_roots;
}

/* Return the height of tree vpar[1..n].
   This is the depth of the deepest vertex, so the eccentricity of the root.
   If vpar is a forest then height is the maximum of its component tree
   heights.
*/
int
nautyextra_height_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_height_vpar() %d\n", n);
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
  struct data {
    int num_children;
    int height;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_height_vpar");
#endif
  int height = -1;  /* for no vertices */

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }
    for (;;) {
      int c_height = data[v].height;
      DEBUG(printf("consider v=%d parent %d\n", v, vpar[v]));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* to parent */
      if (! v) {
        /* v is a root, accumulate its height into the return */
        height = MAX(height,c_height);
        break;
      }

      int new_height = c_height + 1;
      int v_height = data[v].height;
      data[v].height = MAX(v_height, new_height);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        break;
      }
    }
  }
  return height;
}

/* Return the height of the second-highest vertex in tree vpar[1..n].
   If vpar has two vertices both at its height then 2height == height.
*/
#define HEIGHT2_ACCUMULATE(h,h2, new_h,new_h2)  \
  do {                                          \
    if ((new_h) >= (h)) {                       \
      (h2) = MAX(h, new_h2);                    \
      (h) = (new_h);                            \
    } else {                                    \
      (h2) = MAX(h2, new_h);                    \
    }                                           \
  } while (0)
int
nautyextra_2height_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_2height_vpar() ");
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
  struct data {
    int num_children;
    int height;   /* highest        */
    int height2;  /* second highest */
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_2height_vpar");
#endif
  int height  = -1;
  int height2 = -1;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].height2 = -1;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children) continue;
    for (;;) {
      int c_height  = data[v].height;
      int c_height2 = data[v].height2;
      DEBUG(printf("consider v=%d  heights %d,%d\n", v, c_height,c_height2));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* to parent */
      if (! v) {
        /* at a root */
        HEIGHT2_ACCUMULATE(height,height2, c_height,c_height2);
        DEBUG(printf("  at root accumulate to %d,%d\n", height,height2));
        break;
      }

      int prospective_height  = c_height  + 1;
      int prospective_height2 = c_height2 + 1;
      int v_height  = data[v].height;
      int v_height2 = data[v].height2;
      DEBUG(printf("  prospective %d,%d into %d,%d\n",
                   prospective_height,prospective_height2,
                   v_height,v_height2));
      HEIGHT2_ACCUMULATE(v_height,v_height2,
                         prospective_height,prospective_height2);
      data[v].height  = v_height;
      data[v].height2 = v_height2;
      DEBUG(printf("  gives %d,%d\n", v_height,v_height2));

      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;
    }
  }
  return height2;
}


/* ------------------------------------------------------------------------- */
/* Eccentricity */

/* graph g or sparsegraph sg are connected undirected graphs.
   Return the eccentricity of vertex v, being the maximum distance to any
   other vertex (along the shortest path to it).
   If destination_ptr is not NULL then a vertex attaining the eccentricity
   is stored to *destination_ptr.  If there's more than one attaining the
   eccentricity then it's unspecified which is stored.

   FIXME: Not sure about *destination_ptr.
*/
int
nautyextra_eccentricity_of_vertex (const graph *g, int m, int n, int v,
                                   int *destination_ptr)
{
  int to, max_dist = -1;
#if MAXN
  int dist[MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE(int,dist,n+1,"nautyextra_eccentricity_of_vertex");
#endif
  DEBUG(printf("nautyextra_eccentricity_of_vertex() n=%d\n", n));
  assert (0 <= v && v < n);

  find_dist((graph*)g, m,n, v, dist);
  for (to = 0; to < n; to++) {
    int d = dist[to];
    assert (dist[to] < n);
    if (d > max_dist) {
      max_dist = d;
      if (destination_ptr) *destination_ptr = to;
    }
  }
  return max_dist;
}
int
nautyextra_eccentricity_of_vertex_sg (const sparsegraph *sg, int v,
                                      int *destination_ptr)
{
  int to, max_dist = -1;
  int n = sg->nv;
  assert (0 <= v && v < n);
#if MAXN
  int dist[MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE(int,dist,n,"nautyextra_eccentricity_of_vertex_sg");
#endif
  DEBUG(printf("nautyextra_eccentricity_of_vertex_sg() n=%d\n", n));

  distvals((sparsegraph*) sg, v, dist, n);
  for (to = 0; to < n; to++) {
    int d = dist[to];
    assert (d < n);
    if (d > max_dist) {
      max_dist = d;
      if (destination_ptr) *destination_ptr = to;
    }
  }
  return max_dist;
}

/* Set ecc[1..n] to the eccentricities of vertices 1..n.
   This is the same as ecc[v] = nautyextra_eccentricity_of_vertex_vpar(vpar,n,v)
   but calculated all together.
*/
void
nautyextra_eccentricities_vpar (const int *vpar, int n, int *ecc)
{
  DEBUG(printf("nautyextra_eccentricities_vpar() n=%d\n", n);
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
  struct data {
    int num_children;
    int ecc_second;
    int ecc_above;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_eccentricities_vpar");
#endif

  memset(data, '\0', (n+1)*sizeof(*data));
  memset(ecc+1, '\0', n*sizeof(*ecc));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  int rev_v = 0;
  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children) continue;
    for (;;) {
      data[v].ecc_above = rev_v;
      rev_v = v;

      int c_ecc = ecc[v];
      DEBUG(printf("consider v=%d parent %d\n", v, vpar[v]));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* to parent */
      if (! v) break;

      int b = c_ecc + 1;
      int v_ecc = ecc[v];
      if (b > v_ecc) {
        data[v].ecc_second = v_ecc;
        ecc[v] = b;
      } else {
        data[v].ecc_second = MAX(data[v].ecc_second, b);
      }

      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;
    }
  }

  int v = rev_v;
  while (v) {
    int p = vpar[v];
    DEBUG(printf("rev v=%d ecc below %d second %d  p=%d \n",
                 v, ecc[v], data[v].ecc_second, p));
    int next_v = data[v].ecc_above;

    if (p) {
      int p_other = (ecc[p]==ecc[v]+1 ? data[p].ecc_second : ecc[p]);
      int b = MAX(data[p].ecc_above, p_other) + 1;
      DEBUG(printf("  p above %d ecc below %d second %d  use second %d  p other %d  b %d\n",
                   data[p].ecc_above, ecc[p], data[p].ecc_second, ecc[p]==ecc[v]+1, p_other, b));

      data[v].ecc_above = b;
      int v_ecc = ecc[v];
      ecc[v] = MAX(v_ecc,b);
      DEBUG(printf("  gives ecc %d above %d\n",
                   ecc[v], data[v].ecc_above));
    } else {
      data[v].ecc_above = 0;
    }
    v = next_v;
  }
}


/* Set *c1p and *c2p to the centre of tree vpar[1..n].
   vpar must be a single tree, not a forest.
   The centre is the vertices of minimum eccentricity.
   There can be either 1 unique minimum, or 2 vertices of equal minimum.

   Equivalently these are the 1 or 2 vertices which are the middle of any
   path attaining the diameter.  1 when diameter even (so odd number of
   vertices on it), or 2 when diameter odd (so even number of vertices on
   it).

   If there is just 1 centre then set *c1p==*c2p the same.
   If n==0 then there are no vertices at all so set *c1p=*c2p=0.

   FIXME: Waiting for nautyextra_path_centres_vpar().
*/
void
nautyextra_centres_vpar (const int *vpar, int n, int *c1p, int *c2p)
{
  struct nautyextra_diameter_info  info;
  nautyextra_diameter_info_vpar(vpar,n, &info);
  nautyextra_path_centres_vpar (vpar, info.end_v1, info.end_v2,
                                info.end_ancestor, c1p, c2p);
}


/* ------------------------------------------------------------------------- */
/* Wiener Index

        1
   W = --- *    sum     distance(u,v)
        2    vertices u,v

   The factor 1/2 has the effect of counting a distince between vertices u,v
   in just one direction.

   If terminal=0 then the index is between all vertex pairs.
   If terminal=1 then the terminal Wiener index which is paths between
   terminal vertices (leaf vertices) only.
*/

/* Return the Wiener index of graph g.
   Must be an undirected graph, so bi-directional edges both u->v and v->u.
*/
void
nautyextra_Wiener_index (const graph *g, int m, int n, int terminal,
                         struct nautyextra_Wiener_info *out)
{
  int from, to;
  int diameter = -1;
  NAUTYEXTRA_TYPE_ULL W = 0;
#if MAXN
  int dist[MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE(int,dist,n+1,"nautyextra_Wiener_index");
#endif
  assert (n >= 0);
  assert (terminal == 0 || terminal == 1);
  out->diameter_from = 0;
  out->diameter_to   = 0;

  /* setsize(g,m); */

  for (from = 0; from < n; from++) {
    find_dist((graph*)g, m,n, from, dist);
    for (to = 0; to < n; to++) {
      W += dist[to];
      if (dist[to] > diameter) {
        diameter = dist[to];
        out->diameter_from = from;
        out->diameter_to   = to;
      }
    }
  }

  assert ((W & 1) == 0);
  out->W = W/2;
  out->num = n;
  out->diameter = diameter;
  out->diameter_from = -1;
  out->diameter_to = -1;
}

/* sg is an connected undirected (arcs both ways) graph.
   Return the Wiener index.

   ENHANCE-ME: distvals() on each vertex is not efficient for all paths
   shortest paths.
*/
void
nautyextra_Wiener_index_sg (const sparsegraph *sg, int terminal,
                            struct nautyextra_Wiener_info *out)
{
  DEBUG(printf("nautyextra_Wiener_index_sg()  terminal=%d\n", terminal));

  int         n = sg->nv;
  const int *dd = sg->d;
  int from, to;
  NAUTYEXTRA_TYPE_ULL W = 0;
  int diameter = -1;
  int n_endpoints = n;
#if MAXN
  int dist[MAXN];
  assert(sg->nv <= MAXN);
#else
  WORKSPACE(int,dist,n,"nautyextra_Wiener_index_sg");
#endif
  assert (sg->nv >= 0);
  assert (terminal == 0 || terminal == 1);
  out->diameter_from = 0;
  out->diameter_to   = 0;

  for (from = 0; from < n; from++) {
    if (dd[from] > 1 && terminal) {
      n_endpoints--;
      continue;
    }
    distvals((sparsegraph*)sg, from, dist, n);
    for (to = 0; to < n; to++) {
      if (dd[to] > 1 && terminal) continue;
      assert (dist[to] < n);
      W += dist[to];
      if (dist[to] > diameter) {
        diameter = dist[to];
        out->diameter_from = from;
        out->diameter_to   = to;
      }
    }
  }

  DEBUG(printf("  final 2*W = %Lu\n", (unsigned long long) W));
  assert ((W & 1) == 0);

  out->W = W/2;
  out->num = n_endpoints;
  out->diameter = diameter;
}

/* Return the Wiener index of tree vpar[1..n].

   The strategy is to count crossings of each edge.  There is some number b
   vertices below an edge, and the remaining n-b above, for product b*(n-b)
   paths crossing that edge, and so that distance in W.

   At a leaf vertex, its edge to parent is b=1 (the leaf vertex) below.
   Having counted the edge to parent, the vertex is reckoned as removed by
   adding its b to the parent, and decreasing the parent num_children by 1.
   If num_children reaches 0 then the parent has become a leaf, ready to be
   acted on.

   Total work is 3 passes over vpar, being 1 to make num_children, 1 for
   locating leaf nodes work upwards from, and 1 for those leaves upwards (an
   out-of-sequence chain up through all eventually).

   The terminal option starts num_below at -1 on non-leaf vertices, so they
   do not contribute to the num_below accumulated.  This is an extra initial
   pass over the vertices.

   FIXME: Not certain if terminal should be a parameter or separate function.
 */
void
nautyextra_Wiener_index_vpar (const int *vpar, int n, int terminal,
                               struct nautyextra_Wiener_info *out)
{
  int i;
  NAUTYEXTRA_TYPE_ULL W = 0;
  struct data {
    int num_below;     /* number of vertices below v, not including v itself */
    int num_children;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_Wiener_index_vpar");
#endif
  assert (n >= 0);
  assert (terminal == 0 || terminal == 1);
  out->diameter_from = -1;
  out->diameter_to   = -1;

  DEBUG(printf("nautyextra_Wiener_index_vpar()");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  int n_endpoints = n;
  if (terminal) {
    /* num_children == 0 is a leaf,
       except root vpar[i]==0 is num_children <= 1 leaf.
       If root num_children=0 then its W is 0 so just num_children=1 for
       root as leaf. */
    for (i = 1; i <= n; i++) {
      int is_root = (vpar[i] == 0);
      int is_nonleaf = (data[i].num_children > is_root);
      n_endpoints -= is_nonleaf;
      data[i].num_below = - is_nonleaf;
      DEBUG(printf("  v=%d is_root %d is_nonleaf %d  (num_children=%d)\n",
                   i, is_root, is_nonleaf, data[i].num_children));
    }
  }
  DEBUG(printf("  n_endpoints %d\n", n_endpoints));

  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d num_below %d\n",
                 v, data[v].num_children, data[v].num_below));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      int b = data[v].num_below + 1;
      DEBUG(printf("consider v=%d num_below %d\n", v, data[v].num_below));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        DEBUG(printf("  no parent\n"));
        break;
      }

      W += (NAUTYEXTRA_TYPE_ULL) b * (n_endpoints - b);
      DEBUG(printf("  Wprod %Lu\n", (unsigned long long) b * (n - b)));

      data[v].num_below += b;

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d\n", v));
        break;
      }
    }
  }

  out->W = W;
  out->diameter = -1;
  out->num = n_endpoints;
}



/* ------------------------------------------------------------------------- */
/* Matchings */

/* vpar[1..n] is a tree.
   ss is a set of n many bits 0..n-1 representing vertices 1..n which are
   the child vertex v of a pair v and its parent.
   Return TRUE if these edges are a matching in vpar.
   Return FALSE if not.
   This means checking all parents of elements of ss exist and are distinct.

   FIXME: Should the set be numbered 1..n too?
*/
int
nautyextra_is_matching_vpar_set (const int *vpar, int n, const set *ss)
{
  DEBUG(printf("nautyextra_is_matching_vpar_set() n=%d  %s\n",
               n, nautyextra_set_str(ss,n)));

  if (n) {   /* nextelement() doesn't take m=0 */
    int m = SETWORDSNEEDED(n);
#if MAXN
    set seen[SETWORDSNEEDED(MAXN)];  /* p0 numbering 0 to n-1 */
    assert(n <= MAXN);
#else
    WORKSPACE(set,seen,m,"nautyextra_is_matching_vpar_set");
#endif
    memcpy(seen, ss, m*sizeof(seen[0]));

    int v0;
    for (v0 = -1; (v0 = nextelement((set*)ss,m,v0)) >= 0; ) {
      assert (0 <= v0 && v0 < n);
      int v1 = v0 + 1;
      int p1 = vpar[v1];
      if (p1 == 0) return FALSE;   /* ss contains the root, not a matching */
      int p0 = p1 - 1;

      assert (0 <= p0 && p0 < n);
      if (ISELEMENT(seen,p0)) return FALSE;   /* overlap, not a matching */
      ADDELEMENT(seen,p0);
    }
  }
  return TRUE;
}


/* ------------------------------------------------------------------------- */
/* Make Various Graphs */

/* Set g to a path of n vertices, 0 -- 1 -- 2 -- ... -- n-1.
   cf. genspecialg.c makepath() doing similar for sparsegraph, with option
   for digraph too.
   digraph=0 for undirected.
   digraph=1 for forward 0-->1.
   digraph=2 for reverse 0<--1.
*/
void
nautyextra_make_path (graph *g, int m, int n, int digraph)
{
  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g, m,n);

  int to;
  for (to=1; to < n; to++) {
    int from = to-1;
    if (digraph != 2) ADDONEARC(g, from,to, m);
    if (digraph != 1) ADDONEARC(g, to,from, m);
  }
}

/* Set g to an undirected star of n vertices.
   0 is the centre vertex and 1..n-1 the arms.

   digraph=0 for undirected.
   digraph=1 for forward 0-->1.
   digraph=2 for reverse 0<--1.
*/
void
nautyextra_make_star (graph *g, int m, int n, int digraph)
{
  assert(n >= 0);
  assert (digraph == 0 || digraph == 1 || digraph == 2);

  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g, m,n);
  int to;
  for (to = 1 ; to < n; to++) {
    if (digraph != 2) ADDONEARC(g, 0,to, m);
    if (digraph != 1) ADDONEARC(g, to,0, m);
  }
}

/* Set vpar[1..n] to a star of n vertices where each arm has 2 vertices, or
   if n even then one arm 1 vertex and the rest 2 vertices.
   vpar must have space for n+1 ints.

         *                  *
         |                  |
         *     n=8          *     n=9
         |                  |
   *--*--*--*--*      *--*--*--*--*
         |                  |
         *                  *
                            |
                            *
*/
void
nautyextra_make_star_arms2_vpar (int *vpar, int n)
{
  int i;
  if (n >= 1) {
    vpar[1] = 0;  /* root */
    for (i = 2; i <= n; i++)
      vpar[i] = (i&1 ? i-1 : 1);
  }
}

/* nautyextra_make_bistar_vpar() sets vpar[1..n] to a bi-star a,b.
   This is two stars of a and b many vertices respectively, with a further
   edge joining their centres.

           *       *
           |       |
       *---*-------*---*    a=4,b=4,length=1
           |       |        n=8
           *       *

   "length" is the number of edges in the star arms and centre connections.
   length=1 is the usual bi-star.

           *       *
           |       |        a=4,b=4,length=1
           *       *        n=15
           |       |
   *---*---*---*---*---*---*
           |       |
           *       *
           |       |
           *       *

   The length extra inserts inThe number of vertices is

       n =   (a+b)                bi-star vertices
           + (a+b-1)*(length-1)   extra vertices in edges
         = (a+b-1)*(length+1) + 1

   When length=1 this is simply n=a+b.
   vpar must have space for n+1 ints.

   GP-Test  ('a+'b) + ('a+'b-1)*('len-1) == ('a+'b-1)*'len + 1

   FIXME: Leave extra length to subdivide ...
*/
int
nautyextra_make_bistar_n (int a, int b, int length)
{
  return (a==0 && b==0 ? 0 : (a+b-1)*length + 1);
}
void
nautyextra_make_bistar_vpar (int *vpar, int a, int b, int length)
{
  DEBUG(printf("nautyextra_make_bistar_vpar() %d,%d %d  (n=%d)\n",
               a,b,length,
               nautyextra_make_bistar_n(a,b,length)));
  if (a==0 && b==0) return;  /* empty tree */
  int centre = 1;
  vpar[centre] = 0;
  int c;
  int v = 1;
  int arms = a;  /* a many arms, incl middle connection */
  for (c = 1; c <= 2; c++) {
    DEBUG(printf("centre %d\n", centre));
    int i;
    for (i = 1; i <= arms; i++) {
      DEBUG(printf(" arm v=%d\n", v));
      int prev = centre;
      int l;
      for (l = 1; l <= length; l++) {
        v++;
        vpar[v] = prev;
        prev = v;
      }
    }
    centre = v;      /* last arm end as middle of b */
    arms = b-1;      /* b arms */
  }
  DEBUG(printf("end v=%d\n", v));
  assert(v == nautyextra_make_bistar_n(a,b,length));
}

/* Set vpar[1..n] to a complete c-ary tree of height h,
   with n=nautyextra_make_complete_tree_size().
   Vertex 1 is the tree root.
   h=0 is an empty tree.
   vpar must have space for n+1 ints.

             *
           /   \        c=2, h=3
         *       *
        / \     / \
       *   *   *   *
*/
void
nautyextra_make_complete_tree_vpar (int *vpar, int c, int h)
{
  DEBUG(printf("nautyextra_make_complete_tree_vpar() c=%d h=%d\n", c,h));
  if (h >= 1) {
    int *dst = vpar+1;
    *dst++ = 0;  /* vpar[1] = root */
    int rowlen = 1;
    int parent = 1;
    int i;
    for (i = 2; i <= h; i++) {
      DEBUG(printf("row %d at v=%d\n", i, dst-(vpar+1)));
      int j;
      for (j = 0; j < rowlen; j++) {
        int k;
        for (k = 0; k < c; k++) {
          *dst++ = parent;
        }
        parent++;
      }
      rowlen *= c;
    }
    DEBUG(printf("end %d, cf size %d\n",
                 dst-(vpar+1), nautyextra_make_complete_tree_size(c,h)));
    assert (dst-(vpar+1) == nautyextra_make_complete_tree_size(c,h));
  }
}

/* Return the number of vertices in a complete c-ary tree of height h,
   per nautyextra_make_complete_tree_vpar() above.
   This is successive rows 1 + c + c^2 + ... + c^(h-1) = (c^h - 1)/(c-1).

   GP-Test  matrix(5,5,c,h,c++;h--; sum(i=0,h-1,c^i)) == \
   GP-Test  matrix(5,5,c,h,c++;h--; (c^h-1)/(c-1))
*/
int
nautyextra_make_complete_tree_size (int c, int h)
{
  double d = (pow(c,h) - 1) / (c-1);
  int v = (int) d;
  if (v != d) {
    DEBUG(printf("nautyextra_make_complete_tree_size() c=%d h=%d  %g %d\n",
                 c,h, d,v));
    INTERNAL_overflow();
  }
  return v;
}

/* Set g to a cycle of n vertices.

   directed=0 gives undirected, so arcs both v->v+1 and v+1->v.
   directed=1 gives directed with forward edges v->v+1.
   directed=2 gives directed with reverse edges v+1->v.
   v+1 here is modulo n.

   Cf. genspecialg.c makecycle() which is for sg.
*/
void
nautyextra_make_cycle (graph *g, int m, int n, int directed)
{
  /* ENHANCE-ME: Could store row bit patterns like 000101000 directly. */

  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (directed==0 || directed==1 || directed==2);

  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g, m,n);
  int to;
  for (to = 1 ; to < n; to++) {
    int from = to-1;
    if (directed != 2) ADDONEARC(g, from,to, m);
    if (directed != 1) ADDONEARC(g, to,from, m);
  }
  if (n >= 2) {
    int to = 0;
    int from = n-1;
    if (directed != 2) ADDONEARC(g, from,to, m);
    if (directed != 1) ADDONEARC(g, to,from, m);
  }
}

/* Set sg to a cycle of n vertices.
   About the same as genspecialg.c makecycle(), but here allow n=0 too.
*/
void
nautyextra_make_cycle_sg (sparsegraph *sg, int n)
{
  assert (n >= 0);
  int nde = 2*(n - (n==1 || n==2));
  SG_ALLOC(*sg,n,nde, "nautyextra_make_cycle_sg");
  sg->nv = n;
  sg->nde = nde;

  int    *dd = sg->d;
  int    *ee = sg->e;
  size_t *vv = sg->v;

  int epos = 0;
  switch (n) {
  case 0:
    /* empty */
    break;
  case 1:
    /* 1 vertex, no edges */
    dd[0] = 0;
    vv[0] = 0;
    break;
  case 2:
    /* 2 edges only, so path-2 */
    dd[0] = 1;
    dd[1] = 1;
    vv[0] = 0;
    vv[1] = 1;
    ee[0] = 1;
    ee[1] = 0;
    epos = 2;
    break;
  default:
    {
      int v;
      for (v = 0; v < n; v++) {
        dd[v] = 2;
        vv[v] = epos;
        ee[epos++] = (v==0 ? n-1 : v-1);
        ee[epos++] = (v + 1) % n;
      }
    }
    break;
  }
  assert (epos == nde);
}

/* Set g,m,n to an undirected grid graph of xsize by ysize many vertices.
   Must have n == xsize*ysize.
   g must have space for an n vertex graph.
   Any existing contents in g are cleared.

       *---*---*---*---*
       |   |   |   |   |
       *---*---*---*---*   xsize=5, ysize=4
       |   |   |   |   |
       *---*---*---*---*
       |   |   |   |   |
       *---*---*---*---*

   Cf nauty genspecialg.c makegrid() for sparsegraph, any dimensions (not
   just 2) and digraph option.
*/
void
nautyextra_make_grid (graph *g, int m, int n, int xsize, int ysize)
{
  assert (n == xsize*ysize);

  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g,m,n);

  int xfrom, yfrom;
  for (xfrom = 0; xfrom < xsize; xfrom++) {
    for (yfrom = 0; yfrom < ysize; yfrom++) {
      int from = xfrom + yfrom*ysize;

      if (xfrom > 0)       ADDONEEDGE(g, from,from-1, m);
      if (yfrom > 0)       ADDONEEDGE(g, from,from-ysize, m);
      if (xfrom < xsize-1) ADDONEEDGE(g, from,from+1, m);
      if (yfrom < ysize-1) ADDONEEDGE(g, from,from+ysize, m);
    }
  }
}

/* Set g to an undirected hypercube graph of d dimensions.
   Must have n == 2^d and g enough space for m,n.
   Any existing contents in g are cleared.

   Cf nauty genspecialg.c static makehypercube() for sparsegraph.

   FIXME: directed=1,2 for away or to 0 ...
*/
void
nautyextra_make_hypercube (graph *g, int m, int n, int d)
{
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (d >= 0);
  assert (d < CHAR_BIT*sizeof(int));
  assert (n == 1<<d);
  EMPTYGRAPH(g,m,n);
  int from;
  for (from = 0; from < n; from++) {
    int b;
    for (b = 1; b < n; b<<=1) {
      int to = from ^ b;
      ADDONEARC(g, from,to, m);
    }
  }
}

/* Set g to an undirected complete bipartite of a,b vertices.
   Must have n == a+b and g enough space for that.
   Any existing contents in g are cleared.

   cf genspecialg.c static makecompletebipartite() for sparsegraph

   FIXME: directed=1,2 for arcs one side to the other ...
*/
void
nautyextra_make_complete_bipartite (graph *g, int m, int n, int a, int b)
{
  assert (a >= 0);
  assert (b >= 0);
  assert (n >= 0);
  assert (n == a+b);
  assert (m >= SETWORDSNEEDED(n));

  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g,m,n);

  int from;
  for (from = 0; from < a; from++) {
    int to;
    for (to = 0; to < b; to++) {
      int to_plus_a = to + a;
      ADDONEEDGE(g, from,to_plus_a, m);
    }
  }
}

/* Set g to a complete graph of n vertices.
   Any existing contents in g are cleared.

   cf genspecialg.c static makecomplete() for sparsegraph

   FIXME: directed=1,2 for arcs one side to the other ...
*/
void
nautyextra_make_complete (graph *g, int m, int n, boolean directed)
{
  /* ENHANCE-ME: Fill setwords directly.
     ENHANCE-ME: directed=2 for reverse 0<-1.
   */
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (directed == 0 || directed == 1);

  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g,m,n);

  int u;
  for (u = 0; u < n; u++) {
    int v;
    for (v = directed ? u+1 : 0; v < n; v++) {
      if (u != v)
        ADDONEARC(g, u,v, m);
    }
  }
}

int
nautyextra_can_count (int n, int forest)
{
  assert (forest==0 || forest==1 || forest==-1);
  if (n==0) return 1;
  double d = pow(forest==1 ? n+1 : n, forest==-1 ? n-2 : n-1);
  int c = d;
  if (d!=c) INTERNAL_overflow();
  return c;
}

void
nautyextra_can_init (int *can, int n, int forest)
{
  DEBUG(printf("nautyextra_can_init() forest=%d\n", forest));
  assert (n >= 0);
  assert (forest==0 || forest==1 || forest==-1);
  if (n) {
    int canstart = (forest>0 ? 0 : 1);
    DEBUG(printf("  canstart %d\n", canstart));
    int i;
    for (i = 0; i < n-1; i++)
      can[i] = canstart;
    assert(i==n-1);
    can[i] = 0;
  }
}
int
nautyextra_can_next(int *can, int n, int forest)
{
  assert (n >= 0);
  assert (forest==0 || forest==1 || forest==-1);
  if (n) {
    int canstart = (forest>0 ? 0 : 1);
    int i = n - (forest == -1 ? 3 : 2);
    while (i >= 0) {
      assert (canstart <= can[i] && can[i] <= n);
      if (++can[i] <= n) return TRUE;
      can[i] = canstart;
      i--;
    }
  }
  return FALSE;
}


/* ------------------------------------------------------------------------- */
/* Tree Extensions and Modifications */

/* Extend vpar[1..n] to vpar[1..n*(s+1)] by attaching an extra path of s
   additional to each existing vertex.  For example s=1 attaches a
   single extra vertex to each.
   The existing vertices and their parents are unchanged.
   The new vertices are little paths in vpar[n+1..n*(s+1)] attached to
   each existing vertex.
*/
int
nautyextra_join_paths_vpar (int *vpar, int n, int s)
{
  assert (n >= 0);
  int v;
  int end = n;  /* last vertex number */
  for (v = 1; v <= n; v++) {
    int parent = v;
    int l;
    for (l = 0; l < s; l++) {
      vpar[++end] = parent;
      parent = end;
    }
  }
  assert(end == n*(s+1));
  return end;
}

int
nautyextra_join_vpar (int *vpar, int n, int v, int *sub_vpar, int sub_n)
{
  assert (n >= 0);
  assert (sub_n >= 0);
  int i;
  for (i=1; i<=sub_n; i++) {
    int sub_p = sub_vpar[i];
    assert (0 <= sub_p && sub_p <= sub_n);
    vpar[n+i] = sub_p ? sub_p+n : v;
  }
  return n + sub_n;
}

/* ------------------------------------------------------------------------- */
/* Star Replacement */

/* Set graph g_out to a star-replaced copy of g_in,m_in,n_in.
   Each vertex in g_in becomes a star of s many vertices.
   Existing edges in g_in become edges between leaf vertices of the stars
   in g_out.
   g_out must have space for s*n_in vertices.

   FIXME: Maybe how many new edges between arms.
*/
void
nautyextra_star_replacement(graph *g_in,  int m_in,  int n_in,
                            graph *g_out, int m_out,  int s)
{
  int n_out = s*n_in;
  assert((nauty_check(WORDSIZE, m_in, n_in,  NAUTYVERSIONID), 1));
  assert((nauty_check(WORDSIZE, m_out,n_out, NAUTYVERSIONID), 1));

  if (n_out == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g_out, m_out,n_out);

  /* output stars */
  {
    int from_in, i;
    for (from_in = 0; from_in < n_in; from_in++) {
      int from_out = from_in*s;
      for (i = 1; i < s; i++) {
        DEBUG(printf("  star edge %d to %d\n", from_out,from_out+i));
        ADDONEEDGE(g_out, from_out, from_out+i, m_out);
      }
    }
  }

  {
    int from_in, to_in;
    int *leaf_upto = ALLOCS(n_in, sizeof(int));
    if (leaf_upto == NULL) alloc_error("star_replacement()");
    memset((void*)leaf_upto, '\0', n_in*sizeof(int));

    for (from_in = 0; from_in < n_in; from_in++) {
      int from_out = s*from_in;
      set *row_in = GRAPHROW(g_in,from_in,m_in);
      for (to_in = from_in; (to_in = nextelement(row_in,m_in,to_in)) >= 0; ) {
        int from_leaf = ++leaf_upto[from_in];
        int to_leaf   = ++leaf_upto[to_in];
        if (from_leaf > s || to_leaf > s) {
          fprintf(stderr, "  star_replacement() oops, at from_in=%d to_in=%d degree too big, got %d,%d\n",
                  from_in,to_in, from_leaf,to_leaf);
          abort();
        }
        DEBUG(printf("  new edge %d to %d\n",
                     from_out + from_leaf,
                     s*to_in + to_leaf));
        ADDONEEDGE(g_out,  from_out + from_leaf, s*to_in + to_leaf, m_out);
      }
    }
    FREES(leaf_upto);
  }
}

/* Set sg_out to a star-replacement copy of sg_in.
   Each vertex in sg_in becomes a star of s many vertices in sg_out.
   Existing edges in sg_in become edges between leaf vertices of the stars
   in sg_out.

   sg_in must be undirected, so edges both u->v and v->u, and must not have
   any multi-edges (repeat of any u->v) or self-loops (u->u).

   sg_out must be already initialized per SG_INIT, but any existing contents
   are replaced.

   FIXME: Maybe how many new edges between arms.
*/
void
nautyextra_star_replacement_sg (sparsegraph *sg_in, sparsegraph *sg_out, int s)
{
  int nv_in   = sg_in->nv;
  int nde_in  = sg_in->nde;
  int nv_out  = s*nv_in;
  int nde_out = 2*(s-1)*nv_in + nde_in;
  int    *dd_in = sg_in->d, *dd_out;
  int    *ee_in = sg_in->e, *ee_out;
  size_t *vv_in = sg_in->v, *vv_out;
  int from_in;
  int epos = 0;
  int *leaf_upto;

  /* from ---> from_leaf ---> to_leaf ---> to
     .    <---           <---         <---
  */

  DEBUG(printf("star_replacement_sg() nv=%d nde=%d   %p %p\n",
               nv_in, nde_in, sg_in,sg_out));
  DEBUG(nautyextra_print_sg(sg_in));

  leaf_upto = ALLOCS(nv_in, sizeof(int));
  if (leaf_upto == NULL) alloc_error("star_replacement_sg()");
  memset((void*)leaf_upto, '\0', nv_in*sizeof(int));

  DEBUG(printf("  nv_out=%d nde_out=%d\n", nv_out, nde_out));
  SG_ALLOC(*sg_out, nv_out, nde_out, "star_replacement_sg()");
  sg_out->nv  = nv_out;
  sg_out->nde = nde_out;
  ee_out = sg_out->e;
  dd_out = sg_out->d;
  vv_out = sg_out->v;
  memset((void*)dd_out, '\0', nv_out*sizeof(int)); /* degrees 0 */

  /* DEBUG(printf("  OUT:\n")); */
  /* DEBUG(nautyextra_print_sg(sg_out)); */

  for (from_in = 0; from_in < nv_in; from_in++) {
    int from_out  = from_in*s;
    int degree_in = dd_in[from_in];
    int vpos_in   = vv_in[from_in];
    int i;
    DEBUG(printf("  from_in = %d degree %d\n", from_in, degree_in));

    if (degree_in > s) {
      fprintf(stderr, "star_replacement_sg() oops, degree>s at vertex %d\n",
              from_in);
      abort();
    }

    /* star middle to outer */
    vv_out[from_out] = epos;
    dd_out[from_out] = s-1;   /* middle out-degree */
    for (i = 1; i < s; i++) {
      DEBUG(printf("    middle to outer epos=%d\n", epos));
      DEBUG(if (epos >= nde_out) { abort(); });
      ee_out[epos++] = from_out + i;
    }

    if (degree_in > s) {
      fprintf(stderr, "star_replacement() oops, degree>s to vertex %d\n", from_in);
      abort();
    }
    for (i = 0; i < degree_in; i++) {
      int to_in  = ee_in[vpos_in + i];
      DEBUG(printf("    outer vertex edges epos=%d  %d -> %d\n", epos, from_in, to_in));

      if (to_in > from_in) {
        int to_out = s*to_in;
        int from_out_leaf_upto = ++leaf_upto[from_in];
        int from_out_leaf      = from_out + from_out_leaf_upto;
        int to_out_leaf_upto   = ++leaf_upto[to_in];
        int to_out_leaf        = to_out + to_out_leaf_upto;

        DEBUG(printf("      between %d and %d\n", from_out_leaf, to_out_leaf));
        vv_out[from_out_leaf] = epos;
        dd_out[from_out_leaf] = 2;
        DEBUG(if (epos >= nde_out) { abort(); });
        ee_out[epos++] = to_out_leaf;   /* across edge */
        ee_out[epos++] = from_out;      /* to middle */

        vv_out[to_out_leaf] = epos;
        dd_out[to_out_leaf] = 2;
        DEBUG(if (epos >= nde_out) { abort(); });
        ee_out[epos++] = from_out_leaf; /* across edge */
        ee_out[epos++] = to_out;        /* to middle */
      }
    }
  }
  for (from_in = 0; from_in < nv_in; from_in++) {
    int from_out  = from_in*s;
    for (;;) {
      int from_out_leaf_upto = ++leaf_upto[from_in];
      if (from_out_leaf_upto >= s) break;
      DEBUG(printf("  new leaf vertex back to middle, epos=%d\n", epos));
      int from_out_leaf    = from_out + from_out_leaf_upto;
      vv_out[from_out_leaf] = epos;
      dd_out[from_out_leaf] = 1;
      DEBUG(if (epos >= nde_out) { abort(); });
      ee_out[epos++] = from_out;
    }
  }

  FREES(leaf_upto);

  if (epos != nde_out) {
    fprintf(stderr, "star_replacement() oops, final epos=%d but nde_out=%d\n",
            epos, nde_out);
    abort();
  }
}

/* ------------------------------------------------------------------------- */

/* g is a directed graph.
   Add arcs to make it undirected.
   Any arc u-->v gets an arc back the other way v-->u.
   If g is already undirected then this is no change.
*/
void
nautyextra_to_undirected (graph *g, int m, int n)
{
  set *from_row = (set*) g;
  int from;
  for (from = 0; from < n; from++, from_row += m) {
    int to;
    for (to = -1; (to = nextelement(from_row,m,to)) >= 0; ) {
      ADDONEARC(g, to,from, m);
    }
  }
}

/* ------------------------------------------------------------------------- */

/* Set *up,*vp to the vertices of arc number "arc" in g,m,n.
   arc=0 is the first arc.  Return TRUE and set *up,*vp if arc exists.

   If arc exceeds the number of arcs, including if g has no arcs at all,
   then return FALSE and set *up = *vp = -1.

   Arcs u->v are in order first by ascending u and among equal u by
   ascending v.  Arc numbers range 0 <= arc < nautyextra_num_arcs(g,m,n).
   Self-loop arcs are included.
*/
int
nautyextra_nth_arc (const graph *g, int m, int n, int arc, int *up, int *vp)
{
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (arc >= 0);
  assert (up != NULL);
  assert (vp != NULL);

  const setword *p = g;
  int u, i;
  for (u = 0; u < n; u++) {
    for (i = 0; i < m; i++) {
      setword w = *p++;
      int c = POPCOUNT(w);
      int new_arc = arc - c;
      if (new_arc < 0) {
        *up = u;
        *vp = nautyextra_set_nthelement1(w, arc);
        return 1;
      }
      arc = new_arc;
    }
  }
  *up = -1;
  *vp = -1;
  return 0;
}

/* Set *up,*vp to the vertices of a uniformly distributed random arc of g,m,n.
   If g has no arcs then set *up = *vp = -1.
*/
void
nautyextra_random_arc (const graph *g, int m, int n, int *up, int *vp)
{
  /* cumul_arcs[u] is the number of preceding arcs, ie. how many 0..u-1 ->
     something.

     Could record cumulative count for each setword of g using more
     workspace.
  */
  DEBUG(printf("nautyextra_random_arc()\n"));
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (up != NULL);
  assert (vp != NULL);
  int mshrunk = SETWORDSNEEDED(n);
  int cumul_size = n*mshrunk;
#if MAXN
  int cumul_arcs[SETWORDSNEEDED(MAXN)*MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE(int,cumul_arcs,cumul_size,"nautyextra_random_arc");
#endif

  int mextra = m - mshrunk;
  int total_arcs = 0;
  {
    int *pc = cumul_arcs;
    const setword *row = g;
    assert (mextra >= 0);
    int u;
    for (u = 0; u < n; u++,row+=mextra) {
      int i;
      for (i = 0; i < mshrunk; i++) {
        *pc++ = total_arcs;
        setword w = *row++;
        total_arcs += POPCOUNT(w);
      }
    }
  }

  DEBUG(printf("  total_arcs %d\n", total_arcs));
  if (total_arcs == 0) {
    *up = *vp = -1;
    return;
  }
  int arc;
  GETKRAN(total_arcs,arc);
  DEBUG(printf("  arc %d\n", arc));
  assert (0 <= arc && arc < total_arcs);

  /* Bsearch cumul_arcs for biggest u with cumul_arcs[u] <= arc.
     u is lo vertex number, so cumul_arcs[u] <= arc.
     hi is a vertex number with cumul_arcs[u] > arc, or hi=n, so hi is where
     definitely too big.
  */
  assert (n >= 1);
  int lo = 0;
  int hi = cumul_size;
  for (;;) {
    assert (0 <= lo && lo < cumul_size);
    assert (0 <= hi && hi <= cumul_size);
    assert (cumul_arcs[lo] <= arc);
    assert (hi==cumul_size || arc < cumul_arcs[hi]);

    int mid = (hi + lo) >> 1;   /* rounded up */
    if (mid == lo) break;
    if (cumul_arcs[mid] <= arc) lo = mid;
    else                        hi = mid;
  }
  assert (cumul_arcs[lo] <= arc);
  assert (lo==n-1 || arc < cumul_arcs[lo+1]);

  int u = lo / mshrunk;
  int r = lo % mshrunk;
  *up = u;
  int v = r*WORDSIZE;

  arc -= cumul_arcs[lo];
  DEBUG(printf("u=%d r=%d v=%d   arc %d\n", u, r, v, arc));

  setword w = ((setword*) g)[lo + u*mextra];
  v += nautyextra_set_nthelement1(w, arc);
  *vp = v;

  assert (0 <= *up && *up < n);
  assert (0 <= *vp && *vp < n);
}

/* Set *up,*vp to the vertices of a uniformly distributed random arc of sg.
   If sg has no arcs then set *up = *vp = -1.
*/
void
nautyextra_random_arc_sg (const sparsegraph *sg, int *up, int *vp)
{
  DEBUG(printf("nautyextra_random_arc_sg()\n"));
  const int *dd = sg->d;
  int n = sg->nv;
  int nde = sg->nde;
  if (nde == 0) {
    *up = *vp = -1;
    return;
  }

  int arc;
  GETKRAN(nde,arc);
  DEBUG(printf("  arc %d\n", arc));
  assert (0 <= arc && arc < nde);

  int u;
  for (u = 0; u < n; u++) {
    DEBUG(printf("u=%d arc %d\n", u, arc));
    assert (arc >= 0);

    int new_arc = arc - dd[u];
    if (new_arc < 0) {
      *up = u;
      *vp = sg->e[sg->v[u] + arc];
      assert (0 <= *up && *up < n);
      assert (0 <= *vp && *vp < n);
      return;
    }
    arc = new_arc;
  }

  fprintf(stderr, "nautyextra_random_arc_sg() oops, end with arc still %d\n",
          arc);
  abort();
}

/* Set dst[0..m-1] to set union of dst[0..m-1] and src[0..m-1].
 */
void
nautyextra_set_union(set *dst, const set *src, int m)
{
  int i;
  for (i = 0; i < m; i++)
    UNION(dst[i], src[i]);
}

/* g is a graph with an edge u-->v.
   Return TRUE if u-->v is a cut edge, meaning that removing it would
   separate those two vertices, ie. the only path between u and v is the
   edge between them.  Return FALSE if not, so there is some other path.

   Not sure if works properly for directed graph.  For a directed graph the
   idea roughly would be whether arc u-->v is the only directed path u-->v.
*/
int
nautyextra_is_cut_edge (const graph *g, int m, int n, int u, int v)
{
  DEBUG(printf("nautyextra_is_cut_edge() %d to %d\n", u, v));
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (0 <= u && u < n);
  assert (0 <= v && v < n);
  assert (u != v);

  /* FIXME: or maybe allow no actual u->v existing edge, which would be to
     ask whether u,v in same connected component */
  assert (ISELEMENT(GRAPHROW(g,u,m), v));

#if MAXN
  int queue[MAXN];
  set seen[SETWORDSNEEDED(MAXN)];
  assert(n <= MAXN);
#else
  WORKSPACE2(int,queue,2*n,
             set,seen, m,    "nautyextra_is_cut_edge");
#endif

  memset (seen,'\0',m*sizeof(seen[0]));
  ADDELEMENT(seen,v);

  int queue_head = 0;
  queue[queue_head++] = u;
  ADDELEMENT(seen,u);

  int queue_tail = 0;
  while (queue_tail < queue_head) {
    int i = queue[queue_tail++];
    set *row = GRAPHROW(g,i,m);
    DEBUG(printf("i=%d row %s\n", i, nautyextra_set_str(row,n)));
    int to;
    for (to = -1; (to = nextelement(row,m,to)) >= 0; ) {
      DEBUG(printf("  to %d\n", to));
      if (to == v) {
        if (i != u) {
          DEBUG(printf("  another way to v=%d, not a cut edge\n", v));
          return FALSE;
        }
        DEBUG(printf("    is v but from u\n"));
      } else {
        if (ISELEMENT(seen,to)) {
          DEBUG(printf("    seen\n"));
        } else {
          DEBUG(printf("    add to queue\n"));
          assert (queue_head < n);
          queue[queue_head++] = to;
          ADDELEMENT(seen,to);
        }
      }
    }
  }
  DEBUG(printf("no other way to v, is a cut edge\n"));
  return TRUE;
}

/* g is an undirected graph.
   Return the number of cut edges (per nautyextra_is_cut_edge()).

   Currently this is a naive search over all edges, so suitable only for
   small graphs.
*/
int
nautyextra_cut_edge_count (const graph *g, int m, int n)
{
  int count = 0;
  int from;
  for (from = 0; from < n; from++) {
    const set *from_row = GRAPHROW(g,from,m);
    int to;
    for (to = from-1; (to = nextelement((set*)from_row,m,to)) >= 0; ) {
      count += nautyextra_is_cut_edge(g,m,n, from,to);
    }
  }
  return count;
}


/* ------------------------------------------------------------------------- */

/* Set g_out,m_out,n_out to the line graph of g_in,m_in,m_out.
   g_in must be undirected (arcs both directions).
   n_out is the number of vertices in the line graph output.  This must be
   the number of edges in g_in.
   g_out must have enough space for its n_out but any existing contents are
   cleared.

   cf lingraphg.c linegraph() for sparsegraph.
*/
void
nautyextra_linegraph (const graph *g_in, int m_in, int n_in,
                      graph *g_out, int m_out, int n_out)
{
  DEBUG(printf("nautyextra_linegraph() m=%d,n=%d out m=%d,n=%d\n",
               m_in,n_in, m_out,n_out));
  assert(n_in >= 0);
  assert(m_in >= SETWORDSNEEDED(n_in));
  assert(n_out >= 0);

  if (n_in == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */

  struct data {
    int u,v, out;
  };
#if MAXN
  struct data data[2*MAXN];
  assert(n_out <= MAXN);
#else
  WORKSPACE(struct data,data,2*n_out,"nautyextra_linegraph");
#endif

  int datalen = 0;
  int out_v = 0;
  int in_from;
  for (in_from = 0; in_from < n_in; in_from++) {
    set *in_from_row = GRAPHROW(g_in,in_from,m_in);
    int in_to;
    for (in_to = -1;
         (in_to = nextelement(in_from_row,m_in,in_to)) >= 0; ) {
      assert (in_from != in_to);
      assert (0 <= in_to && in_to < n_in);

      int this_out_v;
      if (in_from < in_to) {
        assert(out_v < n_out);
        this_out_v = out_v++;
      } else {
        /* bsearch in data[0 .. datalen-1] */
        assert (datalen >= 1);
        int lo = 0;
        int hi = datalen-1;
        while (lo != hi) {
          assert (0 <= lo && lo < hi && hi < datalen);
          int mid = (lo + hi) >> 1;
          if (data[mid].u < in_to) {
          LO: lo = mid+1;
          } else if (data[mid].u > in_to) {
          HI: hi = mid-1;
          } else if (data[mid].v < in_from) {   /* here u==in_to */
            goto LO;
          } else if (data[mid].v > in_from) {
            goto HI;
          } else {
            lo = mid;
            break;
          }
        }
        assert (data[lo].u == in_to);
        assert (data[lo].v == in_from);
        this_out_v = data[lo].out;
      }
      DEBUG(printf("  edge %d--%d out %d\n", in_from, in_to, this_out_v));
      assert (datalen < 2*n_out);
      data[datalen].u   = in_from;
      data[datalen].v   = in_to;
      data[datalen].out = this_out_v;
      datalen++;
    }
  }
  assert(out_v == n_out);
  assert(datalen == 2*n_out);

  EMPTYGRAPH(g_out, m_out,n_out);

  DEBUG({ int i;
          for (i = 0; i < datalen; i++)
            printf("data[%d] = %d,%d out %d\n",
                   i, data[i].u, data[i].v, data[i].out);
        });

  int i;
  for (i = 0; i < datalen-1; i++) {
    DEBUG(printf("out i=%d  %d-%d out=%d\n",
                 i, data[i].u, data[i].v, data[i].out));
    int out_from = data[i].out;
    int in_from  = data[i].u;
    int j;
    for (j = i + 1; j < datalen && data[j].u == in_from; j++) {
      int out_to = data[j].out;
      DEBUG(printf("  out j=%d is %d-%d out=%d   edge %d--%d\n",
                   j, data[j].u,data[j].v,data[j].out,
                   out_from, out_to));
      assert (0 <= out_from && out_from < n_out);
      assert (0 <= out_to   && out_to   < n_out);
      assert (out_from != out_to);
      ADDONEEDGE(g_out, out_from,out_to, m_out);
    }
  }
}

/* Return TRUE if g is a line graph, or FALSE if not.

   Every vertex must belong to at most 2 cliques.

   NOT WORKING
 */
int
nautyextra_is_linegraph1 (const graph *g, int n)
{
  /* NOT WORKING */

  DEBUG(printf("nautyextra_is_linegraph1() n=%d\n", n));
  abort();

  /*   struct data { */
  /*     setword clique1; */
  /*     setword clique2; */
  /*   }; */
  /* #if MAXN */
  /*   struct data  data[MAXN+1];  /\* 1 to n inclusive *\/ */
  /*   assert (n <= MAXN); */
  /* #else */
  /*   WORKSPACE(struct data,data,n,"nautyextra_is_linegraph1"); */
  /* #endif */

  int v;
  for (v = 0; v < n; v++) {
    DEBUG(printf("v=%d\n", v));
    setword v_bit = bit[v];
    setword clique1 = v_bit;
    setword clique2 = v_bit;
    int to;
    setword v_row = g[v];
    for (to = -1; (to = nextelement((set*)&v_row,1,to)) >= 0; ) {
      setword to_bit = bit[to];
      setword not_to_row = ~(g[to] | to_bit);
      if (! (clique1 & not_to_row)) {
        DEBUG(printf("  to=%d clique 1\n", to));
        clique1 |= to_bit;

        if (clique2 != v_bit && ! (clique2 & not_to_row)) {
          DEBUG(printf("  cliques with >1 vertices in common\n"));
          return FALSE;
        }
      } else if (! (clique2 & not_to_row)) {
        DEBUG(printf("  to=%d clique 2\n", to));
        clique2 |= to_bit;
      } else {
        DEBUG(printf("  to=%d another clique\n", to));
        return FALSE;  /* a third clique at v */
      }
    }
  }
  return TRUE;
}

/* nautyextra_is_linegraph() returns TRUE if g,m,n is a line graph or FALSE
   if not.

   The current code uses the conditions of

       Wilf and van Rooij, "The Interchange Graph of a Finite Graph",
       Acta Mathematica Academiae Scientiarum Hungarica, volume 16, issue 3,
       September 1965, pages 263-269

   g is a line graph iff it is claw-free and every induced diamond subgraph
   has at least one of its two triangles even - ie. any diamond with both
   odd means not a line graph.
*/
static int
INTERNAL_is_linegraph_callback (const graph *g, int m, int n,
                       int a, int b, int c, int d, void *data)
{
  /* return 1 to stop if found both odd (which means not line graph),
     0 to continue otherwise
  */
  return nautyextra_diamond_is_both_odd(g,m, a,b,c,d);
}
int
nautyextra_is_linegraph (const graph *g, int m, int n)
{
  return nautyextra_is_clawfree (g,m,n)
    && ! nautyextra_diamond_search (g,m,n, 1,
                                    &INTERNAL_is_linegraph_callback, NULL);

#if 0
  /*
    .   c
    .  / ?     claw b < c < d
    . a---b    diamond a < b and c < d
    .  \ ?
    .   d
  */
  int a;
  set *a_row = (set*) g;
  for (a = 0; a < n-1; a++, a_row += m) {                  /* a all */
    DEBUG(printf(" a=%d\n", a));

    int b;
    for (b = a; (b = nextelement(a_row,m,b)) >= 0; ) {     /* a < b */
      DEBUG(printf("  b=%d\n", b));
      assert (a < b);
      set *b_row = GRAPHROW(g,b,m);

      int c;
      for (c = -1; (c = nextelement(a_row,m,c)) >= 0; ) {  /* c any, c!=b */
        assert (c != a); /* no self-loop */
        if (c == b) continue;
        assert (c != a); /* no self-loop */
        DEBUG(printf("   c=%d\n", c));
        if (! ISELEMENT(b_row,c)) continue;  /* no b--c edge, not diamond */
        set *c_row = GRAPHROW(g,c,m);

        int d;
        for (d = c; (d = nextelement(a_row,m,d)) >= 0; ) { /* c<d, d!=b */
          assert (d != a); /* no self-loop */
          if (d == b) continue;
          DEBUG(printf("    d=%d\n", d));
          assert (c < d);

          if (! ISELEMENT(b_row,d)) continue;  /* no b--d edge, not diamond */
          if (induced
              && ISELEMENT(c_row,d)) continue; /* have c--d edge, not induced */

          DEBUG(printf("     found %d,%d,%d,%d\n", a,b,c,d));
          int ret = (*callback)(g,m,n, a,b,c,d, data);
          if (ret) return ret;
        }
      }
    }
  }
  return 0;
#endif
}

/* ------------------------------------------------------------------------- */

/* Return true if sub_g,sub_m,sub_n is a subgraph of g,m,n.
   This means some mapping of sub_n vertices to a subset of g which has all
   the sub_g edges present also in g.

   induced=0 allows additional edges between those which sub_g maps to in g.
   induced=1 demands an induced subgraph, so that among the vertices mapped
   in g there are no edges other than the ones corresponding to sub_g.

   This code is a fairly poor exhaustive search, though pruning the search
   when degrees of sub_g and g can't match at some point.
*/
int
nautyextra_is_subgraph (const graph *sub_g, int sub_m, int sub_n,
                        const graph *g, int m, int n, int induced)
{
  DEBUG(printf("nautyextra_is_subgraph()\n"));
  if (sub_n == 0) return TRUE;
  if (sub_n > n) return FALSE;

#if MAXN
  int  map[MAXN];
  int  sub_degrees[MAXN];
  int  degrees[MAXN];
  set  used[SETWORDSNEEDED(MAXN)];
  assert (n <= MAXN);
#else
  WORKSPACE4(int,sub_degrees,sub_n,
             int,degrees,n,
             int,map,sub_n,
             set,used,SETWORDSNEEDED(n),
             "nautyextra_is_subgraph");
#endif

  nautyextra_degrees_array(sub_g,sub_m,sub_n, sub_degrees);
  nautyextra_degrees_array(g,m,n, degrees);
  EMPTYSET(used,m);
  int pos = 0;
  map[pos] = -1;
  for (;;) {
    int v;
  NEXT_OUTER:
    DEBUG(printf("at pos=%d\n", pos));
    v = map[pos];
    if (v >= 0)
      DELELEMENT(used,map[pos]);
    for (;;) {
      v = ++map[pos];
      DEBUG(printf("  step map[%d] to v=%d  used %d\n",
                   pos, v, ISELEMENT(used,v)));
      if (v >= n) {
        pos--;
        DEBUG(printf("backtrack to pos=%d\n", pos));
        if (pos < 0) return FALSE;
        goto NEXT_OUTER;
      }
      if (! ISELEMENT(used,v)) {
        ADDELEMENT(used,v);
        break;
      }
    }

    DEBUG(printf("incremented pos=%d\n", pos));

    if (degrees[v] < sub_degrees[pos]) {
      DEBUG(printf("  graph degree %d < subgraph degree %d\n",
                   degrees[map[pos]], sub_degrees[pos]));
      continue;
    }
    int p;
    for (p = 0; p < pos; p++) {
      /* ### consider: "pos=$pos p=$p graph $map[$p] to $map[$pos]  subgraph $p to $pos" */
      int sub_has_edge = ISELEMENT(GRAPHROW(sub_g,p,sub_m), pos);
      int has_edge = ISELEMENT(GRAPHROW(g,v,m), map[p]);
      if (induced
          ? (sub_has_edge != has_edge)
          : (sub_has_edge > has_edge)) goto NEXT_OUTER;
    }

    DEBUG(printf("  good, descend\n"));
    if (++pos >= sub_n) {
      return TRUE;
    }
    map[pos] = -1;
  }
  return FALSE;
}


/* ------------------------------------------------------------------------- */

/* Set g,m,n to the generalized Petersen graph N,K where N=n/2.
   n must be even.
   g must have enough space but any existing content is cleared.
   The usual Petersen graph is N=5,K=2 which here is n=10, k=2.
*/
void
nautyextra_make_Petersen (graph *g, int m, int n, int k)
{
  DEBUG(printf("nautyextra_make_Petersen() m=%d,n=%d\n", m,n));
  assert (n >= 0);
  assert ((n & 1) == 0);
  assert (m >= SETWORDSNEEDED(n));

  EMPTYGRAPH(g,m,n);
  int N = n>>1;
  int i;
  for (i = 0; i < N; i++) {
    int from = i;
    int to = (i == 0 ? N-1 : i-1);
    DEBUG(printf("  outer around %d -- %d\n", from,to));
    ADDONEEDGE(g, from,to, m);

    to = from + N;
    DEBUG(printf("  outer %d -- inner %d\n", from,to));
    ADDONEEDGE(g, from,to, m);

    from = N + i;
    to = N + ((i + k) % N);
    DEBUG(printf("  inner around %d -- %d\n", from,to));
    ADDONEEDGE(g, from,to, m);
  }
}

/* Set g,m,n to Mobius ladder of n vertices.
   n must be even.
   g must have enough space but any existing content is cleared.

   This is a cyclic ladder

       ...--*---*---*--...
            |   |   |
       ...--*---*---*--...

   with the ends wrapping and a twist so the upper path wraps around to the
   lower and vice versa.
*/
void
nautyextra_make_Mobius_ladder (graph *g, int m, int n)
{
  DEBUG(printf("nautyextra_make_Mobius_ladder() m=%d,n=%d\n", m,n));
  assert (n >= 0);
  assert ((n & 1) == 0);
  assert (m >= SETWORDSNEEDED(n));

  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g,m,n);

  int N = n>>1;
  int i;
  for (i = 0; i < N-1; i++) {
    ADDONEEDGE(g, i,   i+1,   m);
    ADDONEEDGE(g, i,   i+N,   m);  /* across */
    ADDONEEDGE(g, i+N, i+N+1, m);
  }
  ADDONEEDGE(g, 0,   2*N-1, m);  /* twist */
  ADDONEEDGE(g, N-1, 2*N-1, m);  /* across */
  ADDONEEDGE(g, N,   N-1,   m);
}

/* Set g,m,n to the connected graph of n>=6 vertices with most independent
   dominating sets.
   g must have enough space for m,n.  Any existing content is cleared.

   For n>=6 there is a single unique such graph.
   For n<=5 the graph created is unspecified.

   For n>=6 the sets of the relevant graph are all size the independence
   number.  So this graph is also the graph which the most maximum
   independent sets.

   The graph created is per

       J.R. Griggs, C.M. Grinstead and D.R. Guichard, "The Number of Maximal
       Independent Sets in a Connected Graph", Discrete Mathemathics, volume
       68, 1988, pages 211-220.

       Min-Jen Jou and Gerard J. Chang, "The Number of Maximum Independent
       Sets in Graphs", Taiwanese Journal of Mathematics, volume 4, number
       4, pages 685-695, December 2000.

   It is certain triangles or complete-4 connected by edges.
*/
void
nautyextra_make_most_indomsets (graph *g, int m, int n)
{
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));

  if (n == 0) return;    /* EMPTYGRAPH() doesn't take n=0 */
  EMPTYGRAPH(g,m,n);

  int v = 0;
  while (n > 0) {
    if (v)
      ADDONEEDGE(g,0,v,m);  /* to x */

    int u = v;
    int i;
    for (i = 0; i < 3 + ((n%3)!=0); i++) {   /* triangle or complete-4 */
      int j;
      for (j = 0; j < i; j++)
        ADDONEEDGE(g, u+i, u+j, m);
    }
    n -= i;
    v += i;
  }
}

/* ------------------------------------------------------------------------- */
/* Well Covered */

/* Return 1 if graph g,m,n is well-covered.
   This means all its independent dominating sets (nautyextra_is_indomset_sg())
   are the same size.
*/
int
nautyextra_is_well_covered_sg (const sparsegraph *sg)
{
  setword ss;
  int n = sg->nv;
  assert (n >= 0);
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_is_well_covered_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int want_size = -1;
  NAUTYEXTRA_SET1_LOOP (ss,n,
                        if (nautyextra_is_indomset_sg (sg,&ss)) {
                          int ss_size = POPCOUNT(ss);
                          if (want_size >= 0  && ss_size != want_size) return 0;
                          want_size = ss_size;
                        });
  return 1;
}


/* ------------------------------------------------------------------------- */

/* ss is a set of vertices in an m=1 word graph g,n.
   Return TRUE if ss is a clique in g, meaning every vertex in ss has an
   edge to every other vertex in ss.
   g must be undirected.  Any self-loops are ignored.
   An empty set or single vertex ss is always a clique.
*/
int
nautyextra_is_clique1 (const graph *g, int n, setword ss)
{
  DEBUG(printf("nautyextra_is_clique1() n=%d ss=%s\n",
               n, nautyextra_set_str(&ss,n)));
  assert (0 <= n && n <= WORDSIZE);
  assert ((ss & BITMASK(n)) == 0);
  int v;
  setword v_mask = bit[0];  /* 1-bit at position v */
  for (v = 0; v < n; v++, v_mask >>= 1) {
    setword gg = *g++;
    if ((v_mask & ss) && ((gg & ss) | v_mask) != ss) {
      DEBUG(printf("  non-clique at v=%d row=%s\n",
                   v, nautyextra_set_str(&gg,n)));
      return FALSE;
    }
  }
  return TRUE;
}

/* ss is a set of vertices in graph g,m,n, formed with m many setwords.
   Return TRUE if ss is a clique in g, meaning every vertex in ss has an
   edge to every other vertex in ss.
   g must be undirected and have no self-loops.
   An empty set or single vertex ss is always a clique.
*/
int
nautyextra_is_clique (const graph *g, int m, int n, const set *ss)
{
  if (m==1) return nautyextra_is_clique1(g,n,ss[0]);

  /* ENHANCE-ME: A self-loop at v counts as an edge in setinter, tricking
     the count to be == target_setinter when there is a missing edge
     elsewhere.  Would like to ignore self-loop, set_is_subset_except_v,
     or subtract ISELEMENT(v_row,v).
  */
  int target_setinter = setsize((set*)ss,m) - 1;
  int v;
  for (v = -1; (v = nextelement((set*)ss,m,v)) >= 0; ) {
    const set *v_row = GRAPHROW(g,v,m);
    if (setinter((set*)v_row, (set*)ss, m) < target_setinter) {
      return FALSE;
    }
  }
  return TRUE;
}

/* Return the clique number of graph g,m,n.
   This is the number of vertices in the largest clique in g (largest
   complete subgraph).
   The code is a very poor exhaustive search.  cliquer program much better.
*/
int
nautyextra_clique_number (const graph *g, int m, int n)
{
  setword ss;
  assert (n >= 0);
  if (n == 0) return 1;   /* n=0 empty set is only indset */
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_clique_number() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int clique_number = 0;
  NAUTYEXTRA_SET1_LOOP (ss,n,
                        if (nautyextra_is_clique (g,m,n,&ss)) {
                          int ss_size = POPCOUNT(ss);
                          clique_number = MAX(clique_number, ss_size);
                        });
  return clique_number;
}

/* Set *info to some information about similar vertices in g,m,n.
   Must have m=1.

   Vertices u,v are similar when in the same orbit of the automorphism group
   of g, meaning there is some automorphism mapping u->v (and vice versa).

   info->similar is a matrix of flags, in the form of an undirected graph,
   for where pairs u,v are similar.  An edge u,v means u,v are similar.
   This is simply a set of cliques for the orbits of the automorphism group.

   Vertices u,v are removal similar if deleting either u or v gives the same
   remaining graph (ie. isomorphic remainder).

   info->removal_similar is a matrix of flags, in the form of an undirected
   graph, for where pairs u,v are removal similar.  An edge u,v means u,v
   are removal similar.

   Pseudosimilar vertices are those which are removal similar but not
   similar.  Any similar u,v are certainly removal similar, but the converse
   is not always so.  So removal similar = similar union pseudosimilar.

   info->pseudosimilar is a matrix of flags, in the form of an undirected
   graph, for where pairs u,v are pseudosimilar.
*/
void
nautyextra_similar (graph *g, int m, int n, struct nautyextra_similar_info *info)
{
  DEBUG (printf("nautyextra_similar() n=%d\n", n));
  DEBUG (nautyextra_print_matrix(g,m,n));

  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (m == 1);

  static DEFAULTOPTIONS_DENSEGRAPH(options);
  options.getcanon = TRUE;
  statsblk stats;

  DYNALLSTAT(graph,canon_g,canon_g_sz);
  DYNALLSTAT(int, lab,    lab_sz);
  DYNALLSTAT(int, ptn,    ptn_sz);
  DYNALLSTAT(int, orbits, orbits_sz);
  DYNALLSTAT(setword, dels,   dels_sz);

  int del_n = n-1;
  DYNALLOC2(graph,canon_g,canon_g_sz,m,n,"nautyextra_similar");
  DYNALLOC1(int, lab,    lab_sz,    n, "nautyextra_similar");
  DYNALLOC1(int, ptn,    ptn_sz,    n, "nautyextra_similar");
  DYNALLOC1(int, orbits, orbits_sz, n, "nautyextra_similar");
  DYNALLOC1(setword, dels, dels_sz, n*del_n, "nautyextra_similar");

  int v;
  for (v = 0; v < n; v++) {
    delete1(g,canon_g, v,n);
    DEBUG (printf("delete %d to del_n=%d at %p\n", v, del_n, canon_g));
    DEBUG (nautyextra_print_matrix(canon_g, m,del_n));

    graph *canon_del = dels + v*del_n;
    densenauty(canon_g, lab,ptn,orbits,&options,&stats,
               m, del_n, canon_del);
    DEBUG (printf("canonical at %p (%d orbits)\n", canon_del,
                  stats.numorbits));
    DEBUG (nautyextra_print_matrix(canon_del,m,del_n));
    DEBUG (printf("\n"));
  }

  densenauty(g, lab,ptn,orbits,&options,&stats,m,n,canon_g);
  info->numorbits = stats.numorbits;

  DYNALLSTAT(int, orbit_size, orbit_size_sz);
  DYNALLOC1(int, orbit_size, orbit_size_sz, n, "nautyextra_similar");
  memset (orbit_size,'\0',n*sizeof(orbit_size[0]));
  for (v = 0; v < n; v++) {
    orbit_size[orbits[v]]++;
    DEBUG(printf("v=%d orbit rep %d inc to %d\n",
                 v, orbits[v], orbit_size[orbits[v]]));
  }
  int orbit_max = 0;
  for (v = 0; v < n; v++) {
    orbit_max = MAX(orbit_max, orbit_size[v]);
    DEBUG(printf("orbit v=%d size %d\n", v, orbit_size[v]));
  }
  info->orbit_max = orbit_max;

  DYNALLSTAT(graph,similar,similar_sz);
  DYNALLSTAT(graph,removal_similar,removal_similar_sz);
  DYNALLSTAT(graph,pseudosimilar,pseudosimilar_sz);
  DYNALLOC2(graph,similar,similar_sz,m,n,"nautyextra_similar");
  DYNALLOC2(graph,removal_similar,removal_similar_sz,m,n,"nautyextra_similar");
  DYNALLOC2(graph,pseudosimilar,pseudosimilar_sz,m,n,"nautyextra_similar");

  if (n > 0) {;    /* EMPTYGRAPH() doesn't take n=0 */
    EMPTYGRAPH(similar,m,n);
    EMPTYGRAPH(pseudosimilar,m,n);
    EMPTYGRAPH(removal_similar,m,n);

    int u;
    for (u = 0; u < n; u++) {
      for (v = u+1; v < n; v++) {
        int uv_removal_similar = nautyextra_aresame(dels+u*del_n,
                                                    dels+v*del_n, m, del_n);
        int uv_similar = (orbits[u] == orbits[v]);
        int uv_pseudosimilar = (uv_removal_similar && !uv_similar);

        /* any similar is also removal similar */
        assert(uv_removal_similar >= uv_similar);

        if (uv_removal_similar)
          ADDONEEDGE(removal_similar, u,v, m);
        if (uv_similar)
          ADDONEEDGE(similar, u,v, m);
        if (uv_pseudosimilar)
          ADDONEEDGE(pseudosimilar, u,v, m);
      }
    }
  }

  info->similar = similar;
  info->pseudosimilar = pseudosimilar;
  info->removal_similar = removal_similar;
  info->pseudosimilar_clique_size = nautyextra_clique_number(pseudosimilar,m,n);
}



/* ------------------------------------------------------------------------- */


void
nautyextra_find_dist_counts(const graph *g, int m, int n, int v,
                            const int *dist, unsigned long long *count)
{
  DEBUG(printf("nautyextra_find_dist_counts() n=%d v=%d\n", n, v));
  assert (n >= 0);
  assert (0 <= v && v < n);
#if MAXN
  int queue[MAXN];
  set seen[SETWORDSNEEDED(MAXN)];
  assert(n <= MAXN);
#else
  WORKSPACE2(int,queue,n,
             set,seen,m, "nautyextra_find_dist_counts");
#endif

  memset (count,'\0',n*sizeof(count[0]));
  count[v] = 1;

  memset (seen,'\0',m*sizeof(seen[0]));
  ADDELEMENT(seen,v);

  int queue_head = 0;
  queue[queue_head++] = v;
  ADDELEMENT(seen,v);

  int queue_tail = 0;
  while (queue_tail < queue_head) {
    int from = queue[queue_tail++];
    int target_dist = dist[from] + 1;
    set *row = GRAPHROW(g,from,m);
    DEBUG(printf("from=%d row %s\n", from, nautyextra_set_str(row,n)));
    int to;
    for (to = -1; (to = nextelement(row,m,to)) >= 0; ) {
      if (dist[to] == target_dist) {
        DEBUG(printf("  to %d  new count %Lu\n", to, count[to]+count[from]));
        ADD_NO_OVERFLOW_ULL(count[to], count[to], count[from]);
        if (! ISELEMENT(seen,to)) {
          queue[queue_head++] = to;
          ADDELEMENT(seen,to);
        }
      } else {
        DEBUG(printf("  skip %d\n", to));
      }
    }
  }
}

/* ------------------------------------------------------------------------- */

/* Set distances[0..n-1][0..n-1] so distances[u][v] = distance between u and v,
   by the shortest path between.
   distances[] must have space for n*n ints.
*/
void
nautyextra_distances_matrix (const graph *g, int m, int n, int *distances)
{
  assert(n >= 0);
  assert(m >= SETWORDSNEEDED(n));
  int from;
  for (from = 0; from < n; from++) {
    find_dist((graph*)g, m,n, from, distances);
    distances += n;
  }
}

int
nautyextra_is_distance_regular (const graph *g, int m, int n,
                                int *intersection_array)
{
  DEBUG(printf("nautyextra_is_distance_regular() n=%d\n", n));
  assert(n >= 0);
  assert(m >= SETWORDSNEEDED(n));
#if MAXN
  int distances[MAXN*MAXN];
  int intersection_array_workspace[2*MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE2(int,distances,n*n,
             int,intersection_array_workspace, intersection_array ? 0 : 2*n,
             "nautyextra_is_distance_regular");
#endif
  if (! intersection_array)
    intersection_array = intersection_array_workspace;

  nautyextra_distances_matrix (g,m,n, distances);

  int from;
  for (from = 0; from < 2*n; from++)
    intersection_array[from] = -1;

  int diameter = 0;
  int *from_distances = distances;
  for (from = 0; from < n; from++,from_distances+=n) {
    DEBUG(printf(" from=%d\n", from));

    int to;
    for (to = from; to < n; to++) {
      set *to_row = GRAPHROW(g,to,m);
      int to_distance = from_distances[to];
      DEBUG(printf("  to=%d row %p\n", to, to_row));
      assert (0 <= to_distance && to_distance < n);
      diameter = MAX(diameter,to_distance);

      int neighbour;
      int b = 0, c = 0;
      for (neighbour = -1; (neighbour = nextelement(to_row,m,neighbour)) >= 0;){
        int neighbour_distance = from_distances[neighbour];
        int diff = neighbour_distance - to_distance;
        assert (-1 <= diff && diff <= 1);
        b += (diff == 1);  /* neighbour 1 further away */
        c += (diff == -1); /* neighbour 1 closer */
      }
      DEBUG(printf("   dist %d got b=%d c=%d  and previously %d %d\n",
                   to_distance, b, c,
                   intersection_array[to_distance],
                   intersection_array[to_distance+n]));

      if (intersection_array[to_distance] == -1) {
        intersection_array[to_distance] = b;
        intersection_array[to_distance + n] = c;
      } else {
        if (intersection_array[to_distance] != b
            || intersection_array[to_distance + n] != c) {
          return FALSE;
        }
      }
    }
  }
  return TRUE;
}

/* ------------------------------------------------------------------------- */

/* Return the total distance from v to all other vertices in g,m,n.
   g must be undirected and connected.  Any self-loops are ignored.
*/
int
nautyextra_total_distance_of_vertex (const graph *g, int m, int n, int v, const set *profile)
{
  assert(n >= 0);
  assert(m >= SETWORDSNEEDED(n));
  assert(0 <= v && v < n);
#if MAXN
  int distances[MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE(int,distances,n,"nautyextra_total_distance_of_vertex");
#endif
  find_dist((graph*)g, m,n, v, distances);
  int total_distance = 0;
  int to;
  if (profile) {
    for (to = -1; (to = nextelement((set*)profile,m,to)) >= 0; )
      total_distance += distances[to];
  } else {
    for (to = 0; to < n; to++)
      total_distance += distances[to];
  }
  return total_distance;
}

/* Return the number of median vertices in g,m,n.
   g must be undirected and connected.  Any self-loops are ignored.

   A median vertex has sum distance to all other vertices which is the
   minimum in the graph.  Multiple vertices can have equal minimum sum.  The
   return is how many such equal minimums.
*/
int
nautyextra_median_size (const graph *g, int m, int n, const set *profile)
{
  assert(n >= 0);
  assert(m >= SETWORDSNEEDED(n));
  int median_total_distance = 0;
  int median_size = 0;

  int v;
  for (v = 0; v < n; v++) {
    int total_distance = nautyextra_total_distance_of_vertex(g,m,n,v,profile);
    if (median_size == 0 || total_distance < median_total_distance) {
      median_size = 0;
      median_total_distance = total_distance;
      goto INCREMENT;
    }
    if (median_total_distance == total_distance) {
    INCREMENT:
      median_size++;
    }
  }
  return median_size;
}

/* If g,m,n is regular, meaning all vertices have the same degree, then
   return that degree.  If not then return -1.
   An empty graph n=0 is reckoned as degree 0.

   Cf nautycliquer.c graph_test_regular() same for graph_t.
*/
int
nautyextra_regular_degree (const graph *g, int m, int n)
{
  assert(n >= 0);
  assert(m >= SETWORDSNEEDED(n));

  if (n == 0) return 0;
  int degree = setsize((set*)g,m);     /* first row */
  while (--n > 0) {
    g += m;
    if (setsize((set*)g,m) != degree) /* second and subsequent row must same */
      return -1;
  }
  return degree;
}

/* ------------------------------------------------------------------------- */

/* Return the Szeged index of g,m,n.
   This is

       Sz =   sum     numcloser(u) * numcloser(v)
           edge u,v

   where numcloser is the number of vertices closer to u than to v, or
   coversely closer to v than u.

   (For a tree this the same as the Wiener index.)
*/
int
nautyextra_Szeged_index (const graph *g, int m, int n)
{
  assert(n >= 0);
  assert(m >= SETWORDSNEEDED(n));
#if MAXN
  int distances[MAXN*MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE(int,distances,n*n, "nautyextra_Szeged_index");
#endif

  nautyextra_distances_matrix (g,m,n, distances);
  const set *u_row = g;
  int *u_distances = distances;
  unsigned long long Sz = 0;
  int u,v,i;
  for (u = 0; u < n; u++,u_row+=m,u_distances+=n) {
    for (v = u; (v = nextelement((set*)u_row,m,v)) >= 0; ) {
      int *v_distances = distances + v*n;
      int u_count = 0;
      int v_count = 0;
      for (i = 0; i < n; i++) {
        int ud = u_distances[i];
        int vd = v_distances[i];
        u_count += (ud > vd);
        v_count += (ud < vd);
      }
      ADD_NO_OVERFLOW_ULL(Sz, Sz, u_count*v_count);
    }
  }
  return Sz;
}

/* ------------------------------------------------------------------------- */
/* Relabelling */

/* Return TRUE if perm[0..n-1] is the identity permutation, or return FALSE
   if not.  The identity permutation is each perm[i]==i.
*/
int
nautyextra_is_idperm (const int *perm, int n)
{
  int i;
  for (i = 0; i < n; i++)
    if (perm[i] != i)
      return FALSE;
  return TRUE;
}

/* Copy g_in to g_out with random vertex relabelling.
   g_in must be different location than g_out, no in-place relabel.

   Cf ranlabg.c static ranrelabel(), which can also hold some vertex numbers
   fixed.
*/
void
nautyextra_ranrelabel (const graph *g_in, int m, int n, graph *g_out)
{
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (g_in != NULL);
  assert (g_out != NULL);
  assert (g_in != g_out);
#if MAXN
  int perm[MAXN];
#else
  WORKSPACE(int,perm,n,"nautyextra_ranrelabel");
#endif

  /* use updatecan() here since it goes in to out, as opposed to relabel()
     which is in-place by going to nauty workspace and back
  */
  ranperm(perm,n);
  updatecan((graph*)g_in,g_out,perm,0,m,n);
}

void
nautyextra_relabel_swap_vpar_copy (const int *vpar_in, int n, int u, int v, int *vpar_out)
{
  /* UNTESTED */

  assert (n >= 0);
  assert (1 <= u && u <= n);
  assert (1 <= v && v <= n);
  int i;
  for (i = 1; i <= n; i++) {
    int ii = (i==u ? v : i==v ? u : i);
    int p = vpar_in[ii];
    int pp = (p==u ? v : p==v ? u : p);
    vpar_out[i] = pp;
  }
}

/* ------------------------------------------------------------------------- */
/* Bandwidth */

/* Return the dilation of graph g,m,n, sparsegraph sg, or tree vpar[1..n].
   This is the biggest abs(u-v) for an edge u,v.
   Or equivalently this is the matrix bandwidth of the adjacency matrix,
   that being the biggest abs(i-j) among non-0 matrix entries m[i,j].
*/
int
nautyextra_dilation (const graph *g, int m, int n)
{
  int u,b;
  for (b=n-1; b>0; b--) {
    for (u=0; u<n; u++) {
      int v=u+b;
      if (ISELEMENT(GRAPHROW(g,u,m), v)
          || ISELEMENT(GRAPHROW(g,v,m), u))
        return b;
    }
  }
  return 0;
}
int
nautyextra_dilation_sg (const sparsegraph *sg)
{
  int n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int b = 0;
  int from;
  for (from = 0; from < n; from++) {
    const int *from_eptr = ee + vv[from];
    int from_degree = dd[from];
    int i;
    for (i=0; i<from_degree; i++) {
      int to = *from_eptr++;
      int delta = abs(from-to);
      b = MAX(b,delta);
    }
  }
  return b;
}
int
nautyextra_dilation_vpar (const int *vpar, int n)
{
  DEBUG(printf("  nautyextra_dilation_vpar() %p n=%d\n", vpar,n));
  int v, b = 0;
  for (v=1; v<=n; v++) {
    int p = vpar[v];
    if (p) {
      int delta = abs(p-v);
      b = MAX(b,delta);
    }
  }
  return b;
}

/* Set perm[1..n] to permutation number num of the integers 1..n, and set
   perm[0]=0.
*/
void
nautyextra_numtoperm1 (NAUTYEXTRA_TYPE_ULL num, int n, int *perm)
{
  int i;
  for (i=0; i<=n; i++)
    perm[i] = i;
  for (i=1; i<n; i++) {
    int d = n-i;
    int r = (num % d);
    num   =  num / d;
    r += i;
    assert (1 <= i && i <= n-1);  /* swap i,r */
    assert (i <= r && r <= n);
    int t = perm[i]; perm[i] = perm[r]; perm[r] = t;
  }
}

/* Set perm[0..n-1] to permutation number num of the integers 0..n-1.
   FIXME: Args other way?
*/
void
nautyextra_numtoperm (NAUTYEXTRA_TYPE_ULL num, int n, int *perm)
{
  int i;
  for (i=0; i<n; i++)
    perm[i] = i;
  for (i=0; i<n-1; i++) {
    int d = n-i;
    int r = (num % d);
    num   =  num / d;
    r += i;
    assert (0 <= i && i < n-1);  /* swap i,r */
    assert (i <= r && r < n);
    int t = perm[i]; perm[i] = perm[r]; perm[r] = t;
  }
}

unsigned long long
nautyextra_factorial (int n)
{
  unsigned long long f = 1;
  int i;
  for (i = 2; i <= n; i++)
    MUL_NO_OVERFLOW_ULL(f, f, i);
  return f;
}

/* Return the bandwidth of graph g,m,n or tree vpar[1..n].
   This is the smallest dilation (nautyextra_dilation_vpar()) possible for
   any labelling of vpar.
*/
int
nautyextra_bandwidth (const graph *g_orig, int m, int n)
{
  DEBUG(printf("nautyextra_bandwidth() n=%d edges %d\n",
               n, nautyextra_num_edges(g_orig,m,n)));

  assert(n >= 0);
#if MAXN
  setword g[MAXN*SETWORDSNEEDED(MAXN)];
  int perm[MAXN];
  assert(n <= MAXN);
  assert(m <= SETWORDSNEEDED(MAXN));
#else
  WORKSPACE2(setword,g, n*m,
             int,perm,n,
             "nautyextra_bandwidth");
#endif
  int b_min = n;
  int num = 0;
  int limit = nautyextra_factorial(n);
  DEBUG(printf("  limit %d  g temp %p perm temp %p\n", limit, g, perm));

  for (num = 0; num < limit; num++) {
    nautyextra_numtoperm (num,n, perm);
    updatecan((graph*)g_orig,g,perm,0,m,n);
    int b = nautyextra_dilation(g,m,n);
    b_min = MIN(b,b_min);
    DEBUG(printf("  dilation %d\n", b));
  }
  DEBUG(printf("  return %d\n", b_min));
  return b_min;
}
int
nautyextra_bandwidth_vpar (const int *vpar_orig, int n)
{
  DEBUG(printf("nautyextra_bandwidth_vpar() n=%d\n", n));

  assert(n >= 0);
#if MAXN
  int vpar[MAXN+1];
  int perm[MAXN+1];
  assert(n <= MAXN);
#else
  WORKSPACE2(int,vpar,n+1,
             int,perm,n+1,
             "nautyextra_bandwidth_vpar");
#endif
  int b_min = n;
  int num = 0;
  int limit = nautyextra_factorial(n);
  DEBUG(printf("  limit %d  vpar temp %p perm temp %p\n", limit, vpar, perm));

  for (num = 0; num < limit; num++) {
    nautyextra_numtoperm1 (num,n, perm);
    nautyextra_relabel_perm_vpar(vpar_orig,n, perm, vpar);
    int b = nautyextra_dilation_vpar(vpar,n);
    b_min = MIN(b,b_min);
    DEBUG(printf("  dilation %d\n", b));
  }
  return b_min;



  /* NOT WORKING ? */

  memcpy(vpar, vpar_orig, (n+1)*sizeof(vpar[0]));
  perm[1] = 1;
  int pos = 1;
  for (;;) {
    if (++perm[pos] > n) {
      /* backtrack */
      if (--pos < 1) break;
      continue;
    }

    nautyextra_relabel_swap_vpar(vpar,n, pos, perm[pos]); /* swap */
    int b = nautyextra_dilation_vpar(vpar,n);
    b_min = MIN(b,b_min);
    nautyextra_relabel_swap_vpar(vpar,n, pos, perm[pos]); /* unswap */
  }
  return b_min;
}

/* ------------------------------------------------------------------------- */

#define NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION(pos,m,expr,end,out)              \
  do {                                                                        \
    setword NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword;                    \
    int mpos;                                                                 \
    if (pos == -1) {                                                          \
      mpos = 0;                                                               \
      NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword = (expr);                 \
    } else {                                                                  \
      mpos = SETWD(pos);                                                      \
      NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword = (expr);                 \
      NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword &= BITMASK(SETBT(pos));   \
    }                                                                         \
    for (;;) {                                                                \
      if (NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword != 0) {               \
        (out) = TIMESWORDSIZE(mpos)                                           \
          + FIRSTBITNZ(NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword);        \
        break;                                                                \
      }                                                                       \
      if (++mpos >= m) {                                                      \
        goto end;                                                             \
      }                                                                       \
    }                                                                         \
  } while (0)

#define NAUTYEXTRA_LOOP_ROWEXPR(m,expr,var,body)                              \
  do {                                                                        \
    setword NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword;                    \
    int mpos = 0;                                                             \
    for (mpos = 0; mpos < m; mpos++) {                                        \
      NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword = (expr);                 \
      while (NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword) {                 \
        TAKEBIT(var,NAUTYEXTRA_NEXTELEMENT_OF_EXPRESSION__setword);           \
        (var) += TIMESWORDSIZE(mpos);                                         \
        body;                                                                 \
      }                                                                       \
    }                                                                         \
  } while (0)



/* ------------------------------------------------------------------------- */

/* sg is an undirected sparsegraph.
   ss is a set of vertices 0 to sg->nv-1.
   Return TRUE if ss is an inverse dominating set in sg.  Return FALSE if not.
*/
int
nautyextra_is_invdomset_sg (const sparsegraph *sg, const set *ss)
{
  DEBUG(printf("nautyextra_is_invdomset_sg() n=%d  %s\n",
               sg->nv, nautyextra_set_str(ss,sg->nv)));
  int n = sg->nv;

  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int v;
  for (v=0; v<n; v++) {
    const int *eptr = ee + vv[v];
    int degree = dd[v];
    int v_in_ss = (ISELEMENT(ss,v) != 0);
    int i;
    for (i=0; i<degree; i++) {
      int to = *eptr++;
      int to_not_in_ss = (ISELEMENT(ss,to) == 0);
      if (v_in_ss == to_not_in_ss)
        goto NEXT_V;  /* neighbour "to" opposite to v, good */
    }
    DEBUG(printf(" v=%d not inverse dominated\n", v));
    return FALSE;
  NEXT_V: ;
  }
  DEBUG(printf("  is invdomset\n"));
  return TRUE;
}

/* sg is an undirected sparsegraph.
   Return the number of inverse dominating sets in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   few vertices, and currently only uses 1-word sets, so must have sg->nv <=
   WORDSIZE many vertices.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_invdomsets_count_sg (const sparsegraph *sg)
{
  setword count = 0;
  setword ss;
  int n = sg->nv;

  assert (n >= 0);
  if (n == 0) return 1;   /* n=0 empty set is only indset */
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_invdomsets_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  NAUTYEXTRA_SET1_LOOP (ss,n, count += nautyextra_is_invdomset_sg (sg,&ss));
  return count;
}

/* Set *info to the inverse domination number and count of sets of this size
   in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   small number of vertices, and currently only uses 1-word sets, so must
   have sg->nv <= WORDSIZE many vertices.
*/
void
nautyextra_invdomnum_and_count_sg (const sparsegraph *sg,
                                   struct nautyextra_invdomnum_and_count *info)
{
  DEBUG(printf("nautyextra_invdomnum_and_count_sg() n=%d\n", sg->nv));

  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) {
    DEBUG(printf(" n=0 special case\n"));
    info->invdomnum = 0; /* empty set of vertices */
    info->count = 1;  /* 1 way */
    return;
  }
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_invdomnum_and_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int invdomnum = n+1;  /* to be bettered by n vertices or fewer */
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     DEBUG(printf(" consider set %s\n", nautyextra_set_str(&ss,n)));
     if (nautyextra_is_invdomset_sg(sg,&ss)) {
       int ss_size = POPCOUNT(ss);
       if (ss_size < invdomnum) {
         DEBUG(printf(" new low %d\n", ss_size));
         invdomnum = ss_size;
         count = 0;
       }
       if (ss_size == invdomnum) {
         count++;
         DEBUG(printf(" count %u\n", count));
       }
     });
  DEBUG(printf(" end invdomnum %d count %u\n", invdomnum, count));
  info->invdomnum = invdomnum;
  info->count = count;
}

char *
nautyextra_canonical_str (const graph *g, int m, int n, int format)
{
#if MAXN
  graph canon_g[MAXN*SETWORDSNEEDED(MAXN)];
  assert(n <= MAXN);
  assert(m <= SETWORDSNEEDED(MAXN));
#else
  WORKSPACE(setword,canon_g,n*m, "nautyextra_canonical_str");
#endif
  nautyextra_canonise (g,m,n, canon_g);
  switch (format) {
  case 0: return ntog6(canon_g,m,n);
  case 1: return ntos6(canon_g,m,n);
  case 2: return ntod6(canon_g,m,n);
  default: return NULL;
  }
}

int
nautyextra_hog_grep(const graph *g, int m, int n)
{
  const char *str = nautyextra_canonical_str(g,m,n,0);
  const char *home = getenv("HOME");
  int buflen = strlen(str) + 128;
  WORKSPACE2(char,filename,strlen(home)+128,
             char,buf,buflen,
             "nautyextra_hog_grep");
  strcpy(filename,home);
  strcat(filename,"/HOG/all.g6");
  FILE *fp = fopen(filename,"r");
  if (fp == NULL) return 0;
  int found = 0;
  while (fgets(buf,buflen,fp) != NULL) {
    if (strcmp(buf,str) == 0) {
      found = 1;
      break;
    }
  }
  fclose(fp);
  return found;
}

/* ------------------------------------------------------------------------- */


/* Set components[1..n] so each components[v] is the component number of
   vertex v in vpar.

   The first component is 1.  The last is nautyextra_num_roots_vpar(vpar).
   Components are ordered by their smallest vertex.  So component 1 contains
   vertex 1, component 2 contains the smallest of the rest, etc.
*/
void
nautyextra_compnum_array_vpar (const int *vpar, int n, int *components)
{
  assert(n >= 0);
  int upto = 1;
  int i;
  for (i = 1; i <= n; i++) {
    int v = i;
    int prev = 0;
    while (v && !components[v]) {
      components[v] = prev; prev = v; v = vpar[v];
    }
    int c = (v ? components[v] : upto++);
    assert (1 <= c && c < upto);
    while (prev) {
      v = components[prev]; components[prev] = c; prev = v;
    }
  }
}








/* ------------------------------------------------------------------------- */
/* Disjoint Domination Number */

int
nautyextra_disjoint_domnum_vpar (const int *vpar, int n)
{
  int ret = 0;
  int i;
  struct data {
    int  num_children;
    int  with_dom;
    int  with_undom;
    int  without_dom_dom;
    int  without_dom_undom;
    int  without_undom_undom;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_disjoint_domnum_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_domnum_vpar() ");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with_dom            =
      data[i].without_dom_dom   =
      data[i].without_dom_undom = n+1;
    data[i].with_undom          = 1;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children) continue;

    for (;;) {
      DEBUG(printf("consider v=%d (parent %d)\n", v, vpar[v]));
      assert (data[v].num_children == 0);

      int v_dom = MIN(data[v].with_dom, data[v].without_dom_dom);
      data[v].num_children = -1;
      int p = vpar[v];     /* parent */
      if (! p) {
        DEBUG(printf("  no parent\n"));
        ret += v_dom;
        if (ret > n) return -1;  /* top singleton, no sets */
        break;
      }

      int v_with = MIN(data[v].with_dom, data[v].with_undom);
      {
        int extend = data[p].with_dom + MIN(v_with, data[v].without_dom_undom);
        int build  = data[p].with_undom + v_with;
        data[p].with_dom = MIN(extend, build);
      }
      data[p].with_undom = data[p].with_undom + MIN(data[v].with_dom,
                                                    data[v].without_dom_undom);
      {
        int extend = data[p].without_dom_dom + v_dom;
        int build  = data[p].without_dom_undom + data[v].with_dom;
        data[p].without_dom_dom = MIN(extend, build);
      }
      {
        int extend = data[p].without_dom_undom + v_dom;
        int build  = data[p].without_undom_undom + data[v].with_dom;
        data[p].without_dom_undom = MIN(extend, build);
      }
      data[p].without_undom_undom += data[v].without_dom_dom;

      /* DEBUG(print(" to p with "with_dom[p]" "with_undom[p]"  without "without_dom_dom[p]" "without_dom_undom[p]" "without_undom_undom[p])); */

      v = p;
      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d\n", v));
        break;
      }
    }
  }

  DEBUG(printf("  final %d\n", ret));
  return ret;
}

void
nautyextra_disjoint_domnum_and_count_vpar (const int *vpar, int n, struct nautyextra_disjoint_domnum_and_count *info)
{
  int ret = 0;
  NAUTYEXTRA_TYPE_ULL ret_count = 1;
  int i;
  struct data {
    int  num_children;
    int  with_dom;
    int  with_undom;
    int  without_dom_dom;
    int  without_dom_undom;
    int  without_undom_undom;
    NAUTYEXTRA_TYPE_ULL  with_dom_count;
    NAUTYEXTRA_TYPE_ULL  with_undom_count;
    NAUTYEXTRA_TYPE_ULL  without_dom_dom_count;
    NAUTYEXTRA_TYPE_ULL  without_dom_undom_count;
    NAUTYEXTRA_TYPE_ULL  without_undom_undom_count;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_disjoint_domnum_and_count_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_domnum_vpar() ");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with_dom            =
      data[i].without_dom_dom   =
      data[i].without_dom_undom = n+1;
    data[i].with_undom          = 1;
    data[i].with_undom_count =
      data[i].without_undom_undom_count = 1;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children) continue;

    for (;;) {
      DEBUG(printf("consider v=%d parent %d\n", v, vpar[v]);
            printf("  with %d[%Lu] %d[%Lu] without %d[%Lu] %d[%Lu] %d[%Lu]\n",
                   data[v].with_dom, data[v].with_dom_count,
                   data[v].with_undom, data[v].with_undom_count,
                   data[v].without_dom_dom, data[v].without_dom_dom_count,
                   data[v].without_dom_undom, data[v].without_dom_undom_count,
                   data[v].without_undom_undom, data[v].without_undom_undom_count));
      assert (data[v].num_children == 0);

      int                 v_dom;
      NAUTYEXTRA_TYPE_ULL v_dom_count;
      if (data[v].with_dom < data[v].without_dom_dom) {
        v_dom       = data[v].with_dom;
        v_dom_count = data[v].with_dom_count;
      } else {
        v_dom       = data[v].without_dom_dom;
        v_dom_count = data[v].without_dom_dom_count;
        if (data[v].with_dom == data[v].without_dom_dom)
          ADD_NO_OVERFLOW_ULL(v_dom_count,
                              v_dom_count, data[v].with_dom_count);
      }
      DEBUG(printf("  v_dom %d[%Lu]\n", v_dom, v_dom_count));

      data[v].num_children = -1;
      int p = vpar[v];     /* parent */
      if (! p) {
        DEBUG(printf("  no parent\n"));
        if (v_dom > n) {
          info->disjoint_domnum = -1;     /* top singleton, no sets */
          info->count = 0;
          return;
        }
        MUL_NO_OVERFLOW_ULL(ret_count, ret_count, v_dom_count);
        if (ret)   /* double for second and subsequent */
          ADD_NO_OVERFLOW_ULL(ret_count, ret_count,ret_count);
        ret += v_dom;
        break;
      }

      int                 v_with;
      NAUTYEXTRA_TYPE_ULL v_with_count;
      if (data[v].with_dom < data[v].with_undom) {
        v_with       = data[v].with_dom;
        v_with_count = data[v].with_dom_count;
      } else {
        v_with       = data[v].with_undom;
        v_with_count = data[v].with_undom_count;
        if (data[v].with_dom == data[v].with_undom)
          ADD_NO_OVERFLOW_ULL(v_with_count,
                              v_with_count, data[v].with_dom_count);
      }

      /*
        with_dom[p]
        =
        .      with_dom[p]                  OR          with_undom[p]
        .  +  undom  dom without                     +  dom  with
        or +   dom   dom without  \ each way      or + undom with
        or +   dom   dom without  /
        or +  undom  with
        or +   dom   with         \ each way
        or +  with   dom          /
      */
      {
        int extend = MIN(v_with, data[v].without_dom_undom);
        data[p].with_dom += extend;
        int prospective = data[p].with_undom + v_with;
        if (data[p].with_dom <= prospective) {
          NAUTYEXTRA_TYPE_ULL extend_count = 0;
          if (extend == data[v].without_dom_undom)
            extend_count = data[v].without_dom_undom_count;
          if (extend == data[v].without_dom_dom) {
            ADD_NO_OVERFLOW_ULL(extend_count,
                                extend_count, data[v].without_dom_dom_count);
            ADD_NO_OVERFLOW_ULL(extend_count,
                                extend_count, data[v].without_dom_dom_count);
          }
          if (extend == data[v].with_undom)
            ADD_NO_OVERFLOW_ULL(extend_count,
                                extend_count, data[v].with_undom_count);
          if (extend == data[v].with_dom) {
            ADD_NO_OVERFLOW_ULL(extend_count,
                                extend_count, data[v].with_dom_count);
            ADD_NO_OVERFLOW_ULL(extend_count,
                                extend_count, data[v].with_dom_count);
          }
          MUL_NO_OVERFLOW_ULL(data[p].with_dom_count,
                              data[p].with_dom_count, extend_count);
        }

        if (prospective <= data[p].with_dom) {
          NAUTYEXTRA_TYPE_ULL prospective_count;
          MUL_NO_OVERFLOW_ULL(prospective_count,
                              data[p].with_undom_count, v_with_count);
          if (prospective < data[p].with_dom) {
            data[p].with_dom       = prospective;
            data[p].with_dom_count = prospective_count;
          } else {
            ADD_NO_OVERFLOW_ULL(data[p].with_dom_count,
                                data[p].with_dom_count, prospective_count);
          }
        }
      }

      /*
        with_undom[p] = with_undom[p] + min(with_dom[v],without_dom_undom[v])

        with_undom[p]
        =
        .      with_undom[p]
        .  +  undom  dom without
        or +   dom   dom without   \ each way
        or +   dom   dom without   /
        or +  with   dom
      */
      {
        int extend = MIN(data[v].with_dom, data[v].without_dom_undom);
        data[p].with_undom += extend;

        NAUTYEXTRA_TYPE_ULL extend_count = 0;
        if (extend == data[v].without_dom_dom) {
          ADD_NO_OVERFLOW_ULL(extend_count,
                              data[v].without_dom_dom_count,
                              data[v].without_dom_dom_count);
        }
        if (extend == data[v].without_dom_undom)
          ADD_NO_OVERFLOW_ULL(extend_count,
                              extend_count, data[v].without_dom_undom_count);
        if (extend == data[v].with_dom)
          ADD_NO_OVERFLOW_ULL(extend_count,
                              extend_count, data[v].with_dom_count);

        MUL_NO_OVERFLOW_ULL(data[p].with_undom_count,
                            data[p].with_undom_count, extend_count);
      }

      /*
        without_dom_dom[p]
        =                                          (prospective)
        without_dom_dom[p]                OR   without_dom_undom[p]
        .    +   dom  dom  \ each way              +   dom  with
        .    +   dom  dom  /
        . or +  with  dom  \ each way
        . or +  dom  with  /
      */
      {
        data[p].without_dom_dom  += v_dom;

        int prospective = data[p].without_dom_undom + data[v].with_dom;
        if (data[p].without_dom_dom <= prospective) {
          NAUTYEXTRA_TYPE_ULL extend_count;
          ADD_NO_OVERFLOW_ULL(extend_count,
                              v_dom_count, v_dom_count);
          MUL_NO_OVERFLOW_ULL(data[p].without_dom_dom_count,
                              data[p].without_dom_dom_count, extend_count);
        }
        if (prospective <= data[p].without_dom_dom) {
          NAUTYEXTRA_TYPE_ULL prospective_count;
          MUL_NO_OVERFLOW_ULL(prospective_count,
                              data[p].without_dom_undom_count,
                              data[v].with_dom_count);
          if (prospective < data[p].without_dom_dom) {
            data[p].without_dom_dom       = prospective;
            data[p].without_dom_dom_count = prospective_count;
          } else {
            ADD_NO_OVERFLOW_ULL(data[p].without_dom_dom_count,
                                data[p].without_dom_dom_count,
                                prospective_count);
          }
        }
      }

      /*
        without_dom_undom[p]
        =
        without_dom_undom[p]               OR     without_undom_undom[p]
        .    +  dom  dom  without \ each              +    with dom [v]
        . or +  dom  dom  without / way           *2 if not empty
        . or +  with dom

        without_dom_undom[p] = min(without_dom_undom[p] + v_dom,
        .                          without_undom_undom[p] + with_dom[v])
      */
      {
        data[p].without_dom_undom += v_dom;

        int prospective = data[p].without_undom_undom + data[v].with_dom;
        if (data[p].without_dom_undom <= prospective) {
          NAUTYEXTRA_TYPE_ULL extend_count = v_dom_count;
          if (v_dom == data[v].without_dom_dom)
            ADD_NO_OVERFLOW_ULL(extend_count,
                                extend_count, data[v].without_dom_dom_count);
          MUL_NO_OVERFLOW_ULL(data[p].without_dom_undom_count,
                              data[p].without_dom_undom_count, extend_count);
        }

        if(prospective <= data[p].without_dom_undom) {
          NAUTYEXTRA_TYPE_ULL prospective_count;
          MUL_NO_OVERFLOW_ULL(prospective_count,
                              data[p].without_undom_undom_count,
                              data[v].with_dom_count);
          if (data[p].without_undom_undom)
            ADD_NO_OVERFLOW_ULL(prospective_count,
                                prospective_count,prospective_count);

          if (prospective < data[p].without_dom_undom) {
            data[p].without_dom_undom       = prospective;
            data[p].without_dom_undom_count = prospective_count;
          } else {
            ADD_NO_OVERFLOW_ULL(data[p].without_dom_undom_count,
                                data[p].without_dom_undom_count,
                                prospective_count);
          }
        }
      }

      data[p].without_undom_undom       += data[v].without_dom_dom;
      data[p].without_undom_undom_count *= data[v].without_dom_dom_count;

      DEBUG(printf("  sets parent %d[%Lu] %d[%Lu] without %d[%Lu] %d[%Lu] %d[%Lu]\n",
                   data[p].with_dom, data[p].with_dom_count,
                   data[p].with_undom, data[p].with_undom_count,
                   data[p].without_dom_dom, data[p].without_dom_dom_count,
                   data[p].without_dom_undom, data[p].without_dom_undom_count,
                   data[p].without_undom_undom, data[p].without_undom_undom_count));

      v = p;
      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d\n", v));
        break;
      }
    }
  }
  DEBUG(printf("  final %d[%Lu]\n", ret, ret_count));
  info->disjoint_domnum = ret;
  info->count           = ret_count;
}


/* ------------------------------------------------------------------------- */

/* Print the parent vertex list vpar[1..n] in a form suitable for input to
   Pari/GP interpreter.
*/
void
nautyextra_print_vpar_gp (const int *vpar, int n)
{
  int v;
  assert (n >= 0);

  const char *join = "[";
  for (v=1; v<=n; v++) {
    printf("%s%d", join, vpar[v]);
    join = ",";
  }
  putchar(']');
}

/* ------------------------------------------------------------------------- */
/* Semi-Total Domination */

/* sg is an undirected sparsegraph.
   ss is a set of vertices 0 to sg->nv-1.
   Return TRUE if ss is a semi-total dominating set in sg.  Return FALSE if not.
*/
int
nautyextra_is_semitotdomset_sg (const sparsegraph *sg, const set *ss)
{
  DEBUG(printf("nautyextra_is_semitotdomset_sg() n=%d  %s\n",
               sg->nv, nautyextra_set_str(ss,sg->nv)));
  int n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int v;
  for (v=0; v<n; v++) {
    const int *eptr = ee + vv[v];
    int v_degree = dd[v];
    int v_in_ss = (ISELEMENT(ss,v) != 0);
    int v_i;
    for (v_i=0; v_i<v_degree; v_i++) {
      int to = *eptr++;
      int to_in_ss = ISELEMENT(ss,to);
      if (to_in_ss)
        goto NEXT_V;  /* v has neighbour in set */

      if (v_in_ss) {
        int to_degree = dd[to];
        int to_i;
        const int *to_eptr = ee + vv[to];
        for (to_i=0; to_i<to_degree; to_i++) {
          int second = *to_eptr++;
          if (second==v) continue;
          if (ISELEMENT(ss,second))
            goto NEXT_V;  /* v in set has neighbour 2 away */
        }
      }
    }
    DEBUG(printf(" v=%d not semi-total dominated\n", v));
    return FALSE;
  NEXT_V: ;
  }
  DEBUG(printf("  is semitotdomset\n"));
  return TRUE;
}

/* Set *info to the semi-total domination number and count of sets of this
   size in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   small number of vertices, and currently only uses 1-word sets, so must
   have sg->nv <= WORDSIZE many vertices.
*/
void
nautyextra_semitotdomnum_and_count_sg (const sparsegraph *sg,
                                       struct nautyextra_semitotdomnum_and_count *info)
{
  DEBUG(printf("nautyextra_semitotdomnum_and_count_sg() n=%d\n", sg->nv));

  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) {
    DEBUG(printf(" n=0 special case\n"));
    info->semitotdomnum = 0; /* empty set of vertices */
    info->count = 1;  /* 1 way */
    return;
  }
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_semitotdomnum_and_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int semitotdomnum = n+1;  /* to be bettered by n vertices or fewer */
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     DEBUG(printf(" consider set %s\n", nautyextra_set_str(&ss,n)));
     if (nautyextra_is_semitotdomset_sg(sg,&ss)) {
       int ss_size = POPCOUNT(ss);
       if (ss_size < semitotdomnum) {
         DEBUG(printf(" new low %d\n", ss_size));
         semitotdomnum = ss_size;
         count = 0;
       }
       if (ss_size == semitotdomnum) {
         count++;
         DEBUG(printf(" count %u\n", count));
       }
     });
  DEBUG(printf(" end semitotdomnum %d count %u\n", semitotdomnum, count));
  if (count == 0)
    semitotdomnum = -1;
  info->semitotdomnum = semitotdomnum;
  info->count = count;
}

/* Return the total domination number of tree vpar[1..n].
   vpar can be a forest, in which case the result is sum of total domination
   numbers of its component trees.
   An isolated vertex cannot be total dominated.  If any such in vpar then
   return -1.
 */
int
nautyextra_semitotdomnum_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_semitotdomnum_vpar() ");
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  struct data {
    int  num_children;
    int  with_dom;
    int  with_undom;
    int  without_dom_child_dom;
    int without_dom_child_undom;
    int without_undom;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_semitotdomnum_vpar");
#endif
  int ret = 0;
  int i;
  int none = n+1;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with_dom
      = data[i].without_dom_child_dom
      = data[i].without_dom_child_undom = none;
    data[i].with_undom = 1;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf(" v=%d num_children %d  with %d %d without %d %d %d\n", v, data[v].num_children,
                 data[v].with_dom, data[v].with_undom,
                 data[v].without_dom_child_dom, data[v].without_dom_child_undom, data[v].without_undom));
    if (data[v].num_children != 0) continue;

    for (;;) {
      DEBUG(printf(" consider v=%d (parent %d)\n",
                   v, vpar[v]));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      int p = vpar[v];     /* parent */
      if (! p) {
        int d = MIN(data[v].with_dom, data[v].without_dom_child_dom);
        if (d >= none) return -1;
        ret += d;
        break;
      }

      int v_with        = MIN(data[v].with_dom, data[v].with_undom);
      int v_without_dom = MIN(data[v].without_dom_child_dom, data[v].without_dom_child_undom);

      data[p].with_dom = MIN(MIN(data[p].with_dom, data[p].with_undom)
                             + MIN(v_with, v_without_dom),
                             data[p].with_dom + data[v].without_undom);

      data[p].with_undom += data[v].without_undom;

      data[p].without_dom_child_dom
        = MIN3(data[p].without_dom_child_dom   + MIN(v_with,
                                                     data[v].without_dom_child_dom),
               data[p].without_dom_child_undom + v_with,
               data[p].without_undom           + data[v].with_dom);

      data[p].without_dom_child_undom
        = MIN(data[p].without_dom_child_undom + data[v].without_dom_child_dom,
              data[p].without_undom           + data[v].with_undom);

      data[p].without_undom += data[v].without_dom_child_dom;

      v = p;
      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;
    }
  }

  DEBUG(printf(" final %d\n", ret));
  return ret;
}


/* ------------------------------------------------------------------------- */
/* Perfect Domination */

/* sg is an undirected sparsegraph.
   ss is a set of vertices 0 to sg->nv-1.
   Return TRUE if ss is a perfect dominating set in sg.  Return FALSE if not.
*/
int
nautyextra_is_perfect_domset_sg (const sparsegraph *sg, const set *ss)
{
  DEBUG(printf("nautyextra_is_perfect_domset_sg() n=%d  %s\n",
               sg->nv, nautyextra_set_str(ss,sg->nv)));
  int n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int v;
  for (v=0; v<n; v++) {
    if (ISELEMENT(ss,v)) continue;
    const int *eptr = ee + vv[v];
    int degree = dd[v];
    int i, seen=0;
    for (i=0; i<degree; i++) {
      int to = *eptr++;
      if (ISELEMENT(ss,to))
        if (++seen >= 2)
          return FALSE;  /* 2 or more */
    }
    if (! seen)
      return FALSE;  /* not dominated */
  }
  DEBUG(printf("  is perfect_domset\n"));
  return TRUE;
}

/* Set *info to the perfect domination number and count of sets of this size
   in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   small number of vertices, and currently only uses 1-word sets, so must
   have sg->nv <= WORDSIZE many vertices.
*/
void
nautyextra_perfect_domnum_and_count_sg (const sparsegraph *sg,
                                       struct nautyextra_perfect_domnum_and_count *info)
{
  DEBUG(printf("nautyextra_perfect_domnum_and_count_sg() n=%d\n", sg->nv));

  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) {
    DEBUG(printf(" n=0 special case\n"));
    info->perfect_domnum = 0; /* empty set of vertices */
    info->count = 1;  /* 1 way */
    return;
  }
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_perfect_domnum_and_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int perfect_domnum = n+1;  /* to be bettered by n vertices or fewer */
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     DEBUG(printf(" consider set %s\n", nautyextra_set_str(&ss,n)));
     if (nautyextra_is_perfect_domset_sg(sg,&ss)) {
       int ss_size = POPCOUNT(ss);
       if (ss_size < perfect_domnum) {
         DEBUG(printf(" new low %d\n", ss_size));
         perfect_domnum = ss_size;
         count = 0;
       }
       if (ss_size == perfect_domnum) {
         count++;
         DEBUG(printf(" count %u\n", count));
       }
     });
  DEBUG(printf(" end perfect_domnum %d count %u\n", perfect_domnum, count));
  info->perfect_domnum = perfect_domnum;
  info->count = count;
}

/* Return the perfect domination number of tree vpar[1..n].
   vpar can be a forest, in which case the result is sum of total domination
   numbers of its component trees.
   An isolated vertex cannot be total dominated.  If any such in vpar then
   return -1.
 */
int
nautyextra_perfect_domnum_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_perfect_domnum_vpar() ");
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  struct data {
    int  num_children;
    int  with;
    int  without_dom;
    int  without_undom;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_perfect_domnum_vpar");
#endif
  int ret = 0;
  int i;
  int none = n+1;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].without_dom = none;
    data[i].with = 1;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf(" v=%d num_children %d  with %d without %d %d\n",
                 v, data[v].num_children,
                 data[v].with,
                 data[v].without_dom,
                 data[v].without_undom));
    if (data[v].num_children != 0) continue;

    for (;;) {
      DEBUG(printf(" consider v=%d (parent %d)\n",
                   v, vpar[v]));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      int p = vpar[v];     /* parent */
      if (! p) {
        ret += MIN(data[v].with, data[v].without_dom);
        break;
      }
      data[p].without_dom   = MIN(data[p].without_dom + data[v].without_dom,
                                  data[p].without_undom + data[v].with);
      data[p].without_undom = MIN(data[p].without_undom + data[v].without_dom,
                                  none);
      data[p].with          += MIN(data[v].with, data[v].without_undom);

      v = p;
      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;
    }
  }

  DEBUG(printf(" final %d\n", ret));
  return ret;
}


/* ------------------------------------------------------------------------- */
/* Asymmetric Graphs */

/* Return TRUE if graph g,m,n is asymmetric, or return FALSE if not.

   Asymmetric means its automorphism group comprises solely the identity
   permutation.  For that reason this is also called an identity graph, but
   the name "asymmetric" is preferred here to avoid confusion with other
   ideas of identity.

   FIXME: What to do with directedness?
*/
int
nautyextra_is_asymmetric(const graph *g, int m, int n)
{
  static DEFAULTOPTIONS_DENSEGRAPH(options);
  options.digraph  = ! nautyextra_is_undirected(g,m,n);
#if MAXN
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  assert(n <= MAXN);
#else
  WORKSPACE3(int,lab,n,
             int,ptn,n,
             int,orbits,n,
             "nautyextra_is_asymmetric");
#endif
  assert(m >= SETWORDSNEEDED(n));

  statsblk stats;
  densenauty ((graph*)g, lab, ptn, orbits, &options, &stats, m,n, NULL);
  return stats.numorbits == n;
}


/* ------------------------------------------------------------------------- */
/* Stable */

/* Modify g to disconnect vertex v.
   Any edges to or from v are deleted.
 */
void
nautyextra_disconnect_vertex (graph *g, int m, int n, int v)
{
  EMPTYSET(GRAPHROW(g,v,m), m);
  g += SETWD(v);
  setword mask = ~BITT[SETBT(v)];
  int i;
  for (i = 0; i < n; i++, g+=m)   *g &= mask;
}

static TLS_ATTR struct {
  const graph *g;
  int m;
  int good;
} nautyextra_is_semistable_at_vertex__data;

static void
nautyextra_is_semistable_at_vertex__userautomproc (int count, int *perm, int *orbits, int numorbits, int stabvertex, int n)
{
  DEBUG(printf(" userautomproc count=%d\n", count));
  if (nautyextra_is_semistable_at_vertex__data.good
      && ! isautom ((graph*) nautyextra_is_semistable_at_vertex__data.g,
                    perm,
                    0, /* digraph */
                    nautyextra_is_semistable_at_vertex__data.m, n)) {
    DEBUG(printf("  not an automorphism\n"));
    nautyextra_is_semistable_at_vertex__data.good = 0;
  }
}
static int
nautyextra_is_semistable_at_vertex__usercanonproc (graph *g, int *lab, graph *canong, int count, int code, int m, int n)
{
  /* return 1 to stop processing, if bad */
  int ret = 1 - nautyextra_is_semistable_at_vertex__data.good;
  DEBUG(printf(" usercanonproc return %d\n", ret));
  return ret;
}

/* Return 1 if graph g,m,n is semi-stable at vertex v.
   This means automorphisms of g with v deleted are the same as the
   automorphisms of g which have v fixed, so that deleting v does not
   introduce new automorphisms.

   The current implementation doesn't delete v but disconnects it (zap edges
   to and from) and colours it to avoid automorphisms with any other
   singletons.  The automorphisms of this are the same as with v deleted,
   but still in the numbering of g.  If any of the generators found are not
   automorphisms of g then v deleted has introduced more.  Roughly speaking,
   edges to v may be causing neighbouring vertices or blocks to be different
   from other parts of the graph, which when deleted become the same.

   Not re-entrant.
*/
int
nautyextra_is_semistable_at_vertex (const graph *g, int m, int n, int v)
{
  DEBUG (printf("nautyextra_is_semistable_at_vertex() n=%d v=%d\n", n, v));
  DEBUG (nautyextra_print_matrix(g,m,n));
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (m == 1);
  assert (v >= 0);
  assert (v < n);

  static DEFAULTOPTIONS_DENSEGRAPH(options);
  options.defaultptn = FALSE;
  options.userautomproc = nautyextra_is_semistable_at_vertex__userautomproc;
  options.usercanonproc = nautyextra_is_semistable_at_vertex__usercanonproc;
  statsblk stats;

  DYNALLSTAT(graph,fix_g,fix_g_sz);
  DYNALLOC2(graph,fix_g,fix_g_sz,m,n, "nautyextra_is_semistable_at_vertex");

  DYNALLSTAT(int, lab,    lab_sz);
  DYNALLSTAT(int, ptn,    ptn_sz);
  DYNALLSTAT(int, orbits, orbits_sz);
  DYNALLOC1(int, lab,    lab_sz,    n, "nautyextra_is_semistable_at_vertex");
  DYNALLOC1(int, ptn,    ptn_sz,    n, "nautyextra_is_semistable_at_vertex");
  DYNALLOC1(int, orbits, orbits_sz, n, "nautyextra_is_semistable_at_vertex");

  memcpy(fix_g, g, m*n*sizeof(setword));
  nautyextra_disconnect_vertex (fix_g,m,n, v);
  DEBUG (printf("disconnected v\n"));
  DEBUG (nautyextra_print_matrix(fix_g,m,n));

  lab[0] = v;     /* separate partition for v */
  {
    int pos = 1;
    int i;
    for (i = 0; i < v; i++)   lab[pos++] = i;
    for (i = v+1; i < n; i++) lab[pos++] = i;
  }
  {
    int i;
    ptn[0] = 0;
    for (i = 1; i < n-1; i++)  ptn[i] = 1;
    ptn[n-1] = 0;
  }

  nautyextra_is_semistable_at_vertex__data.g = g;
  nautyextra_is_semistable_at_vertex__data.m = m;
  nautyextra_is_semistable_at_vertex__data.good = 1;
  densenauty(fix_g, lab,ptn,orbits, &options,&stats, m,n, NULL);

  DEBUG(printf(" semistable return %d\n",
               nautyextra_is_semistable_at_vertex__data.good));
  return nautyextra_is_semistable_at_vertex__data.good;
}

/* Return 1 if graph g,m,n is completely semi-stable.
   This means it is nautyextra_is_semistable_at_vertex() at every vertex of g.

   Not re-entrant.
*/
int
nautyextra_is_completely_semistable (const graph *g, int m, int n)
{
  int v;
  for (v=0; v<n; v++)
    if (! nautyextra_is_semistable_at_vertex (g,m,n,v))
      return 0;
  return 1;
}

/* Find a maximal stabilizing sequence in graph g,m,n.
   The return is the length of the sequence found.  This is the stability
   index of the graph.
   seq[0..ret-1] are vertices which can be removed from g successively, in
   order, without introducing new automorphisms at any stage.
   seq can be NULL to just return the length.
   There might be many maximal stabilizing sequences.  The exact one
   returned is unspecified.

   The current implementation is an exhaustive search over all stabilizing
   sequences looking for either ret==n which is g stable and a length n
   sequence, or otherwise for the maximum length in the search.  This is not
   very efficient.

   Not re-entrant.

   ENHANCE-ME: Watch for removal-similar vertices and don't descend again.
*/
int
nautyextra_max_stabilizing_sequence (const graph *g, int m, int n, int *seq)
{
  DEBUG (printf("nautyextra_max_stabilizing_sequence() g=%p m=%d n=%d seq=%p\n",
                g,m,n, seq));
  DEBUG (nautyextra_print_matrix(g,m,n));
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));
  assert (m == 1);
  if (n==0) return 1;

  static DEFAULTOPTIONS_DENSEGRAPH(options);
  options.defaultptn = FALSE;
  options.userautomproc = nautyextra_is_semistable_at_vertex__userautomproc;
  options.usercanonproc = nautyextra_is_semistable_at_vertex__usercanonproc;
  statsblk stats;

  int g_words = m*n;
  DYNALLSTAT(graph,fix_g,fix_g_sz);
  DYNALLOC1(graph,fix_g,fix_g_sz,n*g_words, "nautyextra_max_stabilizing_sequence");
  DEBUG (printf(" fix_g at %p [%d],  g_words = %d\n", fix_g, fix_g_sz, g_words));

  int *try    = alloca(4*n*sizeof(int));
  int *lab    = try + n;
  int *ptn    = lab + n;
  int *orbits = ptn + n;

  nautyextra_is_semistable_at_vertex__data.m = m;
  int ret = 0;
  try[0] = -1;
  int pos = 0;
  for (;;) {
  AGAIN:
    DEBUG (printf(" at pos=%d\n", pos));
    memset(orbits, '\0', n*sizeof(int));
    int i;
    for (i=0; i < pos; i++) {
      int v = try[i];
      assert (0 <= v && v < n);
      orbits[v] = 1;
    }

    int v;
    for (;;) {
      assert (0 <= pos && pos < n);
      v = ++try[pos];
      if (v >= n) {
        DEBUG (printf("  backtrack\n"));
        if (--pos < 0) goto DONE;
        goto AGAIN;
      }
      if (! orbits[v]) break;
    }
    DEBUG (printf("  try v=%d\n", v));
    assert (0 <= v && v < n);
    orbits[v] = 1;
    DEBUG(printf("  used flags "); for(i=0;i<n;i++) printf("%d,",orbits[i]); printf("\n"));

    graph *this_g = fix_g + pos*n*m;
    const graph *prev_g = pos==0 ? g : this_g - g_words;
    DEBUG (printf("  this_g = %p, copy from prev_g = %p\n", this_g, prev_g));
    nautyextra_is_semistable_at_vertex__data.g = prev_g;
    memcpy(this_g, prev_g, g_words*sizeof(setword));
    nautyextra_disconnect_vertex (this_g,m,n, v);
    DEBUG (printf("  disconnected matrix\n"));
    DEBUG (nautyextra_print_matrix(this_g,m,n));

    int upto;
    for (upto = 0; upto <= pos; upto++) {
      assert (0 <= upto && upto < n);
      ptn[upto] = 0;
      lab[upto] = try[upto];
    }
    assert (upto == pos+1);
    for (i=0; i < n; i++) {
      if (!orbits[i]) {
        assert (0 <= upto && upto < n);
        lab[upto] = i;
        ptn[upto] = 1;
        upto++;
      }
    }
    ptn[n-1] = 0;
    DEBUG(printf("  lab "); for(i=0;i<n;i++) printf("%d,",lab[i]); printf("\n"));
    DEBUG(printf("  ptn "); for(i=0;i<n;i++) printf("%d,",ptn[i]); printf("\n"));
    assert (upto == n);

    nautyextra_is_semistable_at_vertex__data.good = 1;
    densenauty(this_g, lab,ptn,orbits, &options,&stats, m,n, NULL);

    if (nautyextra_is_semistable_at_vertex__data.good) {
      pos++;
      if (pos > ret) {
        DEBUG (printf("  new longest\n"));
        ret = pos;
        if (seq) memcpy(seq, try, ret*sizeof(int));
      }
      if (pos >= n) {
        DEBUG (printf("  is stable\n"));
        break;
      }
      DEBUG (printf("  descend\n"));
      assert (0 <= pos && pos < n);
      try[pos] = -1;
    } else {
      DEBUG (printf("  not\n"));
    }
  }
 DONE:
  DEBUG(printf(" return %d\n", ret));
  return ret;
}

/* Return 1 if graph g,m,n is stable.
   This means there exists some ordering of the vertices where removing each
   successively does not introduce any new automorphisms in what remains.

   Not re-entrant.
*/
int
nautyextra_is_stable (const graph *g, int m, int n)
{
  return nautyextra_max_stabilizing_sequence(g,m,n,NULL) == n;
}

  
/* ------------------------------------------------------------------------- */
/*
  Maybe:
  union_vpar
  join_vpar
  nautyextra_relabel_depth_first_vpar(vpar_in,n, x,y, vpar_out)
  nautyextra_relabel_breadth_first_vpar(vpar,n, x,y)
  nautyextra_relabel_preorder_vpar(vpar,n, x,y)
  nautyextra_find_dist_vpar(vpar,n, v, *dist)
*/
